/* Copyright 2006, 2007 Columbia University */
/* This file is part of THINC.

    THINC is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 2 of the License, or
    (at your option) any later version.

    THINC is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with THINC.  If not, see <http://www.gnu.org/licenses/>.
*/
import java.awt.*;
import java.awt.geom.AffineTransform;
import java.awt.image.*;
import java.nio.ByteBuffer;
import java.nio.IntBuffer;
import java.util.Hashtable;
import javax.swing.JPanel;
/**
 * Canvas concept to draw incoming messages from network. extended from JPanel class from java. 
 * @author  Young Jin Yoon
 * @since  2006
 * @version  0.3
 */

public class ThincClientCanvas extends JPanel{
	/**
	 * serialVersionUID generated from compiler
	 */
	private static final long serialVersionUID = 2056853419545819891L;

	public static final String CURSOR_NAME = "THINC_CURSOR";

	/* flags for cache operation */
	public static final int IMG_CACHE = 0;
	public static final int PIX_CACHE = 1;
	public static final int BIT_CACHE = 2;
	private static final int NUM_CACHE = 3;
	
	/* flags for drawFrameBuffer() method */
	public static final int TYPE_RAW = 0;
	public static final int TYPE_PNG = 1;
	public static final int TYPE_BITMAP = 2;
	public static final int TYPE_TRANSPARENT = 3;

	/* flags for drawFrameBufferScaled method */
	public static final int TYPE_VIDEO = 4;
	public static final int TYPE_SCALED_BITMAP = 5;
	public static final int TYPE_SCALED_RAW = 6;
	public static final int TYPE_SCALED_TRANSPARENT = 7;

	/* pointers to refer another object */
	private ThincClient client;
	private ThincMsgHandler handler;
	private ThincFrame frame;
	
	/* cache implementation used with Hashtable in java */
	private Hashtable cache[] = new Hashtable[NUM_CACHE];
	private int cacheMask[] = new int[NUM_CACHE];
	
	/* for drawing operation */
	private int[] rawPixelIntArray;
	private IntBuffer rawPixelIntBuffer;
	private MemoryImageSource rawPixel  = null;
	private Image rawPixelImage = null;
	
	/* for cursor operation */
	private int[] cursorIntArray;
	private MemoryImageSource cursor = null;
	private Image cursorImage = null;
	private Point cursorHot = null;
	private Point currentCursor = null;
	private int fgCursor = 0;
	private int bgCursor = 0;
	private int widthCursor = 0;
	private int heightCursor = 0;
	private byte[] bgbCursor;
	private byte[] fgbCursor;
	private boolean isARGB = false;

	/* invisible cursors */
	private int[] pixels = new int[16 * 16];
	private Image invCursor = Toolkit.getDefaultToolkit().createImage(
			new MemoryImageSource(16, 16, pixels, 0, 16));
	private boolean cursorVisible = false;
	
	/* BufferedImage object which contains entire screen pixels.
	   It can be treated as frame buffer */
	private BufferedImage screenImage = null;
	private Graphics2D screenGraphics = null; // Graphics2D object from screenImage

	/* width, height, bpp and depth information for frame buffer */
	private short fbWidth;
	private short fbHeight;
	private int bpp;
	private int depth;

	/* ColorModel object to draw image into frame buffer */
	private ColorModel rawColorModel; // for normal images
	private ColorModel alphaColorModel; // for transparent images
		
	/* for drawing tile image into frame buffer */
	private BufferedImage tile = null;
	private Rectangle tileRect = null;

	/* resizing information */
	private boolean clientResize;
	private AffineTransform resized;
	private AffineTransform uniform;
	private BufferedImage resizedImage = null;
	private Graphics2D resizedGraphics = null;
	
	/**
	 * ThincClientCanvas
	 * 
	 * creating ThincClientCanvas object. it disables focus traversal and
	 * simply put the pointer of ThincClient object into client variable.
	 * @param tc ThincClient class which constructs this object
	 */
	public ThincClientCanvas(ThincClient tc)
	{
		setFocusTraversalKeysEnabled(false);
        client = tc;
        clientResize = tc.getClientResize();
        uniform = new AffineTransform();
        uniform.scale(1.0, 1.0);
	}
	/**
	 * setMsgHandler()
	 * set the ThincMsgHandler into handler variable.
	 * @param h ThincMsgHandler object to set
	 */
	public void setMsgHandler(ThincMsgHandler h){handler = h;}

	/**
	 * setFrame() sets the ThincFrame into frame variable.
	 * @param f  ThincFrame object to set
	 */
	public void setFrame(ThincFrame f) {frame = f;}
		
	/**
	 * setListener()
	 *
	 * set the listener to listen any changes or interactions from
	 * user.
	 */
	public void setListener()
	{
		addMouseListener(new ThincCMsgEventMouseButton(handler,this));
        addMouseMotionListener(new ThincCMsgEventMouseMove(handler,this));
        addKeyListener(new ThincCMsgEventKey(handler,this));
        addHierarchyBoundsListener(new ThincCMsgResizeEvent(handler,this,frame));
        requestFocus(); // request focus on this ThincClient Canvas object!
    }
	
	/**
	 * createFrameBuffer()
	 * 
	 * creates frame buffer from the network message from the server.
	 * It also creates ColorModel objects to draw frame buffer later.
	 * 
	 * @param w width for frame buffer
	 * @param h height for frame buffer
	 * @param bpp bit per pixel for frame buffer
	 * @param depth depth for each pixel
	 */
	public void createFrameBuffer(short w, short h, int bpp, int depth)
	{
		fbWidth = w;
		fbHeight = h;
		this.bpp = bpp;
		this.depth = depth;
		createColorModel(bpp, depth);
		
		screenImage = getGraphicsConfiguration().createCompatibleImage(fbWidth,fbHeight);
			// create the frame buffer 
		screenGraphics = screenImage.createGraphics();
			// get the Graphics2D object and save it

		/* set the size of ThincFrame object and itself */
		
//		frame.setSize(fbWidth, fbHeight);
//		setSize(fbWidth,fbHeight);
		
		frame.setResizable(false);
		frame.setVisible(true);

//		while(!getSize().equals(new Dimension(fbWidth,fbHeight))) 
//		while(!frame.getSize().equals(new Dimension(fbWidth,fbHeight))) 
		setSize(fbWidth, fbHeight);
		this.setPreferredSize(new Dimension(fbWidth, fbHeight));
//		frame.setSize(fbWidth, fbHeight);
//		frame.setBounds(0,0, fbWidth, fbHeight);
		frame.pack();
		frame.setMaximumSize(new Dimension(frame.getWidth(),frame.getHeight()));
		frame.setMaximizedBounds(new Rectangle(frame.getWidth(),frame.getHeight()));
		
		/* preparing to receive frame buffer data from the server */
		rawPixelIntArray = new int[w*h];
		rawPixelIntBuffer= IntBuffer.wrap(rawPixelIntArray);
	}
	
	/**
	 * createColorModel()
	 * 
	 * creates the ColorModel objects to draw frame buffer later
	 * @param bpp Bit per pixel of frame buffer
	 * @param depth Depth for each pixel
	 */
	private void createColorModel(int bpp, int depth)
	{
		int rgb = 1;
		int alpha = 1;
		for(int i=0; i<depth/3-1; i++) // fill the rgb masks
		{
			rgb <<= 1;
			rgb = (rgb | 1);  
		}
		for(int i=0; i<bpp-depth-1;i++) // fill the alpha masks
		{
			alpha <<=1;
			alpha |= 1;
		}
		rawColorModel = new DirectColorModel(depth, rgb ,rgb << (depth/3), rgb <<(depth*2/3));		
		alphaColorModel = new DirectColorModel(bpp, rgb << ((depth/3)+(bpp-depth)),rgb << (bpp-depth), rgb, alpha << depth);
	}
	
	
	/**
	 * createCursor()
	 * 
	 * creates a cursor image if it comes as a RGB image
	 * 
	 * @param imageArray target int array for cursor
	 * @param bb ByteBuffer object which contains RGB image for cursor
	 */
	public void createCursor(int[] imageArray, ByteBuffer bb)
	{
		//RGBA -> ABGR
		byte alpha;
		isARGB = true;
		for(int i=0; i<imageArray.length; i++)
		{
			imageArray[i] = bb.getInt();
			alpha = (byte) (imageArray[i] & 0xff);
			imageArray[i] >>= 8;
			imageArray[i] |= (alpha << 24);
			imageArray[i] = argbToAbgr(imageArray[i]);
		}
	}
	
	/**
	 * createCursor()
	 * 
	 * creates the cursor image as an int array if it comes as a bitmap image
	 * @param imageArray target int array for cursor
	 * @param fgb byte array which contains bitmap image for foreground
	 * @param bgb byte array which contains bitmap image for background
	 * @param fg color information for foreground
	 * @param bg color information for background
	 * @param width width for cursor
	 */
	public void createCursor(int[] imageArray, byte[] fgb, byte[] bgb, int fg, int bg,int width) {
		isARGB = false;
		bgbCursor = bgb;
		fgbCursor = fgb;
		fgCursor = fg;
		bgCursor = bg;
		createBitMapImage(imageArray,fgb,bgb,fg,bg,width); 
			// invokes createBitMapImage to create an image from bitmap
	}
	/**
	 * createCursor()
	 * 
	 * creates the cursor image as an int array 
	 * if it comes as a bitmap image.
	 * this method is used when modifying the cursor into 
	 * different images.
	 * @param imageArray target int array for cursor
	 * @param fgb byte array which contains bitmap image for foreground
	 * @param bgb byte array which contains bitmap image for background
	 */
	public void createCursor(int[] imageArray, byte[] fgb, byte[] bgb) {
		isARGB = false;
		createBitMapImage(imageArray,fgb,bgb,fgCursor,bgCursor,widthCursor);
		// invokes createBitMapImage to create an image from bitmap
	}
	
	/**
	 * colorChangeCursor()
	 * 
	 * modifies the cursor image color.
	 * @param fg color information for foreground
	 * @param bg color information for background
	 */
	public void colorChangeCursor(int fg, int bg)
	{
		if(isARGB) return; // if ARGB, do nothing
		for(int i=0;i < cursorIntArray.length; i++) {
			if(cursorIntArray[i] == bgCursor) cursorIntArray[i] = bg;
			if(cursorIntArray[i] == fgCursor) cursorIntArray[i] = fg;
		}
		fgCursor = fg;
		bgCursor = bg;
		/* set modified cursor onto canvas */
		setThincCursor(cursorIntArray,cursorHot.x,cursorHot.y,widthCursor,heightCursor);
	}
	
	/**
	 * createBitmapImage()
	 * 
	 * creates the image as an int array from bitmap image information.
	 * 
	 * @param imageArray target int array for cursor
	 * @param fgb byte array which contains bitmap image for foreground
	 * @param bgb byte array which contains bitmap image for background
	 * @param fg color information for foreground
	 * @param bg color information for background
	 * @param width width for cursor
	 */
	public void createBitMapImage(int[] imageArray, byte[] fgb, byte[] bgb, int fg, int bg,int width)
	{
		byte mask;
		for(int i=0,j=0,widthCount = 0; i< fgb.length; i++) 
		{
			mask = 1;
			/* fill each int in imageArray from each bit in fgb and bgb.
			 * because java does not support bit assigned type, used
			 * mask and & operator to extract each bit from byte type.
			 */
			while(mask != 0){ // for each bit in byte bgb[i] and fgb[i],
				/* fill the background if available */
				if(bgb != null)
					imageArray[j] =((bgb[i] & mask) == mask)?bg:0;
				else
					imageArray[j] = 0;

				/* fill the foreground */
				imageArray[j] =((fgb[i] & mask) == mask)?fg:
									((bgb == null)?bg:imageArray[j]); 

				mask <<= 1; // bit shift to refer the next bit in byte bgb[i] and fgb[i]
				j++;
				widthCount++;
				if(widthCount == width) { 
					// if over the width, go another for loop outside
					// from while loop.
					widthCount = 0;
					break;
				}
			}
		}
		
	}

	/**
	 * setThincCursor()
	 * 
	 * sets the cursor to the canvas which is generated from createThincCursor().
	 * used when cursor is changed after initialization.
	 * @param imageArray image array generated from createThincCursor()
	 * @param hotx hot x coordinate in cursor(ex> 0 for normal arrow cursor)
	 * @param hoty hot y coordinate in cursor(ex> 0 for normal arrow cursor)
	 */
	public void setThincCursor(int[] imageArray, int hotx, int hoty){
		setThincCursor(imageArray,hotx,hoty,widthCursor,heightCursor);
	}
	/**
	 * setThincCursor()
	 * 
	 * sets the cursor to the canvas which is generated from createThincCursor().
	 * used when cursor is changed after initialization.
	 * 
	 * @param imageArray image array generated from createThincCursor()
	 */
	public void setThincCursor(int[] imageArray) {
		setThincCursor(imageArray,cursorHot.x,cursorHot.y,widthCursor,heightCursor);
	}
	
	/**
	 * setThincCursor()
	 * 
	 * sets the cursor to the canvas which is generated from createThincCursor().
	 * used when it wants to initialize the cursor.
	 * 
	 * @WARNING: because java only supports 32 x 32 images for cursor, if the
	 * cursor is larger then 32 x 32, it should be resized automatically.
	 * 
	 * @param imageArray image array generated from createThincCursor()
	 * @param hotx hot x coordinate in cursor(ex> 0 for normal arrow cursor)
	 * @param hoty hot y coordinate in cursor(ex> 0 for normal arrow cursor)
	 * @param w width for cursor
	 * @param h height for cursor
	 */
	public void setThincCursor(int[] imageArray, int hotx, int hoty,int w, int h){
		widthCursor = w;
		heightCursor = h;
		cursorIntArray = imageArray.clone();
		cursorHot = new Point(hotx, hoty);
		cursor = new MemoryImageSource(w,h,alphaColorModel,cursorIntArray,0,w);
		cursor.setAnimated(false);
		cursor.setFullBufferUpdates(false);
		cursorImage = createImage(cursor);

		// TODO: because JAVA automatically resizes its cursor if cursor image is too big,
		// is there any idea to deal with this?
		// 1: crop the image to proper size --> could be a problem with hot spot
		// 2: let it go --> too small to handle it 
		// 3: make transparent cursor and try to simulate it! --> hot spot problem
		int[] pixels = new int[16 * 16];
		Image invCursor = Toolkit.getDefaultToolkit().createImage(
				new MemoryImageSource(16, 16, pixels, 0, 16));
		Cursor c = getToolkit().createCustomCursor(invCursor, new Point(0,0), CURSOR_NAME);
//		Cursor c = getToolkit().createCustomCursor(cursorImage, cursorHot, CURSOR_NAME);
		setCursor(c);
		repaint(30);
	}
	
	/**
	 * showThincCursor()
	 * 
	 * shows or hides the cursor from the user
	 * 
	 * @param value true for show, false for hide
	 */
	public void showThincCursor(boolean value) {
		if(value == cursorVisible){	
			Image temp = cursorImage;
			cursorImage = invCursor;
			invCursor = temp;
			cursorVisible = !cursorVisible;
			repaint(30);
		}
	}

	/**
	 * moveThincCursor()
	 * 
	 * moves the cursor to the specific location
	 * @param x x coordinate to move
	 * @param y y coordinate to move
	 * @throws ThincException
	 */
	public void moveThincCursor(int x, int y) throws ThincException
	{
		Robot r;
		try {
			r = new Robot();
		} catch (AWTException e) {
			throw new ThincException("moveThincCursor: error on Robot class",e);
		}
		r.mouseMove(x, y);
		if(currentCursor == null) currentCursor = new Point(x,y);
		else {
			currentCursor.x = x;
			currentCursor.y = y;
		}
		repaint(30);
	}
	

	/**
	 * drawThincCursor()
	 * 
	 * Actually draws imitate ThincCursor.
	 * @param x x coordinate to draw
	 * @param y y coordinate to draw
	 */
	public void drawThincCursor(int x, int y)
	{
		currentCursor.x = x;
		currentCursor.y = y;
		repaint(30);
	}
	public void drawThincCursor()
	{
		if(currentCursor != null) repaint(30,currentCursor.x,currentCursor.y,widthCursor,heightCursor);
	}
	
	
	/**
	 * setFrameBuffer()
	 * 
	 * sets the frame buffer from the latest position.
	 * @param bb images to fill in
	 */
	public void setFrameBuffer(int[] bb)
	{
			/*translate normal rgb into abgr.
			 default BufferedImage object contains each pixel as ABGR, not RGB*/
			for(int i=0; i<bb.length; i++) { 
				bb[i] >>= 8;
				bb[i] |= 0xff000000;
			}
			rawPixelIntBuffer.put(bb);
	}
	
	/**
	 * drawFrameBuffer()
	 *
	 * draws frame buffer onto canvas. when it used,
	 * the frame buffer should be filled with setFrameBuffer
	 * methods
	 */
	public void drawFrameBuffer()
	{
		/* creating images from int array */
		rawPixel = new MemoryImageSource(fbWidth,fbHeight,rawColorModel,rawPixelIntArray,0,fbWidth);
		rawPixel.setAnimated(true);
		rawPixel.setFullBufferUpdates(false);
		rawPixelImage = this.createImage(rawPixel);
		
		/* draw the image */
		screenGraphics.drawImage(rawPixelImage,0,0,null);

		/*make the ThincFrame object available again*/
//		frame.setVisible(true);
		frame.setResizable(true); 
		repaint();
	}
	
	/**
	 * drawFillRect()
	 *
	 * draws a rectangle onto canvas. 
	 *  
	 * @param x x coordinate for rectangle
	 * @param y y coordinate for rectangle
	 * @param w width for rectangle
	 * @param h height for rectangle
	 * @param color color for rectangle
	 */
	public void drawFillRect(int x, int y, int w, int h, int color)
	{
		/* simply draw fillrect by using fillRect method in Graphics2D object */
		screenGraphics.setColor(new Color(color));
		screenGraphics.fillRect(x, y, w, h);
		repaint(30);
	}

	/**
	 * drawImage()
	 * 
	 * draws the image represented as IntBuffer object.
	 * 
	 * @param bb IntBuffer object which contains the image
	 * @param x x coordinates to draw
	 * @param y y coordinates to draw
	 * @param w width to draw
	 * @param h height to draw
	 * @param type type to draw - TYPE_RAW, TYPE_PNG, TYPE_BITMAP, and TYPE_TRANSPARENT 
	 * @throws ThincException
	 */
	public void drawImage(IntBuffer bb,int x, int y, int w, int h, int type) throws ThincException
	{
		/* simple type checking */
		if(type < TYPE_RAW || type > TYPE_TRANSPARENT) throw new ThincException("drawFrameBuffer: unproper use");

		/* copy IntBuffer object from the server into rawPixel*/
		rawPixelIntArray = new int[bb.limit()];
		bb.get(rawPixelIntArray);

		/*translate normal rgb into abgr.
		 default BufferedImage object contains each pixel as ABGR, not RGB*/
		if(type == TYPE_RAW) {
			for(int i=0; i<rawPixelIntArray.length; i++) {
				rawPixelIntArray[i] >>= 8;
				rawPixelIntArray[i] |= 0xff000000;
			}
		}
		/* set the ColorModel for each case */
		switch(type) { 
		case TYPE_RAW:
		case TYPE_PNG:
		case TYPE_BITMAP:
			rawPixel = new MemoryImageSource(w,h,rawColorModel,rawPixelIntArray,0,w);
		break;
		case TYPE_TRANSPARENT:
			rawPixel = new MemoryImageSource(w,h,alphaColorModel,rawPixelIntArray,0,w);
		break;
		}
		/* set various configuration for creating images */
		rawPixel.setAnimated(true);
		rawPixel.setFullBufferUpdates(false);

		/* create image using createImage method */
		rawPixelImage = this.createImage(rawPixel);
		
		/* draw image onto canvas */
		if (ThincSMsgFlagResize.isResized() && !clientResize &&
				(type == TYPE_RAW || type == TYPE_PNG)) screenGraphics.setTransform(uniform);
		screenGraphics.drawImage(rawPixelImage,x,y,null);
		if (ThincSMsgFlagResize.isResized() && !clientResize &&
				(type == TYPE_RAW || type == TYPE_PNG)) screenGraphics.setTransform(resized);
		repaint(30);
	}

	/**
	 * drawImageScaled()
	 * 
	 * draws the scaled image onto canvas. it is as same as
	 * drawImage methods except scaling
	 * 
	 * @param bb IntBuffer object which contains the image
	 * @param x x coordinates to draw
	 * @param y y coordinates to draw
	 * @param w original width to draw
	 * @param h original height to draw
	 * @param dst_w resized width to draw
	 * @param dst_h resized height to draw
	 * @param type type to draw - TYPE_CIDEO, TYPE_SCALED_RAW, TYPE_SCALED_BITMAP, and TYPE_SCALED_TRANSPARENT 
	 * @throws ThincException
	 */
	public void drawImageScaled(IntBuffer bb,int x, int y, int w, int h, int dst_w, int dst_h, int type) throws ThincException
	{
		/* simple type checking */
		if(type < TYPE_VIDEO || type > TYPE_SCALED_TRANSPARENT) throw new ThincException("drawFrameBuffer: unproper use");

		/* set the scaling factors onto AffineTransform object */
		AffineTransform trans = new AffineTransform();
		trans.setToScale(((double)dst_w/(double)w), ((double)dst_h/(double)h));

		/* copy IntBuffer object from the server into rawPixel*/
		rawPixelIntArray = new int[bb.limit()];
		bb.get(rawPixelIntArray);

		/*translate normal rgb into abgr.
		 default BufferedImage object contains each pixel as ABGR, not RGB*/
		if(type == TYPE_SCALED_RAW) {
			for(int i=0; i<rawPixelIntArray.length; i++) {
				rawPixelIntArray[i] >>= 8;
				rawPixelIntArray[i] |= 0xff000000;
			}
		}

		/* set the ColorModel for each case */
		switch(type) {
		case TYPE_VIDEO:
		case TYPE_SCALED_TRANSPARENT:
			rawPixel = new MemoryImageSource(w,h,alphaColorModel,rawPixelIntArray,0,w);
			break;
		case TYPE_SCALED_RAW:
		case TYPE_SCALED_BITMAP:
			rawPixel = new MemoryImageSource(w,h,rawColorModel,rawPixelIntArray,0,w);
			break;
		}
		/* set various configuration for creating images */
		rawPixel.setAnimated(true);
		rawPixel.setFullBufferUpdates(false);

		/* create original image using createImage method */
		rawPixelImage = this.createImage(rawPixel);
		
		BufferedImage videoImage = getGraphicsConfiguration().createCompatibleImage(dst_w, dst_h);
		Graphics2D videoGraphics = videoImage.createGraphics();
		videoGraphics.setTransform(trans);
		videoGraphics.drawImage(rawPixelImage,0,0,null);
		videoGraphics.dispose();
		
		/* draw the scaled image onto canvas */
		if (ThincSMsgFlagResize.isResized() && !clientResize) screenGraphics.setTransform(uniform);
		screenGraphics.drawImage(videoImage,x,y,null);
		repaint(30,x,y,dst_w,dst_h);		
		//paintImmediately(x, y, dst_w, dst_h);
		if (ThincSMsgFlagResize.isResized() && !clientResize) screenGraphics.setTransform(resized);
	}
	
	/**
	 * copyRegion()
	 * copies the region from the frame buffer
	 * 
	 * @param sx source x coordinates
	 * @param sy source y coordinates
	 * @param dx destination x coordinates
	 * @param dy destination y coordinates
	 * @param width width for region
	 * @param height height for region
	 */
	public void copyRegion(int sx, int sy, int dx, int dy, int width, int height)
	{
		/* simply using copyArea method provided by java */
		if (ThincSMsgFlagResize.isResized() && !clientResize) screenGraphics.setTransform(uniform);
		screenGraphics.copyArea(sx, sy, width, height, dx - sx, dy -sy);
		if (ThincSMsgFlagResize.isResized() && !clientResize) screenGraphics.setTransform(resized);
		repaint(30);
	}
	
	/**
	 * setTile()
	 * 
	 * sets the tile to be filled later by using drawTile() method
	 * 
	 * @param x x coordinates for tile
	 * @param y y coordinates for tile
	 * @param width width for tile
	 * @param height height for tile
	 * @param pixels int array to create the tile
	 */
	public void setTile(int x, int y, int width, int height, int[] pixels)
	{
		/* NOTE: because 1 pixel cannot be set as tile, it is set by Color */
		if(pixels.length == 1) {
			//because it is a single pixel, we need to convert it into RGB from ABGR,
			int bgr = argbToAbgr(pixels[0]>> 8);
			screenGraphics.setColor(new Color(bgr));
			return;
		}
		/* create new image for tile */
		if (null == tile || tile.getTileWidth() != width
		            || tile.getTileHeight() != height) {
			tile = new BufferedImage(width,height,BufferedImage.TYPE_INT_ARGB);
		} else
			tile.flush();
		/* set the configuration for tile */
		tile.getRaster().setDataElements(0,0,width,height,pixels);
		if(tileRect == null) tileRect = new Rectangle();
		tileRect.setBounds(x, y, width, height);

		/* define the tile using setPaint method in Graphics2D class */
		screenGraphics.setPaint(new TexturePaint(tile, tileRect));
	}
	
	/**
	 * drawTile()
	 * 
	 * fills the region using tile defined by setTile() method.
	 * 
	 * @param x x coordinates to be filled with
	 * @param y y coordinates to be filled with
	 * @param width width to be filled with
	 * @param height height to be filled with
	 */
	public void drawTile(int x, int y, int width, int height)
	{
		/* simply using fillRect method */
		screenGraphics.fillRect(x, y, width, height);
		repaint(30);
	}
	
	/**
	 * createCache()
	 * 
	 * creates the cache. 
	 * @param img number of image cache. total number of slot in cache should be the power of 2 
	 * @param bit number of bitmap cache. total number of slot in cache should be the power of 2 
	 * @param pix number of pixmap cache. total number of slot in cache should be the power of 2 
	 */
	public void createCache(int img, int bit, int pix) {
		cache[IMG_CACHE] = new Hashtable(2<<img);
		cache[BIT_CACHE] = new Hashtable(2<<bit);
		cache[PIX_CACHE] = new Hashtable(2<<pix);

		// make cacheMask
		cacheMask[IMG_CACHE] = 1;
		cacheMask[BIT_CACHE] = 1;
		cacheMask[PIX_CACHE] = 1;
		for(int i=0; i<=img; i++) {
			cacheMask[IMG_CACHE] <<= 1;
			cacheMask[IMG_CACHE] |= 1;
		}
		for(int i=0; i<=bit; i++) {
			cacheMask[BIT_CACHE] <<= 1;
			cacheMask[BIT_CACHE] |= 1;
		}
		for(int i=0; i<=pix; i++) {
			cacheMask[PIX_CACHE] <<= 1;
			cacheMask[PIX_CACHE] |= 1;
		}
	}
	
	
	/**
	 * getCache()
	 * 
	 * gets the object from the cache.
	 * @param id identifier to get an object
	 * @param type cache type - IMG_CACHE, BIT_CACHE, and PIX_CACHE
	 * @return object refered by id and type
	 * @throws Exception
	 */
	public Object getCache(int id, int type) throws ThincException {
		Object target;
		/* simple type checking */
		if(type > BIT_CACHE || type < IMG_CACHE) 
			throw new ThincException("type out of range");
		
//		id &= cacheMask[type];
		
		/* get the cache */
		if(!cache[type].containsKey((Object)id)) throw new ThincException("cache contents does not exist!");
		target = cache[type].get((Object)id);
		if (target == null) throw new ThincException ("cache contents empty");
		
		return target;
	}
	
	/**
	 * setCache()
	 * 
	 * sets the object into cache.
	 * 
	 * @param contents actual object to be contained
	 * @param id id to be refered
	 * @param type cache type - IMG_CACHE, BIT_CACHE, and PIX_CACHE
	 * @throws Exception
	 */
	@SuppressWarnings("unchecked")
	public void setCache(Object contents, int id, int type) throws ThincException{
		if(type > BIT_CACHE || type < IMG_CACHE) 
			throw new ThincException("type out of range:"+type);

//		id &= cacheMask[type];

		/* error checking! */
		if(contents == null) throw new ThincException("setCache: try to put null in cache");		
		if(cache[type].containsKey((Object)id)) cache[type].remove((Object)id);
		cache[type].put((Object)id, contents);
	}
	
	/**
	 * argbToAbgr()
	 * 
	 * converts ARGB image to ABGR images.
	 * @param src source ARGB pixel
	 * @return ABGR image pixel
	 */
	public int argbToAbgr(int src) {
		int r,g,b,a;
		/* extract alpha, red, green, blue separately*/
		a = (src & 0xff000000); 
		r = (src & 0xff0000);
		g = (src & 0xff00);
		b = (src & 0xff);

		/* switch the position for red and blue*/
		r >>= 16;
		b <<= 16;	
		
		/* put the switched alpha, blue, green, red onto src */
		src = a;
		src |= r;
		src |= g;
		src |= b;

		return src;
	}

	public void setResized(int newWidth, int newHeight) {
		resized = new AffineTransform();
		resized.scale((double)newWidth/(double)fbWidth, (double)newHeight/(double)fbHeight);
		screenGraphics.setRenderingHint(RenderingHints.KEY_INTERPOLATION, RenderingHints.VALUE_INTERPOLATION_BILINEAR);
		if(!clientResize) screenGraphics.setTransform(resized);
		else {
			resizedImage = getGraphicsConfiguration().createCompatibleImage(newWidth, newHeight); 
			resizedGraphics = resizedImage.createGraphics();
			resizedGraphics.setTransform(resized); // set graphics as resized
			resizedGraphics.setRenderingHint(RenderingHints.KEY_INTERPOLATION, RenderingHints.VALUE_INTERPOLATION_NEAREST_NEIGHBOR);
		}

	}
	public int getBpp(){return bpp;}
	public short getFBHeight(){return fbHeight;}
	public short getFBWidth(){return fbWidth;}
	public boolean getClientResize() {return clientResize;}
	public int getCursorHeight() {return heightCursor;}
	public int 	getCursorWidth() {return widthCursor;}
	/**
	 * paint()
	 * simply draws frame buffer onto canvas.
	 */
	@Override
	public void paint(Graphics g) {
		/* if canvas does resize by itself, and if it is resized, do it by canvas */
		if(clientResize && ThincSMsgFlagResize.isResized()) {
			resizedGraphics.drawImage(screenImage,0,0,this); // draw itself as resized
			g.drawImage(resizedImage, 0, 0, this);
		} else 
			g.drawImage(screenImage, 0, 0, this);
		if(cursorImage != null && currentCursor != null) g.drawImage(cursorImage,currentCursor.x-cursorHot.x,currentCursor.y-cursorHot.y,this);
	}

	/**
	 * update()
	 * simply invokes paint method
	 */
	@Override
	public void update(Graphics g) {
		paint(g);    
	}

	/**
	* Always returns true because it should always be opaque.
	*
	* @return true
	*/
	@Override
	public boolean isOpaque() {
		return true;
	}

	/**
	* Always returns true in order to properly handle key events.
	*
	* @return true
	*/
	@Override
	public boolean isFocusable() {
		return true;
	}

}
