/* Copyright 2006, 2007 Columbia University */
/* This file is part of THINC.

    THINC is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 2 of the License, or
    (at your option) any later version.

    THINC is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with THINC.  If not, see <http://www.gnu.org/licenses/>.
*/
import java.io.IOException;
import java.nio.ByteBuffer;
import java.nio.ByteOrder;
/**
 * handles the general init reply/request messages from/to the server.  it recoginizes the reply message's type then pass them into  the proper message object. Moreover, it also takes the control from client request messages to send a message. 
 * @author  Young Jin Yoon
 * @see ThincHandler
 * @see  ThincMsgHandler
 */
public final class ThincMsgInitReplyHandler extends ThincHandler implements ThincInitFlag {
	public static final int REPLY_SIZE = ThincProto2.T_SERV_REPLY_SIZE;
	public static final byte[] DEFAULT_HANDSHAKE_ORDER 
					= { FBINFO, CURSOR, FBDATA, ENCODER,
						CACHE_SIZE, VIDEO, VIDEO_SERV_FMTS, VIDEO_CLIENT_FMTS};//,
					  //KEEPALIVE, APP_GET, APP_LIST, APP_SHARING};
	private byte hsOrder[]; // handshake order
	private int currentHandShake; // handshake pointer which is sent recently.

	/* server message and client message array for init request and reply */
	private ThincSMsgInitReply replyMessage[] = new ThincSMsgInitReply[MAX_INIT_ARRAY];
	private ThincCMsgInitRequest requestMessage[] = new ThincCMsgInitRequest[MAX_INIT_ARRAY];

	/* pointer for ThincClientCanvas and parent*/
//	private ThincClientCanvas canvas;
	private ThincMsgHandler handler;
	
	private ByteBuffer hd; // ByteBuffer to receive header
	
	/**
	 * creates ThincMsgInitReplyHandler object. it initializes all
	 * parameters in object and init handshake messages for both
	 * client message and server message
	 * @param tc
	 * @param c
	 * @param h
	 */
	public ThincMsgInitReplyHandler(ThincClient tc, ThincClientCanvas c,ThincMsgHandler h) {
		super(tc,c);
		/* set the pointer */
//		canvas = c;
		handler = h;
		sc = h.getSocketChannel(); // set the socket from ThincMsgHandler object

		hsOrder = DEFAULT_HANDSHAKE_ORDER; // set the handshake order
		currentHandShake = -1; // currentHandShake is not started yet (set to -1)
		initMessage(tc,c); // initialize server and client init message classes

		/* allocate ByteBuffer to receive header */
		hd = ByteBuffer.allocateDirect(REPLY_SIZE);
		hd.order(ByteOrder.BIG_ENDIAN);
	}
	
	/**
	 * initServerMessage()
	 *
	 * registers server and client init reply/request message into array named replyMessage and
	 * requestMessage. when reply from server comes, the handler automatically choose the proper 
	 * message and pass the control to it. if you want to add another server/client messages, 
	 * you can register the message by adding the line of here.
	 * Moreover, you can also activate sending client message by modifying 
	 * <b>DEFAULT_HANDSHAKE_ORDER</b> variable of this class.
	 */
	private void initMessage(ThincClient tc, ThincClientCanvas c) {

		/* register proper replyMessage element with proper classes */
		replyMessage[OK] =					new ThincSMsgInitReplyOK(tc,handler,c);
		replyMessage[NOTOK] =				new ThincSMsgInitReplyNotOK(tc,handler,c);
		replyMessage[UNKNOWN] =				new ThincSMsgInitReplyUnknown(tc,handler,c);
		replyMessage[FBINFO] =				new ThincSMsgInitReplyFBInfo(tc,handler,c);
		replyMessage[CACHE_SIZE] = 			new ThincSMsgInitReplyCacheSize(tc,handler,c);
		replyMessage[CURSOR] = 				new ThincSMsgInitReplyCursor(tc,handler,c);
		replyMessage[ENCODER] = 			new ThincSMsgInitReplyEncoder(tc,handler,c);
		replyMessage[FBDATA] = 				new ThincSMsgInitReplyFBData(tc,handler,c);
		replyMessage[VIDEO_SERV_FMTS] = 	new ThincSMsgInitReplyVideoServerFormat(tc,handler,c);

		/* deprecated list */
		//replyMessage[APP_GET] = 			new ThincSMsgInitReplyAppGet(tc,handler,c);
		//replyMessage[APP_LIST] = 			new ThincSMsgInitReplyAppList(tc,handler,c);
		//replyMessage[APP_SHARING] = 		new ThincSMsgInitReplyAppSharing(tc,handler,c);
		//replyMessage[KEEPALIVE] = 			new ThincSMsgInitReplyKeepAlive(tc,handler,c);
	
		
		/* register proper requestMessage element with proper classes */
		requestMessage[FBINFO] =				new ThincCMsgInitRequestFBInfo(handler,c);
		requestMessage[CACHE_SIZE] = 			new ThincCMsgInitRequestCacheSize(handler,c);
		requestMessage[CURSOR] = 				new ThincCMsgInitRequestCursor(handler,c);
		requestMessage[ENCODER] = 				new ThincCMsgInitRequestEncoder(handler,c);
		requestMessage[FBDATA] = 				new ThincCMsgInitRequestFBData(handler,c);
		requestMessage[NOVIDEO] = 				new ThincCMsgInitRequestNoVideo(handler,c,this);
		requestMessage[VIDEO] = 				new ThincCMsgInitRequestVideo(handler,c);
		requestMessage[VIDEO_CLIENT_FMTS] = 	new ThincCMsgInitRequestVideoClientFormat(handler,c);
		requestMessage[VIDEO_SERV_FMTS] = 		new ThincCMsgInitRequestVideoServerFormat(handler,c);
		
		/* deprecated list */
//		requestMessage[APP_GET] = 				new ThincCMsgInitRequestAppGet(handler,c);
//		requestMessage[APP_LIST] = 			new ThincCMsgInitRequestAppList(handler,c);
//		requestMessage[APP_SHARING] = 			new ThincCMsgInitRequestAppSharing(handler,c);
//		requestMessage[KEEPALIVE] = 			new ThincCMsgInitRequestKeepAlive(handler,c);

	}
	
	/**
	 * start()
	 * 
	 * actually starts to receive the init reply messages from server. if it comes,
	 * this parses init reply header and pass the control to the proper
	 * server init reply message class.
	 * @throws ThincException
	 */
	public void start() throws ThincException {
		short reply;
		/* receive init reply header*/
		try {
			recvBytes(hd);
		} catch (ThincException e) {
			throw new ThincException("receiver: Error on recvBytes",e);
		}
		hd.flip(); // flips the ByteBuffer to get the content of it
		/* get the reply number*/
		reply = hd.getShort();
		hd.getShort();
		hd.clear();
		/* pass the control to the proper message*/
		if(replyMessage != null)
			replyMessage[reply].process();

		/* try to sent next init request to the server by using sendNextMessage() method*/
		try{
			sendNextMessage();
		} catch (ThincException e){
			throw new ThincException("receiver: error on sendNextMessage",e);
		}
	}
	
	/**
	 * sendNextMessage()
	 * 
	 * send next init request message to the server.
	 * @throws ThincException
	 */
	public void sendNextMessage() throws ThincException
	{
		do { // try to find next message and send it
			/* if all request is completed send ThincCMsgInitClientDone to say initialization is done */
			if(currentHandShake == hsOrder.length-1) { 
				ThincCMsgInitClientDone done = new ThincCMsgInitClientDone(handler,canvas);
				done.process();
				canvas.setListener();
				return;
			}
			/* if it is not done, try to increase the pointer to point next send message */
			else currentHandShake++;
		}while(hsOrder[currentHandShake] == 0); 
		/* send next message refered by currentHandShake */
		try {
			requestMessage[hsOrder[currentHandShake]].process();
			lastSent = hsOrder[currentHandShake];
		} catch (ThincException e) {
			throw new ThincException("sendNextMessage: error on initialization",e);
		}
	}	
}
