/**
 * Thinc Protocol specification for version 0.2, JAVA
 * @author Ricardo Baratto
 */
/* 
 * NOTES
 * This is the THINC protocol spec, version 0.2. The goal with this version
 * is to make the connection stage of the protocol not suck so much. As such,
 * all the changes here refer to the handshake stage. Messages used after
 * a client is fully connected do not change from 0.1.
 *
 * The basic idea with the new handshake is to make it possible at connection
 * time for the client and server to agree on a set of parameters that will
 * be used for the session. Indeed, 0.1 cannot even do that!
 *
 * A lot of this is modeled on the ssh protocol. Maybe one day we will just
 * scrap a lot of this and use something like SIP. But for now, we're in the
 * nih world. And we're still a hack.
 */
/* HANDSHAKE
 * 1 Server: version string (no header)
 * - Client: version string (no header)
 * 2 Server: security caps
 * - Client: security caps (which are a subset of the server caps)
 * - Server: session sec caps (or notok) (basically an ack from server)
 * 3 if(security caps.ssl) ssl_handshake (started by client)
 * 4 if(security caps.auth) auth (started by client, server replies ok/notok.
 *   TODO: Add multiple auth attempts)
 * 5 client starts sending requests to server to negotiate
 *   aspects of the connection. If the server does not understand a request,
 *   it sends back a reply with "unknown message". It should try not to close 
 *   the connection but i guess it's ok to send back a NOTOK if the request
 *   is just too insulting for the server.
 *   a server reply may be:
 *   => info client requested
 *   => do not support, can't send info, etc.
 *   => unknown request.
 *   if a request triggers multiple message exchanges, each part of the
 *   conversation should be identified by a different request. we need this
 *   to simplify the implementation of the server.
 *   Requests have their own message id space, and all are wrapped by a
 *   generic "request' structure.
 *   e.g. 
 *   client: request fb info
 *   server: send fb info
 *   client: request cursor
 *   server: no cursor.
 * 6 client: done. no more requests, we're all done, let's start doing
 *   real work.
 *
 * at this point, handshake is finished and normal operation begins.
 *
 * NOTES
 * At any point, a reject (not ok) may be sent to terminate the connection.
 */
/* SPECS
 * - Version String: "THINC_[major].[minor] [additional info]\0"
 * => [major] and [minor] are ascii numbers
 * => [additional info]: optional, maybe useful for debugging purposes
 * => must be NULL terminated
 * - Caps: 32bit integer with bits set marking specific caps.
 */

public interface ThincProto2 {

	/**********************
	 * common definitions *
	 **********************/

	public static final String THINC_HEADER_PROTOCOL_VERSION = "0.2";

	/*****************************
	 * server -> client messages *
	 *****************************/

	/*** message types ***/
		public static final int T_SERV_OK = 1;		/* server positive ack */
		public static final int T_SERV_NOTOK= 2;		/* server negative and fatal ack */
		public static final int T_SERV_SEC_CAPS = 3;	/* server security caps */
		public static final int T_SESS_SEC_CAPS = 4;	/* the session security caps */
		public static final int T_SERV_REPLY = 5;		/* server reply to a client request */
		public static final int T_SERV_LAST = 6;		/* just a marker */
		

	/* list of security capabilities */
		public static final int T_CAPS_SEC_ENC =		0x0001;	/* encryption */
		public static final int T_CAPS_SEC_AUTH =	 	0x0002;	/* authentication */

	/*** message definitions ***/

	/* T_SERV_SEC_CAPS
	 * Description: The security capabilities of the server. See above for
	 * available capabilities
	 * Size: 4
	 * Data: None
	 */
	public class thinc_serverSecCaps {
		int caps;//:32;	/* 1 bit for each capability */
	} 

	public static final int T_SERV_SEC_CAPS_SIZE = 4;

	/* T_SESS_SEC_CAPS
	 * Description: The security capabilities for the session. 
	 * Size: 4
	 * Data: None
	 */
	public class thinc_sessionSecCaps {
		int caps;//:32;	/* 1 bit for each capability */
	} 

	public static final int T_SESS_SEC_CAPS_SIZE =	4;

	/* T_SERV_OK
	 * Description: A good reply from the server.
	 * Size: 0
	 * Data: None
	 */
	/*** no need for struct, just the type in the header is enough ***/

	public static final int T_SERV_OK_SIZE =	0;

	/* T_SERV_NOTOK
	 * Description: A negative ack from the server. After this is sent, the
	 * connection should be closed.
	 * Size: 4
	 * Data: (Optional) reason for NOTOK, plaintext string
	 */
	public class thinc_serverNotOk {
		int reason_len;//:32;	/* if 0, no reason */
		/* possibly followed by string describing reason for reject */
	}

	public static final int T_SERV_NOTOK_SIZE =	4;

	/*****************************
	 * client -> server messages *
	 *****************************/

	/*** message types ***/
		public static final int T_CLIENT_OK = 1;		/* client ack */
		public static final int T_CLIENT_NOTOK = 2;		/* client negative ack */
		public static final int T_CLIENT_SEC_CAPS = 3;	/* client security caps */
		public static final int T_CLIENT_AUTH = 4;		/* client auth info */
		public static final int T_CLIENT_REQ = 5;		/* client handshake request */
		public static final int T_CLIENT_DONE = 6;		/* client is done with handshake */
		public static final int T_CLIENT_LAST = 7;		/* just a marker */


	/* Hacks so we can live with v1 */
//	#undef T_CAUTH
	public static final int T_CAUTH	= T_CLIENT_AUTH;

	/*** message definitions ***/

	/* T_CLIENT_SEC_CAPS
	 * Description: The security capabilities of the client. See above for 
	 * available capabilities
	 * Size: 4
	 * Data: None
	 */
	public class thinc_clientSecCaps {
		int caps;//:32;	/* 1 bit for each capability */
	}

	public static final int T_CLIENT_SEC_CAPS_SIZE = 4;

	/* T_CLIENT_NOTOK
	 * Description: A negative reply from the client. After this is sent, the
	 * connection should be closed.
	 * Size: 4
	 * Data: (Optional) reason for NOTOK, plaintext string
	 */
	public class thinc_clientNotOk {
		int reason_len;//:32;	/* if 0, no reason */
		/* possibly followed by string describing reason for reject */
	}

	public static final int T_CLIENT_NOTOK_SIZE	= 4;

	/* T_CLIENT_OK
	 * Description: Positive reply from client.
	 * Size: 0
	 * Data: None
	 */
	/*** no need for struct, just the type in the header is enough ***/

	public static final int T_CLIENT_OK_SIZE = 0;

	/* T_CLIENT_DONE
	 * Description: Tells server that the handshake is done
	 * Size: 0
	 * Data: None
	 */
	/*** no need for struct, just the type in the header is enough */

	public static final int T_CLIENT_DONE_SIZE = 0;

	/********************************
	 * session features negotiation *
	 ********************************/

	/* T_CLIENT_REQ
	 * Description: A handshake request from the client. Identifies and precedes 
	 * the specific request being made. You should think of it not as the client
	 * asking for something from the server, but as the beginning of a 
	 * negotiation with the server, as requests can carry client information too,
	 * or they may just be used by the client to inform the server of "something"
	 * (in which case the server will just reply "OK")
	 * Size: 4
	 * Data: None
	 */
	public class thinc_clientReq extends thinc_reqGen{
		short request;	//:16;
		short unused;	//:16;
	} 

	public static final int T_CLIENT_REQ_SIZE = 4;

	/*NOTE: This is not provided in java.*/
/*	#define t_req_head thinc_clientReq	req
	#define t_req_size(req)	(T_CLIENT_REQ_SIZE + req##_SIZE)
	#define t_req_type(r)	(r)->req.request
	#define t_req_set_type(r, type) do {	\
		t_req_type(r) = htons(type);	\
		(r)->req.unused = 0;		\
	} while(0)
	#define t_req_read_type(r)	ntohs(t_req_type(r))
*/
	/* T_SERV_REPLY
	 * Description: A reply to a client request. The id may be either one of
	 * the thinc_serverReplies below or the id of the client's request to 
	 * which we're replying. It may wrap around an additional reply message
	 * containing data needed for the reply.
	 * IMPORTANT: Use the header.length field so that client can read the whole
	 * reply easily without having to figure out which reply you sent first
	 * Size: 4
	 * Data: None
	 */
	public class thinc_serverReply extends thinc_replyGen{
		short reply;//:16;		/* id of the reply */
		short unused;//:16;
	}

	public static final int T_SERV_REPLY_SIZE =	4;

	/*NOTE: This is not provided in java.*/
/*	#define t_reply_head	thinc_serverReply reply
	#define t_reply_size(reply)	(T_SERV_REPLY_SIZE + reply##_SIZE)
	#define t_reply_type(r)	(r)->reply.reply
	#define t_reply_set_type(r, type) do {	\
		t_reply_type(r) = htons(type);	\
		(r)->reply.unused = 0;		\
	} while(0)
	public static final int t_reply_read_type(r)	ntohs((r)->reply.reply)*/

	/* basic replies */
		public static final int T_REPLY_OK = 0;			/* positive reply */
		public static final int T_REPLY_NOTOK = 1;			/* negative (non-fatal) reply */
		public static final int T_REPLY_UNKNOWN = 2;		/* unknown request */
		public static final int T_REPLY_LAST =3;			/* marker for requests below */

	/* list of requests */
	/* TODO: 
	 * - Resize: so client can tell server it's current size (probably before
	 * it sends the initial fb data)
	 * - Clip: for scrolling. useful on pda client.
	 * - Appsharing: more complicated. server needs to tell client which windows
	 *   can be shared, and client must tell server which ones it wants.
	 */

		public static final int T_REQ_FBINFO = T_REPLY_LAST;					/* basic framebuffer info */
		public static final int T_REQ_CURSOR = T_REPLY_LAST+1;					/* cursor information and data */
		public static final int T_REQ_FBDATA = T_REPLY_LAST+2;					/* contents of framebuffer */
		public static final int T_REQ_ENCODER = T_REPLY_LAST+3;					/* how is image data encoded */
		public static final int T_REQ_CACHESZ = T_REPLY_LAST+4;					/* size of various caches */
		public static final int T_REQ_VIDEO = T_REPLY_LAST+5;					/* server supports video? */
		public static final int T_REQ_NOVIDEO = T_REPLY_LAST+6;					/* client telling server that it 
						   														can't support video (see below) */
		public static final int T_REQ_VIDEO_SERV_FMTS = T_REPLY_LAST+7;			/* video formats supported by server */
		public static final int T_REQ_VIDEO_CLIENT_FMTS = T_REPLY_LAST+8;		/* client informs server of its video
						   														formats */
		public static final int T_REQ_KEEPALIVE = T_REPLY_LAST+9;				/* client supports or not keepalives
						   														does server support them? */
		public static final int T_REQ_APP_SHARING = T_REPLY_LAST+10;			/* does the server support application
						   														sharing? */
		public static final int T_REQ_APP_LIST = T_REPLY_LAST+11;				/* list of available applications */
		public static final int T_REQ_APP_GET = T_REPLY_LAST+12;				/* request a particular application.
						   														client may send this message
						   														multiple times */
		public static final int T_REQ_LAST = T_REPLY_LAST+13;
	/* request and reply messages. 
	 * for requests, first member _must_ be thinc_clientReq. for replies
	 * first member _must_ be thinc_serverReply.
	 * use t_request_head and t_reply_head for this
	 * BIG ASSUMPTION: compiler won't add pad after this struct since it's
	 * aligned to 4-byte boundary.
	 *
	 * The *_SIZE #defines do not include the size of the reply/request header.
	 * use t_reply_size() and t_req_size() to compute the total size if 
	 * needed
	 */

	/* simple replies */
	public class thinc_replyOK extends thinc_reqGen{
		thinc_serverReply reply;
	} 

	public static final int T_REPLY_OK_SIZE	= 0;

	public class thinc_replyNotOK extends thinc_reqGen{
		thinc_serverReply reply;
	}

	public static final int T_REPLY_NOTOK_SIZE = 0;

	public class thinc_replyUnknown extends thinc_reqGen{
		thinc_serverReply reply;
	}

	public static final int T_REPLY_UNKNOWN_SIZE = 0;

	/**
	 * FBINFO *****************
	 */
	public class thinc_reqFBInfo extends thinc_reqGen {
		thinc_clientReq	req;
	} 

	public static final int T_REQ_FBINFO_SIZE = 0; 

	public class thinc_replyFBInfo extends thinc_reqGen{
		thinc_serverReply reply;
		boolean byteOrder;					//:1;		/* T_BIG | T_LITTLE (thincProto.h) */
		boolean mbz[] = new boolean[15];	//:15;		/* reserved for future use */
		byte depth; 						//:8;
		byte bpp;							//:8;		/* bits per pixel */
		short width;						//:16;
		short height;						//:16;
	} 

	public static final int T_REPLY_FBINFO_SIZE	= 8;

	/**
	 * CURSOR *****************
	 */
	public class thinc_reqCursor extends thinc_reqGen{
		thinc_clientReq	req;
	} 

	public static final int T_REQ_CURSOR_SIZE = 0;

	public class thinc_replyCursor extends thinc_reqGen{
		thinc_serverReply reply;
		short flags;	//:16;	/* for now just: is it argb or not */
		short size;		//:16;	/* size of cursor data (bytes) */
		byte width;		//:8;	/* dimensions */
		byte height;	//:8;
		byte xhot;		//:8;	/* hot spot */
		byte yhot;		//:8;
		short x;		//:16;	/* position */
		short y	;		//:16;
		int fg;			//:32;	/* colors. ignore if ARGB */
		int bg;			//:32;
		/* followed by size bytes representing the cursor */
	} 

	public static final int T_REPLY_CURSOR_SIZE	= 20;

	/* flags */
	/* reuse from cursor change message
	public static final int T_CURSOR_ARGB	0x01
	*/

	/**
	 * ENCODER ****************
	 */
	public class thinc_reqEncoder extends thinc_reqGen{
		thinc_clientReq	req;
	} 

	public static final int T_REQ_ENCODER_SIZE = 0;

	public class thinc_replyEncoder extends thinc_reqGen{
		thinc_serverReply reply;
		boolean which[] = new boolean[2];		//:2;	/* zlib, png, none. see below */
		boolean reserved[] = new boolean[30];	//:30;	/* reserved for future use */
	} 

	public static final int T_REPLY_ENCODER_SIZE = 4;

	public static final int T_ENCODER_NONE	=	0x00;
	public static final int T_ENCODER_PNG	=	0x01;
	public static final int T_ENCODER_ZLIB	=	0x02;

	/**
	 * FBDATA ******************
	 */
	public class thinc_reqFBData extends thinc_reqGen{
		thinc_clientReq	req;
	} 

	public static final int T_REQ_FBDATA_SIZE =	0;

	public class thinc_replyFBData extends thinc_reqGen{
		thinc_serverReply reply;
		int flags;	//:32;
		/* if compressed, followed by fbZData. if not, followed by
		 * pixel data */
	} 

	public static final int T_REPLY_FBDATA_SIZE	= 4;

	/* (already in thincProto.h)
	public static final int T_FB_COMPRESSED		0x01
	 */

	/**
	 * CACHESIZE **************
	 */
	/* TODO: Should the client tell the server which caches it supports? */
	public class thinc_reqCacheSz extends thinc_reqGen{
		thinc_clientReq	req;
	} 

	public static final int T_REQ_CACHESZ_SIZE = 0;

	public class thinc_replyCacheSz extends thinc_reqGen{
		thinc_serverReply reply;
		/* all cache sizes are in bits */
		byte		img;//:8;
		byte		bit;//:8;
		byte		pix;//:8;
		byte		reserved;//:8; 
	} 

	public static final int T_REPLY_CACHESZ_SIZE = 4;

	/**
	 * VIDEO ******************
	 */
	/* server replies ok or notok */
	public class thinc_reqVideo extends thinc_reqGen{
		thinc_clientReq	req;
	} 

	public static final int T_REQ_VIDEO_SIZE = 0;

	/* server doesn't reply (unless he or she feels insulted) */
	public class thinc_reqNoVideo extends thinc_reqGen{
		thinc_clientReq	req;
	} 

	public static final int T_REQ_NOVIDEO_SIZE = 0;

	public class thinc_reqVideoServFmts extends thinc_reqGen{
		thinc_clientReq	req;
	} 

	public static final int T_REQ_VIDEO_SERV_FMTS_SIZE = 0;

	public class thinc_replyVideoServFmts extends thinc_reqGen{
		thinc_serverReply reply;
		ThincProtoVideo.thinc_vidSInit	fmts;	/* from thincProtoVideo.h */
	} 

	public static final int T_REPLY_VIDEO_SERV_FMTS_SIZE = ThincProtoVideo.T_VIDEO_SINIT_SIZE;

	/* server replies ok or notok */
	public class thinc_reqVideoClientFmts extends thinc_reqGen{
		thinc_clientReq	req;
		ThincProtoVideo.thinc_vidCInit	fmts;	/* from thincProtoVideo.h */
	} 

	public static final int T_REQ_VIDEO_CLIENT_FMTS_SIZE = ThincProtoVideo.T_VIDEO_CINIT_SIZE;

	/**
	 * APPLICATION SHARING ***
	 */
	/* T_REQ_APP_SHARING
	 * server replies ok or notok. If server replies ok (see thincProtoApp.h) each
	 * thinc_header sent will be followed by a thinc_appHeader */
	public class thinc_reqAppSharing extends thinc_reqGen{
		thinc_clientReq	req;
	}

	public static final int T_REQ_APP_SHARING_SIZE = 0;

	/* T_REQ_APPLIST
	 * client asks the server which applications can be shared. The idea is to
	 * let the user choose which ones it wants.
	 */
	public class thinc_reqAppList extends thinc_reqGen{
		thinc_clientReq	req;
	} 

	public static final int T_REQ_APP_LIST_SIZE = 0;

	/* how many apps we're sending, followed by count thinc_appDesc structs */
	public class thinc_replyAppList extends thinc_reqGen{
		thinc_serverReply reply;
		byte count;			//:8;
		byte mbz;			//:8;
	}

	public static final int T_REPLY_APP_LIST_SIZE = 2;

	/* T_REQ_APPLIST
	 * ask the server for a particular app (from the list sent by APP_LIST). 
	 * The server may reply notok... 
	 */
	public class thinc_reqAppGet extends thinc_reqGen{
		thinc_clientReq	req;
		byte app_id;		//:8;
		byte mbz;			//:8;
	}

	public static final int T_REQ_APP_GET_SIZE = 2;

	public class thinc_replyAppGet extends thinc_reqGen{
		thinc_serverReply reply;
	}

	public static final int T_REPLY_APP_GET_SIZE = 0;

	/* this union allows the server to read the request in one swoop (using
	 * header.length field for size). */
	/* Young Jin Yoon : In java, I used inheritance rather than unions */
	public class thinc_reqGen {}
/*
 * 		Inherited classes... 
 * 
 * 		thinc_clientReq	req;
 *		thinc_reqFBInfo 	fbinfo;
 *		thinc_reqCursor 	cursor;
 *		thinc_reqEncoder	encoder;
 *		thinc_reqFBData		fbdata;
 *		thinc_reqCacheSz	cachesz;
 *		thinc_reqVideo		video;
 *		thinc_reqNoVideo	novideo;
 *		thinc_reqVideoServFmts	video_serv_fmts;
 *		thinc_reqVideoClientFmts video_client_fmts;
 *		thinc_reqAppSharing	app_sharing;
 *		thinc_reqAppList	app_list;
 *		thinc_reqAppGet		app_get;
 */
	
	
	/* this union allows the client to read the request in one swoop (using
	 * header.length field for size). */
	/* Young Jin Yoon : In java, I used inheritance rather than unions */
	public class thinc_replyGen {}
/*
 * 		Inherited classes... 
 * 
 * 		thinc_serverReply reply;
 *		thinc_replyOK		ok;
 *		thinc_replyNotOK	notok;
 *		thinc_replyUnknown	unknown;
 *		thinc_replyFBInfo 	fbinfo;
 *		thinc_replyCursor 	cursor;
 *		thinc_replyEncoder	encoder;
 *		thinc_replyFBData	fbdata;
 *		thinc_replyCacheSz	cachesz;
 *		thinc_replyVideoServFmts video_serv_fmts;
 *		thinc_replyAppList	app_list;
 *		thinc_replyAppGet	app_get;
 */
	/*******
	 * end *
	 *******/

}
