/**
 * THINC video protocol specification, JAVA
 * @author Ricardo Baratto
 */

	/* Bird's-eye view:
	 * Upon initialization, the server and client negotiate video formats.
	 * If a consensus is reached, video support is enabled.
	 * When a video needs to be played, the server sends a message to the client
	 * requesting a video channel. The request contains details about the
	 * video such as position on the screen, size, and format. If the client
	 * is able to support it, it sends back an acknowledgement to the server.
	 * Subsequently, the server streams the video to the client. The way
	 * the video is sent is dependent on the format.
	 */
public interface ThincProtoVideo {
	

	/* server -> client */
	public static final int T_VIDEO_SINIT =	100;	/* thinc_vidSInit */
	public static final int T_VIDEO_START =	110;	/* thinc_vidStart */
	public static final int T_VIDEO_NEXT  =	111;	/* thinc_vidNext */
	public static final int T_VIDEO_END	  =	112;	/* thinc_vidEnd */
	public static final int T_VIDEO_MOVE  =	113;	/* thinc_vidMove */
	public static final int T_VIDEO_SCALE =	114;	/* thinc_vidScale */
	public static final int T_VIDEO_RESIZE=	115;	/* thinc_vidResize */

	/* T_VIDEO_SINIT
	 * Description: Server video inialization msg. Accompanied by formats 
	 * supported by the server. Contains number of video formats.
	 * Size: 4 bytes
	 * Data: The video formats
	 */
	public class thinc_vidSInit {
		int numformats;		//:32;
		/* followed by array of numformats thinc_vidFormat (below) */
	}
	
	public static final int T_VIDEO_SINIT_SIZE = 4;

	/* For now we only use FOURCC format descriptions, which both the server
	 * and client should be able to understand. In the future, a more thorough
	 * description of the format may be necessary */
	public class thinc_vidFormat {
		int id;		//:32;	/* FOURCC id of this format */
	} 

	public static final int T_VIDEO_FORMAT_SIZE	= 4;

	/* some format identifiers */
	public static final int T_FOURCC_YV12 = 0x32315659;  /* {'2','1','V','Y'} */
	public static final int T_FOURCC_YUY2 = 0x32595559;  /* {'2','Y','U','Y'} */
	public static final int T_FOURCC_UYVY = 0x59565955;  /* {'Y','V','U','Y'} */

	/* T_VIDEO_START
	 * Description: Asks the client to start playing video which will come in 
	 * the format specified in the message. It contains information about the 
	 * video, such as format (which must be one of the agreed formats), 
	 * dimensions, position, etc. It also defines an ID for this video stream.
	 * The flags field in the header will have additional information about the
	 * video stream, e.g. if it will be sent compressed.
	 * Size: 16 bytes
	 * Data: None
	 */
	public class thinc_vidStart {
		int id;		//:32;	/* id for this video stream */
		/* format id */
		int fmt_id;	//:32;	/* format. Only send id, which should have
					  	//been set on initialization */
		/* position */
		short x;		//:16;
		short y;		//:16;
		/* dimensions */
		short width; 	//:16;
		short height;	//:16;
		/* for client scaling */
		short dst_width;		//:16;
		short dst_height; 	//:16;
		/* followed by struct with format specific info (possible list
		 * below) */
		/*
		struct thinc_vidInfoYV12 yv12;
		*/
	} 
	
	public static final int T_VIDEO_COMPRESSED	= 0x01;

	public static final int T_VIDEO_START_SIZE = 20;

	/* format specific messages */
	public class thinc_vidInfoYV12 {
		int yuv_yoffset;//:32;
		int yuv_uoffset;//:32;
		int yuv_voffset;//:32;
		int yuv_ypitch;//:32;
		int yuv_upitch;//:32;
		int yuv_vpitch;//:32;
	} 

	public static final int T_VIDEO_INFO_YV12_SIZE = 24;

	public class thinc_vidInfoYUY2 {
		int pitch;//:32;
	}

	public static final int T_VIDEO_INFO_YUY2_SIZE = 4;

	public class thinc_vidInfoUYVY {
		int pitch;//:32;
	} 

	public static final int T_VIDEO_INFO_UYVY_SIZE	= 4;

	/* T_VIDEO_NEXT
	 * Description: Data that should be used next to continue playing video
	 * Size: 12 bytes
	 * Data: The video data
	 */
	public class thinc_vidNext {
		int id;		//:32;	/* stream id */
		int size;	//:32;	/* size of video data */
		int timestamp;	//:32;	/* timestamp for synchronization */
		/* followed by video data */
	} 

	public static final int T_VIDEO_NEXT_SIZE = 12;

	/* T_VIDEO_END
	 * Description: Video for stream id has finished. Client should clean things
	 * up.
	 * Size: 4 bytes
	 * Data: None
	 */
	public class thinc_vidEnd {
		int id;		//:32;
	}

	public static final int T_VIDEO_END_SIZE = 4;

	/* T_VIDEO_MOVE
	 * Description: Move the position of the video stream id to the new
	 * coordinates
	 * Size: 8 bytes
	 * Data: None
	 */
	public class thinc_vidMove {
		int id;		//:32;
		short x;		//:16;
		short y;		//:16;
	}

	public static final int T_VIDEO_MOVE_SIZE = 8;

	/* T_VIDEO_SCALE
	 * Description: Change destination dimensions of the video stream id
	 * Size: 8 bytes
	 * Data: None
	 */
	public class thinc_vidScale {
		int id;		//:32;
		short dst_width;	//:16;
		short dst_height;	//:16;
	} 

	public static final int T_VIDEO_SCALE_SIZE = 8;

	/* T_VIDEO_RESIZE
	 * Description: Change the source size of the video. Source size represents
	 * how big the frames being sent by the server to the client are. Used when
	 * client's window is smaller than original video.
	 * Size: 8 bytes
	 * Data: None
	 */
	public class thinc_vidResize {
		int id;		//:32;
		short width;	//:16;
		short height;	//:16;
		/* followed by struct with format specific info (possible list
		 * below) */
		/*
		struct thinc_vidInfoYV12 yv12;
		*/
	} 

	public static final int T_VIDEO_RESIZE_SIZE	= 8;

	/************** client -> server *******************/
	public static final int T_VIDEO_CINIT	= 100;	/* thinc_vidCInit */
	public static final int T_VIDEO_STARTOK	= 110;	/* thinc_vidStartOK */

	/* T_VIDEO_CINIT
	 * Client initialization msg. Sent in response to a VIDEO_SINIT message.
	 * It contains the subset of formats which the client can support. If the 
	 * list is empty, video initialization fails */
	public class thinc_vidCInit {
		int numformats;	//:32;
		/* followed by array of numformats thinc_vidFormat (above) */
	} 

	public static final int T_VIDEO_CINIT_SIZE = 4;

	/* T_VIDEO_STARTOK
	 * Response to T_VIDEO_START. If ok is T_VIDEO_OK, client is able to play 
	 * the video. Otherwise, video can't be played and server should refrain 
	 * from sending video data */
	public class thinc_vidStartOK {
		int		id;	//:32;	/* video stream for which client is replying */
		byte 	ok;	//:8;	/* T_VIDEO_OK/T_VIDEO_NOTOK */
		byte pad[] = new byte[3];
	} 

	public static final int T_VIDEO_OK = 1;
	public static final int T_VIDEO_NOTOK = 0;

	public static final int T_VIDEO_STARTOK_SIZE = 5;

}
