/* Copyright 2006, 2007 Columbia University */
/* This file is part of THINC.

    THINC is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 2 of the License, or
    (at your option) any later version.

    THINC is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with THINC.  If not, see <http://www.gnu.org/licenses/>.
*/
import java.nio.ByteBuffer;
import java.nio.IntBuffer;

/**
 * class for filling the regions with pixel-map images to the frame buffer. it is almost as same as <code>ThincSMsgFillBilevel</code>
 * @author  Young Jin Yoon
 * @see ThincSMsgFB
 * @see  ThincSMsgFBFillBilevel
 */
public final class ThincSMsgFBFillPixMap extends ThincSMsgFB {
	public static final int MESSAGE_TYPE = ThincProto.T_FB_PFILL;
	public static final int MESSAGE_SIZE = ThincProto.T_FB_PFILL_SIZE;
	public static final int RECT_MESSAGE_SIZE = ThincProto.T_RECTREGION_SIZE;
	
	public static final int FLAG_CACHED = ThincProto.T_FB_PIXMAP_CACHED;
	public static final int FLAG_ADDCACHE = ThincProto.T_FB_PIXMAP_ADDCACHE;
	public static final int FLAG_RESIZED = ThincProto.T_FB_PFILL_RESIZED;

	private ThincSMsgFlagCache cache;

	/**
     * creates ThincSMsgFBFillPixMap object.
     * @param h <code>ThincMsgHandler</code> to handle the message
     * @param c <code>ThincClientCanvas</code> to get the information
	 */
	public ThincSMsgFBFillPixMap(ThincMsgHandler h, ThincClientCanvas c) {
		super(h, c);
		cache = new ThincSMsgFlagCache(h,c);// create object corresponding with cache flag 
	}


	/**
	 * process()
	 * 
	 * get the source pixel-map image from cache or network and try to draw an image 
	 * using drawTile(), drawImage(), drawImageScaled() methods from 
	 * <code>ThincClientCanvas</code> class.
	 * @throws ThincException
	 */
	@Override
	public void process() throws ThincException {
		ByteBuffer bb;
		short numRects;
		short x, y, width, height;
		short dx,dy,dw,dh;
		int[] content = new int[0];
		IntBuffer contentWrap;
		int flag;

		/* get header message and parse it */
		try {
			bb = handler.getMessage(MESSAGE_SIZE);
		} catch (ThincException e) {
			throw new ThincException ("process: error on FBPixmap",e);
		}
		x = bb.getShort(); // x to draw
		y = bb.getShort(); // y to draw
		width = bb.getShort(); // width to draw
		height = bb.getShort(); // height to draw
		numRects = bb.getShort();	// number of regions represented by thinc_rectRegion.
		  							// if it is 0, just draw image from (x,y) to (x+w,y+h).

        /* get flag from ThincMsgHandler */
		flag = ((ThincMsgHandler)handler).getFlag();

        /* flag handling */
		if((flag & FLAG_CACHED) > 0) // if cached, get the pixel-map image from cache.
			content = (int[])cache.getCache(ThincClientCanvas.PIX_CACHE);
		
		if((flag & FLAG_ADDCACHE) > 0){ // if it is required to save, get the cache id information
			cache.setCacheId(ThincClientCanvas.PIX_CACHE);
		}
		if((flag & FLAG_CACHED) == 0) {
			//if it is not cached already, try to receive pixel-map image from the network
			try {
				contentWrap = handler.getMessage(width * height * canvas.getBpp()/8).asIntBuffer();
			} catch (ThincException e) {
				throw new ThincException ("process: error on FBPixmap",e);
			}
			content = new int[contentWrap.limit()];
			contentWrap.get(content);
		}
		/* set tile using width and height */
		canvas.setTile(x, y, width, height, content);

		/* get thinc_rectRegion and draw it using for loop */
		for(int i=0; i<numRects; i++)
		{
			try {
				bb = handler.getMessage(RECT_MESSAGE_SIZE);
			} catch (ThincException e) {
				throw new ThincException ("process: error on FBPixmap",e);
			}
			dx = bb.getShort(); // get x
			dy = bb.getShort(); // get y
			dw = bb.getShort(); // get width
			dh = bb.getShort(); // get height

			/*draw the image using drawTile in ThincClientCanvas class */
			canvas.drawTile(dx, dy, dw, dh);
		}
		if((flag & FLAG_ADDCACHE) > 0) { //if it is required to save, save it into cache
			cache.addCache(content.clone());
		}
	}

	
}
