/* Copyright 2006, 2007 Columbia University */
/* This file is part of THINC.

    THINC is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 2 of the License, or
    (at your option) any later version.

    THINC is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with THINC.  If not, see <http://www.gnu.org/licenses/>.
*/
import java.awt.Rectangle;
import java.nio.ByteBuffer;
import java.nio.IntBuffer;

/**
 * class for drawing an image into the frame buffer.
 * @author  Young Jin Yoon
 * @see  ThincSMsgFB
 */
public final class ThincSMsgFBRawUpdate extends ThincSMsgFB {
	public static final int MESSAGE_TYPE = ThincProto.T_FB_RUPDATE;
	public static final int MESSAGE_SIZE = ThincProto.T_FB_RUPDATE_SIZE;
	
	public static final int FLAG_COMPRESSED = ThincProto.T_FB_RUPDATE_COMPRESSED;
	public static final int FLAG_CACHED = ThincProto.T_FB_RUPDATE_CACHED;
	public static final int FLAG_ADDCACHE = ThincProto.T_FB_RUPDATE_ADDCACHE;
	public static final int FLAG_RESIZED = ThincProto.T_FB_RUPDATE_RESIZED;
	
	private ThincSMsgFlagCache cache;
	private ThincSMsgFlagFBZData zipData;
	private ThincSMsgFlagResize resize;
	private int flag;
	
	/**
     * creates ThincSMsgFBRawUpdate object.
     * @param h <code>ThincMsgHandler</code> to handle the message
     * @param c <code>ThincClientCanvas</code> to get the information
	 */
	public ThincSMsgFBRawUpdate(ThincMsgHandler h, ThincClientCanvas c) {
		super(h, c);
		cache = new ThincSMsgFlagCache(h,c); // create object corresponding with cache flag 
		zipData = new ThincSMsgFlagFBZData(h); // create object corresponding with compressed flag 
		resize = new ThincSMsgFlagResize(h,c); // create object corresponding with resize flag 
	}
	
	/**
	 * process()
	 * 
	 * get the image to draw from network or cache and try to draw an image 
	 * using drawImage() and drawImageScaled() method from <code>ThincClientCanvas</code> class.
	 * @throws ThincException
	 */
	@Override
	public void process() throws ThincException {
		ByteBuffer bb;
		short x, y, width, height;
		int zipType = ThincSMsgFlagFBZData.get(), drawType;
		/* identify the zipType and try to decide drawType which will be used 
		   in drawImage() and drawImageScaled() method in ThincClientCanvas. */
		if(zipType == ThincSMsgFlagFBZData.PNG) drawType = ThincClientCanvas.TYPE_PNG;
		else drawType = ThincClientCanvas.TYPE_RAW;

		/* get header message and parse it */
		try {
			bb = handler.getMessage(MESSAGE_SIZE);
		} catch (ThincException e) {
			throw new ThincException ("process: error on FBRawUpdate",e);
		}
		x = bb.getShort(); // x to draw
		y = bb.getShort(); // y to draw
		width = bb.getShort(); // width to draw
		height = bb.getShort(); // height to draw

		/* initialize the result array representing a final image */
		int[] content;
		IntBuffer contentWrap = null;
		
	    /* get flag from ThincMsgHandler */
		flag = ((ThincMsgHandler)handler).getFlag();

		/* flag handling */
		/*FIXME: different from protocol design, resize comes first */
		if((flag & FLAG_RESIZED) > 0) resize.process(new Rectangle(x,y,width,height));
		if((flag & FLAG_CACHED) > 0) { // if cached, get the image from cache.
			content = (int[])cache.getCache(ThincClientCanvas.IMG_CACHE);
			contentWrap = IntBuffer.wrap(content);
		}
		if((flag & FLAG_ADDCACHE) > 0) // if it is required to save, get the cache id information
			cache.setCacheId(ThincClientCanvas.IMG_CACHE);
		// compare resized flag message into actual resize using process() method in ThincSMsgFlagResize class */
		if((flag & FLAG_CACHED) > 0) { //if it is cached, just draw the image into frame buffer
			if(resize.isProceed())
				canvas.drawImage(contentWrap, x, y, width, height,drawType);
			resize.isProceed(true);
			return;
		}	
		if((flag & FLAG_COMPRESSED) > 0) contentWrap = zipData.process();
		// if compressed, receive the message and unzip it using ThincSMsgFBZData class.
		else { // if not, just receive the message and save it into int array
			try {
				contentWrap = handler.getMessage(width).asIntBuffer();
			} catch (ThincException e) {
				throw new ThincException("process: error on FBRawUpdate",e);
			}
		}
		/* if it requires to save, save it into cache */
		if((flag & FLAG_ADDCACHE) > 0) 
		{
			content = new int[contentWrap.limit()];
			contentWrap.get(content);
			cache.addCache(content.clone());
			contentWrap.flip();
		}
		/* draw image using drawImage() method from ThincClientCanvas Class.
		 * isProceed() methods from ThincSMsgFlagResize class represents
		 * the target image can be proceed as a proper resized image or not.
		 * default value should be true. so we can use it for not resized image. */
		if(resize.isProceed()) 
			canvas.drawImage(contentWrap, x, y, width, height,drawType);
		resize.isProceed(true);
	}

}
