/* Copyright 2006, 2007 Columbia University */
/* This file is part of THINC.

    THINC is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 2 of the License, or
    (at your option) any later version.

    THINC is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with THINC.  If not, see <http://www.gnu.org/licenses/>.
*/
import com.sixlegs.png.*;
import java.awt.image.BufferedImage;
import java.awt.image.PixelGrabber;
import java.io.ByteArrayInputStream;
import java.io.IOException;
import java.nio.ByteBuffer;
import java.nio.IntBuffer;
import java.util.zip.DataFormatException;
import java.util.zip.Inflater;
/**
 * class for compress flag message and unzipping the compressed image.
 * @author  Young Jin Yoon
 * @see ThincSMsgFlag
 * @see  ThincSMsgFBRawUpdate
 */
public final class ThincSMsgFlagFBZData extends ThincSMsgFlag {
	public static final int MESSAGE_SIZE = ThincProto.T_FB_ZDATA_SIZE;

	/* zip flags */
	public static final int NONE = ThincProto2.T_ENCODER_NONE; // raw image
	public static final int PNG = ThincProto2.T_ENCODER_PNG;	// PNG image
	public static final int ZLIB = ThincProto2.T_ENCODER_ZLIB;	// zip image
	
	/* for ziped Image */
	private static Inflater decompressor = new Inflater();
	private static int state = ZLIB;

	private ThincMsgHandler handler;
	private static ByteBuffer hd;
	
	private static int zsize;
	private static int usize;
	private static byte[] ziped;
	private static byte[] unziped;
	private static IntBuffer resultInt;
	private int[] pixels = null;
	
	/**
     * creates ThincSMsgFlagFBZData object.
     * @param h <code>ThincMsgHandler</code> to handle the message
	 */
	public ThincSMsgFlagFBZData(ThincMsgHandler h) {
		handler = h;
	}
	/**
	 * set()
	 * set the decompressing method by using flags.
	 * @param flag flag to use for decompression - NONE, PNG and ZLIB
	 * @return boolean value to set properly or not
	 */
	public static boolean set(int flag) {
		if(flag != NONE && flag != PNG && flag != ZLIB) return false;
		state = flag;
		return true;
	}
	/**
	 * get()
	 * gets the current decompressing method
	 * @return flag for decompressing method
	 */
	public static int get(){return state;}
	
	/**
	 * process()
	 * 
	 * receives the header for compression and try to unzip the
	 * compressed image.
	 * @return IntBuffer type of unziped image
	 * @throws ThincException
	 */
	public IntBuffer process() throws ThincException {
		/* get the compression information header from the network */
		try {
			hd = handler.getMessage(MESSAGE_SIZE);
		} catch (ThincException e) {
			throw new ThincException("decompress: error on process",e);
		}
		zsize = hd.getInt();
		usize = hd.getInt();
		
		/* allocate the byte array for ziped image and unziped image */
		ziped = new byte[zsize];
		unziped = new byte[(zsize<<1 > usize)?(zsize<<1):usize];
		
		/* process the image with proper compression algorithm */
		if(state == ZLIB) return processZLIB();
		else if(state == PNG) return processPNG();
		else return null;
	}
	

	private IntBuffer processPNG() throws ThincException {
		ByteBuffer bb;
		/* get ziped image from the network */
		try {
			bb = handler.getMessage(zsize);
		} catch (ThincException e) {
			throw new ThincException("process: error on Contents of FBData",e);
		}
		bb.get(ziped);
		/* generate PNG image using PNG library */
		ByteArrayInputStream tar = new ByteArrayInputStream(ziped,0, zsize);
		PngImage a = new PngImage();
		BufferedImage buf = null;
		try {
			buf = a.read(tar, true);
		} catch (IOException e1) {
			throw new ThincException("process: error on Contents of FBZData");
		}
		
		/* grab the pixel of image and converts it into int array */
		if(pixels == null || pixels.length < buf.getWidth()*buf.getHeight()) 
			pixels = new int[buf.getWidth()*buf.getHeight()];
		PixelGrabber pxg = new PixelGrabber(buf, 0, 0, buf.getWidth(), buf.getHeight(), pixels,0,buf.getWidth());
		try {
			pxg.grabPixels();
		} catch (InterruptedException e) {
			throw new ThincException("process: error on Contents of FBZData",e);
		}
		/* wrap int array into IntBuffer class and returns it */
		resultInt = IntBuffer.wrap(pixels);
		resultInt.limit(buf.getWidth()*buf.getHeight());
		return resultInt;
        
	}
	
	public IntBuffer processZLIB() throws ThincException {
		int retVal;
		ByteBuffer bb;
		/* get ziped image from the network */		
		try {
			bb = handler.getMessage(zsize);
		} catch (ThincException e) {
			throw new ThincException("process: error on Contents of FBData",e);
		}
		bb.get(ziped);

		/* set input for decompressor */
		decompressor.setInput(ziped,0,zsize);

		/* decompress it */
		try {
		retVal = decompressor.inflate(unziped);
		decompressor.reset();
		} catch (DataFormatException e) {
			throw new ThincException("decompress: error on unzip zlib",e);
		}
		
		/* wrap int array into IntBuffer class and returns it */
		resultInt = ByteBuffer.wrap(unziped).asIntBuffer();
		resultInt.limit(retVal/4);
		return resultInt;
	}
}
