/* Copyright 2006, 2007 Columbia University */
/* This file is part of THINC.

    THINC is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 2 of the License, or
    (at your option) any later version.

    THINC is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with THINC.  If not, see <http://www.gnu.org/licenses/>.
*/
import java.nio.ByteBuffer;
import java.nio.ByteOrder;

/**
 * handles the sound messages only from/to the server.  it recoginizes the message's type, flag and length then pass them into  the proper message object. Moreover, it also takes the control from client messages to send a message.
 * @author  Young Jin Yoon
 * @see  ThincHandler
 */
public final class ThincSoundHandler extends ThincHandler implements Runnable {	
	private static final int MAX_MESSAGE_TYPE = ThincProtoSound.T_SND_VOLUME_UP;
	private static final int FIRST_MESSAGE_TYPE = ThincProtoSound.T_SND_DATA;
	private ThincSMsgSound[] serverMessage;
	protected ByteBuffer hd = null;
	
	private static final byte[] TC_SND_VERSION  = new byte[] {0,1,1};

	/**
	 * create <code>ThincSoundHandler</code> object. it also send initialize
	 * message to the server.
	 * @param tc <code>ThincClient</code> object which invokes this
	 * @throws ThincException
	 */	
	public ThincSoundHandler(ThincClient tc, ThincClientCanvas ca) throws ThincException{ 
		super(tc,ca);

		/* initializing serverMessage array by using initMessage() */
		initMessage();
		
		/* try to make the sound connection to the server */
		try {
			createSocketChannel(client.getHost(),client.getSndPort());
		} catch (ThincException e) {
			throw new ThincException("ThincSoundHandler: cannot create socketchannel",e);
		}

		/* allocating header ByteBuffer object */
		hd = ByteBuffer.allocateDirect(THINC_HEADERSIZE);
		hd.order(ByteOrder.BIG_ENDIAN);
		

		/* do SSL handshake first */
		if((client.getSecCap() & client.SEC_ENC)> 0) setSSLSocketChannel(client.getHost(),client.getSndPort());

		/* send initial message to the server - currently do nothing*/
		sendStart();

		/* receive 16 bits and throw out */
		hd.limit(2);
		try {
			recvBytes(hd);
		} catch (ThincException e) {
			System.err.println("initialize: Error on recvBytes");
			e.printStackTrace();
			return;
		}
		hd.flip();
		hd.clear();
	
		System.out.println("<SOUND> Sound init Complete!");
	}

	private void initMessage() {
		/* allocating serverMessage array */
		serverMessage = new ThincSMsgSound[MAX_MESSAGE_TYPE - FIRST_MESSAGE_TYPE];

		/* initialize all serverMessage element to null */
		for(int i=0; i< serverMessage.length; i++)
			serverMessage[i] = null;

		/* register proper serverMessage element with proper sound classes */
		serverMessage[ThincSMsgSoundOpen.MESSAGE_TYPE - FIRST_MESSAGE_TYPE] = new ThincSMsgSoundOpen(this);
		serverMessage[ThincSMsgSoundClose.MESSAGE_TYPE - FIRST_MESSAGE_TYPE] = new ThincSMsgSoundClose(this);
		serverMessage[ThincSMsgSoundPlay.MESSAGE_TYPE - FIRST_MESSAGE_TYPE] = new ThincSMsgSoundPlay(this);
	}
	/**
	 * run()
	 * 
	 * implementation of Runnable interface from Java.
	 * it starts to receive message from server. if message comes,
	 * this parses general thinc header and pass the control to the proper
	 * server message class.
	 */
	public void run() {
		byte type, flag;
		short length;
		while(true) { // keep trying to receive message 
			try {
				recvBytes(hd);
			} catch (ThincException e) {
				System.err.println("receiver: Error on recvBytes");
				e.printStackTrace();
				break;
			}
			hd.flip();

			/* get the type, flag and length of the message */
			type = hd.get();
			flag = hd.get();
			length = hd.getShort();
			hd.clear();			
			
			/* pass the control to the proper message*/		
			if(type >= FIRST_MESSAGE_TYPE && type < MAX_MESSAGE_TYPE && serverMessage[type-FIRST_MESSAGE_TYPE] != null) {
				try {
					serverMessage[type-FIRST_MESSAGE_TYPE].process();
				} catch (ThincException e) {
					System.err.println("receiver: cannot process serverMessage");
					e.printStackTrace();
					break; // exit this thread by using break statement.
				}
			}
			else {
				System.err.println("receiver: Unknown"+type);
				break;
			}
		}
	}
	/**
	 * sentStart()
	 * 
	 * sends initialization message to the server.
	 * @throws ThincException
	 */
	private void sendStart() throws ThincException{
		/* create CMsgSoundStart object and try to send it using process() method */
/*		ThincCMsgSoundStart sndStart = new ThincCMsgSoundStart(this,TC_SND_VERSION);
		try {
			sndStart.process();
		} catch (ThincException e) {
			throw new ThincException("sendStart: sound init failed",e);
		}*/
	}		
}
