/* Copyright 2004, 2005, 2006, 2007 Ricardo Baratto */
/* Copyright 2004, 2005, 2006, 2007 Columbia University */
/* This file is part of THINC.

    THINC is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 2 of the License, or
    (at your option) any later version.

    THINC is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with THINC.  If not, see <http://www.gnu.org/licenses/>.
*/
/*
 * Resize support for thinc client
 *
 * Author:
 * Ricardo Baratto
 * Network Computing Lab - Columbia University
 *
 */

/* set tabstop 8 */

#include "thinc_client.h"

#include "tc_x11.h"
#include "thinc_resize.h"
#include <stdlib.h>
#include <netinet/in.h>
#include <string.h>
#include <sys/mman.h>
#include <errno.h>
#include <string.h>

/* when wm uses opaque resizing we need to set up a timer to not resize on
 * every intermediate size sent by the wm. This is how long to wait until
 * we try to resize the window (in us) */
#define TC_RESIZE_DELAY	100000

/* delta in size must be larger than this (w*h) for us to ask the server
 * for updated content 
 * currently: 10x10 pixels
 */
#define TC_DSIZE_LIMIT	100

/* TODO: Maybe we can provide the resize library with the temporary
 * storage needed to do rows & columns... */
/* current and original size are stored in the fb struct */
struct tc_resize {
	Bool do_resize;
	/* new size (resizing is done asynchronously) */
	int new_width;
	int new_height;
	double factor_x;
	double factor_y;
	/* local copy of the resized fb */
	XImage *img;
	unsigned char update_img;	/* if TRUE, should do a GetImage 
					   before using it as source */
	int (*resize)(char *, char *,
		unsigned int, unsigned int,     /* width */
		unsigned int, unsigned int,     /* height */
		unsigned int, unsigned int);    /* stride */
};

/* local funcs */
static int tc_resize_update_server(struct tc_fb *fb, struct tc_resize *resize);
static int tc_resize_ask_server(struct tc_fb *fb, struct tc_resize *resize);
#if 1	/* NOT YET */
static int tc_resize_handle_old_update(struct tc_fb *fb, thinc_reszData *d,
	thinc_header *h);
static int tc_resize_read_data(struct tc_fb *fb, thinc_reszData *d,
	thinc_header *h, char *data, unsigned int width, 
	unsigned int height, unsigned int pitch);
static int tc_resize_read_zdata(struct tc_fb *fb, char *data,
	int total, unsigned int width, unsigned int height, unsigned int pitch);
#endif

int tc_resize_init(struct tc_fb *fb)
{
	struct tc_resize *resize;

	if(!fb->do_resize) {
		return TC_OK;
	}

	fb->resize = malloc(sizeof(struct tc_resize));
	if(!fb->resize) {
		tclient_msg("Error: Couldn't allocate memory for resize "
			"structure\n");
		return TC_ERROR;
	}

	resize = fb->resize;
	resize->new_width = resize->new_height = 0;
	resize->factor_x = resize->factor_y = 0;
	resize->do_resize = T_FALSE;
	resize->img = NULL;
	resize->update_img = T_TRUE;	/* just in case */

	switch(fb->bpp) {
		case 8:
			resize->resize = thinc_resize8;
			break;
		case 16:
			resize->resize = thinc_resize16;
			break;
		case 24:
		case 32:
			resize->resize = thinc_resize32;
			break;
		default:
			tclient_msg("Error: Unsupported depth for resizing "
				"%d bpp\n", fb->bpp);
			return TC_ERROR;
	}

	return TC_OK;
}

void tc_resize_shutdown(struct tc_fb *fb)
{
	if(!fb->do_resize)
		return;
#ifndef NO_X
	if(fb->resize->img)
		XDestroyImage(fb->resize->img);
#endif	/* ! NO_X */
	free(fb->resize);
	fb->resize = NULL;
}

/*** handle resizing ***/

/* queue a resize event */
int tc_resize_resize(struct tc_fb *fb, int width, int height)
{
	int err;

	if(!fb->do_resize)
		return TC_OK;

	if( 	(width == fb->curr_width) &&
		(height == fb->curr_height) ) {
		/* do nothing */
		fb->resize->do_resize = T_FALSE;
		return TC_OK;
	}

	tclient_msg("resized to %ux%u\n", width, height);

	fb->resize->new_width = width;
	fb->resize->new_height = height;
	fb->resize->do_resize = T_TRUE;

	/* for stupid opaque resizing :( this ends up being the response time 
	 * of the client to a resize event. Need to play with number */
	set_timer(ALARM_RESIZE, 0, 200000, err);
	if(err == -1) {
		tclient_msg("Error: Couldn't set resize timer: %s\n",
			strerror(errno));
		return TC_ERROR;
	}

	return TC_OK;
}

/* does the resizing.
 *
 * if new size is smaller, everything's local. If not, need to ask
 * server for help 
 *
 * what to do:
 * - identify source pix. if not resized, use original pix (need flag)
 * => create new pix
 * => fill it with source pix
 * => destroy source pix if not original
 * => tell server about new size
 * => resize video
 * if new size > current
 * 	=> ask server to update everything
 * if new == original
 * 	=> destroy resized pix and invalidate everything related to resizing
 */
int tc_resize_do_resize(struct tc_fb *fb)
{
	const char *func = __func__;
	struct tc_x11 *x11 = fb->x11;
	struct tc_resize *resize = fb->resize;
	Pixmap src, dst;
	XImage *srcimg, *dstimg;
	/* 4 cases:
	 * free_rsz && replace_rsz
	 * => from resized size to resized size
	 * !free_rsz && replace_rsz
	 * => from normal size to resized size
	 * free_rsz && !replace_rsz
	 * => from resized size to normal size
	 * !free_rsz && !replace_rsz
	 * => BUG. going from normal to normal
	 */
	int 	free_rsz, /* destroy resizedP pixmap at the end */
		replace_rsz; /* replace resized pixmap with dst at the end */
	int err;
	XRectangle rect;

	if(!resize->do_resize)
		return TC_OK;

	tclient_dbg(1, ("%s: resized to %dx%d\n", func, resize->new_width,
		resize->new_height));

	/* first tell server about new size to minimize updates with wrong
	 * size */
	err = tc_resize_update_server(fb, resize);
	if(err == TC_ERROR) {
		tclient_msg("Error: %s: Couldn't inform server of size "
			"change\n", func);
		return TC_ERROR;
	}

	/* if needed, ask server for updated content */
	if(	resize->new_width > fb->curr_width ||
		resize->new_height > fb->curr_height) {
		tclient_dbg(1, ("%s: size increased, asking server for help\n",
			func));
		err = tc_resize_ask_server(fb, resize);
		if(err == TC_ERROR) {
			tclient_msg("Error: %s: Couldn't request data update "
				"from server\n", func);
			return TC_ERROR;
		}
	}

#ifndef NO_X
	/* where are we coming from? */
	if(x11->resized) {
		tclient_dbg(1, ("%s: we were resized\n", func));
		src = x11->resizedP;
		if(resize->update_img) {
			tclient_dbg(1, ("%s: our image is out of date, "
				"refreshing\n", func));
			/* oops, need to update image before we can use it */
			XGetSubImage(x11->disp, x11->resizedP, 0, 0, 
				fb->curr_width, fb->curr_height, 0xFFFFFFFF, 
				ZPixmap, resize->img, 0, 0);
			resize->update_img = T_FALSE;
		}
		srcimg = resize->img;
		free_rsz = T_TRUE;
		XDestroyRegion(x11->resizedRgn);
		x11->resizedRgn = NULL;
	} else {
		tclient_dbg(1, ("%s: we were not resized\n", func));
		/* ugh.. local image is never up to date so need to do
		 * XGetImage() before we can use it :( */
		/* TODO: 0xFFFFFFFF -> should check bpp and pass appropiate
		 * mask */
		XGetSubImage(x11->disp, x11->DesktopP, 0, 0, fb->width, 
			fb->height, 0xFFFFFFFF, ZPixmap, x11->img, 0, 0);
		src = x11->DesktopP;
		srcimg = x11->img;
		free_rsz = T_FALSE;
	}

	rect.x = 0;
	rect.y = 0;
	rect.width = resize->new_width;
	rect.height = resize->new_height;

	/* where are we going to? */
	if(  	(resize->new_width == fb->width) &&
		(resize->new_height == fb->height) ) {
		tclient_dbg(1, ("%s: we're back to original size\n", func));
		dst = x11->DesktopP;
		dstimg = x11->img;
		resize->factor_x = 1.0;
		resize->factor_y = 1.0;
		XUnionRectWithRegion(&rect, x11->damageRgn, x11->damageRgn);
		replace_rsz = T_FALSE;
		x11->resized = T_FALSE;
	} else {
		tclient_dbg(1, ("%s: we're now resized\n", func));
		/* now create image */
		dst = XCreatePixmap(x11->disp, x11->Main, resize->new_width, 
			resize->new_height, x11->w_depth);
		dstimg = XGetImage(x11->disp, dst, 0, 0, resize->new_width,
			resize->new_height, 0xFFFFFFFF, ZPixmap);
		resize->factor_x = (double)resize->new_width/fb->width;
		resize->factor_y = (double)resize->new_height/fb->height;
		x11->resizedRgn = XCreateRegion();
		if(!x11->resizedRgn) {
			tclient_msg("Error: %s: Couldn't allocate necessary "
				"memory for resizing operation\n", func);
			XFreePixmap(x11->disp, dst);
			XDestroyImage(dstimg);
			return TC_ERROR;
		}
		XUnionRectWithRegion(&rect, x11->resizedRgn, x11->resizedRgn);
		replace_rsz = T_TRUE;
		x11->resized = T_TRUE;
	}

	tclient_dbg(1, ("%s: finished with preamble, resizing data to %dx%d\n", 
		func, resize->new_width, resize->new_height));

	err = resize->resize((char*)srcimg->data, 
		(char*)dstimg->data, fb->curr_width,
		fb->curr_height, resize->new_width, resize->new_height,
		srcimg->bytes_per_line, dstimg->bytes_per_line);
	if(err == T_ERROR) {
		tclient_msg("Error: %s: Could not resize current "
			"framebuffer\n", func);
		if(replace_rsz) {
			XFreePixmap(x11->disp, dst);
			XDestroyImage(dstimg);
			XDestroyRegion(x11->resizedRgn);
			x11->resizedRgn = NULL;
		}
		return TC_ERROR;
	}
	XPutImage(x11->disp, dst, x11->theGC, dstimg, 0, 0, 0, 0,
		resize->new_width, resize->new_height);
#endif	/* ! NO_X */

	/* final touches */
	tclient_dbg(1, ("%s: resized data, final touches..\n", func));

	fb->curr_width = resize->new_width;
	fb->curr_height = resize->new_height;
	x11->updated = T_TRUE;

	if(fb->use_xvideo) {
		tc_resize_video_all(fb);
		tc_resize_move_video_all(fb);
	}
	resize->do_resize = T_FALSE;

	/* clean up */
	tclient_dbg(1, ("%s: almost there, cleaning up\n", func));
#ifndef NO_X
	/* FOR NEW RESIZE: Should have persistent image?? */
	if(free_rsz) {
		/* src == resizedP */
		XFreePixmap(x11->disp, x11->resizedP);
		x11->resizedP = (Pixmap) -1;
		XDestroyImage(resize->img);
		resize->img = NULL;
	}
	if(replace_rsz) {
		x11->resizedP = dst;
		resize->img = dstimg;
		resize->update_img = T_FALSE;
	}
#endif	/* ! NO_X */
	return TC_OK;
}

/* called when somebody drew to the resized pixmap but not the resized image */
inline void tc_resize_img_outdated(struct tc_fb *fb) 
{
	if(fb->do_resize)
		fb->resize->update_img = T_TRUE;
}

/* keep the resized pixmap up to date by resizing/copying from the original
 * pixmap to the resized pixmap the contents of rectangle rect. rectangle
 * out has the coordinates which were used in the resized pixmap
 * (output param) */
static int tc_resize_update(struct tc_fb *fb, int xo, int yo, int wo, int ho, 
	int xr, int yr, int wr, int hr)
{
	tc_dbg_name(func)
	struct tc_resize *resize = fb->resize;
	char *from, *to;
	unsigned int sfrom, sto; /* strides */
	int err;

	tclient_dbg(1, ("%s\n", func));

	if(!fb->do_resize) {
		tclient_dbg(1, ("%s: resize is disabled. finished\n", func));
		return TC_OK;
	}

#ifndef NO_X
	/* update resized image. need to get pointers and strides to normal
	 * and resized image, and then call resize() */
	tc_getDataPtr(fb, fb->fbdata, fb->width, xo, yo, from, 
		sfrom);
	tc_getDataPtr(fb, (char*)resize->img->data, fb->curr_width, 
		xr, yr, to, sto);
	err = resize->resize(from, to, wo, ho, wr, hr, sfrom, sto);
	if(err == TC_ERROR) {
		tclient_msg("Error: Could not update resized window\n");
		return TC_ERROR;
	}
#endif	/* ! NO_X */

	tclient_dbg(1, ("%s: finished\n", func));
	return TC_OK;
}

int tc_resize_update_rect(struct tc_fb *fb, XRectangle *orig, XRectangle *rszd)
{
	tclient_dbg(1, ("%s: rendering rectangle (%d,%d)x(%d,%d) "
		"to (%d,%d)x(%d,%d)\n", __func__, orig->x, orig->y,
		orig->width, orig->height, rszd->x, rszd->y, rszd->width, 
		rszd->height));

	return tc_resize_update(fb, orig->x, orig->y, orig->width, 
		orig->height, rszd->x, rszd->y, rszd->width, rszd->height);
}

int tc_resize_update_box(struct tc_fb *fb,tc_xlib_box *orig,tc_xlib_box *rszd)
{
	return tc_resize_update(fb, orig->x1, orig->y1, tc_box_width(orig),
		tc_box_height(orig), rszd->x1, rszd->y1, tc_box_width(rszd),
		tc_box_height(rszd));
}

int tc_resize_update_rgn(struct tc_fb *fb, const tc_xlib_rgn *rgn, 
	const tc_xlib_rgn *rgn_rsz)
{
	const char *func = __func__;
	struct tc_x11 *x11 = fb->x11;
	struct tc_resize *resize = fb->resize;
	register tc_xlib_box *orig, *rszd;
	int err, i, count;
	XRectangle rect;

	tclient_dbg(1, ("%s\n", func));

	if(!fb->do_resize) {
		tclient_dbg(1, ("%s: resize is disabled. finished\n", func));
		return TC_OK;
	}

	if(tc_rgn_numrects(rgn) != tc_rgn_numrects(rgn_rsz) ) {
		tclient_msg("Error: %s: number of rectangles doesn't match\n",
			func);
		return TC_ERROR;
	}

	tc_rgn_extents_rect(rgn, &rect);

	count = tc_rgn_numrects(rgn);
	orig = tc_rgn_rects(rgn);
	rszd = tc_rgn_rects(rgn_rsz);
	for(i=0; i<count; ++i, ++orig, ++rszd) {
		if(!tc_box_width(rszd) || !tc_box_height(rszd) )
			continue;
		tclient_dbg(1, ("%s: rendering rectangle (%d,%d)x(%d,%d) "
			"to (%d,%d)x(%d,%d)\n", func, orig->x1, orig->y1,
			tc_box_width(orig), tc_box_height(orig), rszd->x1,
			rszd->y1, tc_box_width(rszd), tc_box_height(rszd) ));
		err = tc_resize_update_box(fb, orig, rszd);
		if(err == T_ERROR) {
			tclient_msg("Error: %s: Could not resize rectangle "
				"number %d\n", func, i);
			return TC_ERROR;
		}
#ifndef NO_X
		XPutImage(x11->disp, x11->resizedP, x11->theGC, resize->img,
			rszd->x1, rszd->y1, rszd->x1, rszd->y1, 
			tc_box_width(rszd), tc_box_height(rszd));
#endif	/* ! NO_X */
	}

	tclient_dbg(1, ("%s: finished\n", func));
	return TC_OK;
}

int tc_resize_update_region_rects(struct tc_fb *fb, const Region r, 
	const Region rz)
{
	struct tc_x11 *x11 = fb->x11;
	tc_xlib_rgn *rgn = (tc_xlib_rgn*)r;
	tc_xlib_rgn *rgn_rsz = (tc_xlib_rgn*)rz;
	XRectangle ext;

	if(!fb->do_resize)
		return TC_OK;

#ifndef NO_X
	/* local image is never up to date so need to do
	 * XGetImage() before we can use it :( */
	XClipBox(r, &ext);
	/* TODO: 0xFFFFFFFF -> should check bpp and pass appropiate
	 * mask */
	XGetSubImage(x11->disp, x11->DesktopP, ext.x, ext.y, ext.width, 
		ext.height, 0xFFFFFFFF, ZPixmap, x11->img, ext.x, ext.y);
#endif	/* ! NO_X */

	return tc_resize_update_rgn(fb, rgn, rgn_rsz);
}

int tc_resize_update_region(struct tc_fb *fb, const Region r, 
	const Region rz)
{
	struct tc_x11 *x11 = fb->x11;
	XRectangle ext, extz;
	int err;

	if(!fb->do_resize)
		return TC_OK;

#ifndef NO_X
	/* local image is never up to date so need to do
	 * XGetImage() before we can use it :( */
	XClipBox(r, &ext);
	/* TODO: 0xFFFFFFFF -> should check bpp and pass appropiate
	 * mask */
	XGetSubImage(x11->disp, x11->DesktopP, ext.x, ext.y, ext.width, 
		ext.height, 0xFFFFFFFF, ZPixmap, x11->img, ext.x, ext.y);

	XClipBox(rz, &extz);
#endif	/* ! NO_X */
	err = tc_resize_update_rect(fb, &ext, &extz);
	if(err == TC_ERROR) {
		tclient_msg("Error: Could not update resized region\n");
		return TC_ERROR;
	}
#ifndef NO_X
	XPutImage(x11->disp, x11->resizedP, x11->theGC, fb->resize->img,
		extz.x, extz.y, extz.x, extz.y, extz.width, extz.height);
#endif	/* ! NO_X */
	return TC_OK;
}

/*** server functions ***/
/* tell server we're resized */
static int tc_resize_update_server(struct tc_fb *fb, struct tc_resize *resize)
{
	thinc_reszEvent e;
	int err;

	e.width = htons(resize->new_width);
	e.height = htons(resize->new_height);

	err = tc_writeMsg(fb, T_RESZ_EVENT, &e, T_RESZ_EVENT_SIZE);
	if(err == TC_ERROR) {
		tclient_msg("Error: %s: Couldn't send resize event to "
			"server\n", __func__);
		return TC_ERROR;
	}

	return TC_OK;
}

/* asks the server to send an update for all the fb */
static int tc_resize_ask_server(struct tc_fb *fb, struct tc_resize *resize)
{
	thinc_reszUpdate u;
	thinc_rectRegion r;
	int err, currs, news;

	/* check if increase in size is really significant */
	news = resize->new_width * resize->new_height;
	currs = fb->curr_width * fb->curr_height;

	if(news <= currs)
		return TC_OK;
	else if(news - currs < TC_DSIZE_LIMIT) 
		return TC_OK;

	u.numrects = htonl(1);
	r.x = htons(0);
	r.y = htons(0);
	r.width = htons(fb->width);
	r.height = htons(fb->height);

	err = tc_writeMsg(fb, T_RESZ_UPDATE, &u, T_RESZ_UPDATE_SIZE);
	if(err == TC_ERROR) {
		tclient_msg("Error: %s: Couldn't send update request to "
			"server\n", __func__);
		return TC_ERROR;
	}

	err = writeData(fb, &r, T_RECTREGION_SIZE);
	if(err == TC_ERROR) {
		tclient_msg("Error: %s: Couldn't send update request to "
			"server\n", __func__);
		return TC_ERROR;
	}

	return TC_OK;
}

/* handle a resized RAW udpate from the server */
static int tc_resize_add_cache(struct tc_fb *fb, unsigned long cache_id, 
	char *data);
static char* tc_resize_get_cached(struct tc_fb *fb, 
	thinc_rawUpdate *r);

int tc_resize_handle_update(struct tc_fb *fb, thinc_rawUpdate *r,
	thinc_header *h)
{
	tc_dbg_name(func)
	thinc_reszData d;
	int err;
	struct tc_x11 *x11 = fb->x11;
	char *data = NULL;
	XImage *resized;
	int offset = 0, bytes_per_line = 0, bitmap_pad = 32;
	unsigned long int cache_id = 0;
	unsigned char cache_update = T_NO, cached = T_NO;
	unsigned int update_size = 0;

	tclient_dbg(1, ("%s: reading resized RAW update\n", func));

	if(!fb->do_resize) {
		tclient_msg("Error: Received resized update, but resize "
			"is disabled!\n");
		return TC_ERROR;
	}

	err = readData(fb, &d, T_RESZ_DATA_SIZE, T_TRUE);
	if(err == TC_ERROR || err == TC_CLOSED) {
		tclient_msg("Error: Couldn't read resized message header\n");
		return TC_ERROR;
	}

	d.fb_width = ntohs(d.fb_width);
	d.fb_height = ntohs(d.fb_height);
	d.size = ntohl(d.size);

	tclient_dbg(1, ("%s: got resized RAW update: w %d, h %d, size %d\n",
		func, d.fb_width, d.fb_height, d.size));

	if(h->flags & T_FB_RUPDATE_CACHED) {
		data = tc_resize_get_cached(fb, r);
		if(!data) {
			tclient_msg("Error: Couldn't handle cached update\n");
			return TC_ERROR;
		}
		update_size = 0; /* so we don't count this update's size */
		cached = T_YES;
		tclient_dbg(1, ("%s: got cached resize update\n", func));
	}
	if(h->flags & T_FB_RUPDATE_ADDCACHE) {
		err = tc_handleUpdateAddCache(fb, &cache_id);
		if(err == TC_ERROR) {
			tclient_msg("Error: Couldn't handle cache add "
				"message\n");
			return TC_ERROR;
		}
		tclient_dbg(1, ("%s: got udpate to cache. id: %lu\n", func,
			cache_id));
		cache_update = T_YES;
	}

	if( 	(d.fb_width != fb->curr_width) || 
		(d.fb_height != fb->curr_height) ) {
		/* ignore. need to read the data tho */
		tclient_dbg(1, ("%s: ignoring resized update for old size: "
			"%dx%d\n", func, d.fb_width, d.fb_height));

		if(cached)
			return TC_OK;

		/* TODO: What happens if we're supposed to cache this 
		 * update? */
		err = tc_resize_handle_old_update(fb, &d, h);
		if(err == TC_ERROR) {
			tclient_msg("Error: Couldn't handle outdated resize "
				"update\n");
			return TC_ERROR;
		} else
			return TC_OK;
	}

	tclient_dbg(1, ("%s: updating region (%d,%d)x(%d,%d)\n", func,
		r->x, r->y, r->width, r->height));

#ifndef NO_X
	resized = XCreateImage(x11->disp, x11->visual, x11->w_depth,
		ZPixmap, offset, NULL, r->width, r->height, 
		bitmap_pad, bytes_per_line);
	if(!resized) {
		tclient_msg("Error: Couldn't create image for resized "
			"update\n");
		return TC_ERROR;
	}
	/* need to use this for correct pad when bpp != 32 */
	update_size = resized->bytes_per_line * r->height;
#else	/* ! NO_X */
	update_size = d.size;
#endif	/* ! NO_X */
	if(!cached) {
		data = malloc(update_size);
		if(!data) {
			tclient_msg("Error: Couldn't allocate memory for "
				"resized udpate\n");
			return TC_ERROR;
		}
#ifndef NO_X
		err = tc_resize_read_data(fb, &d, h, data, r->width, 
			r->height, resized->bytes_per_line);
#else
		err = tc_resize_read_data(fb, &d, h, data, r->width, 
			r->height, pix_to_bytes(fb, r->width));
#endif	/* ! NO_X */
		if(err == TC_ERROR) {
			tclient_msg("Error: Couldn't read update data\n");
			XDestroyImage(resized);
			return TC_ERROR;
		}
	}
#ifndef NO_X
	resized->data = (char*) data;
#endif	/* ! NO_X */

	tclient_dbg(1, ("%s: got image data, copying it to pixmap\n", func));

	/* now put back the image */
#ifndef TC_DONT_UPDATE_X
	XPutImage(x11->disp, x11->resizedP, x11->theGC, resized, 0, 0, 
		r->x, r->y, r->width, r->height);

	tclient_dbg(1, ("%s: put image into pixmap, destroying it\n", func));

	tc_x11update_resized(fb, r->x, r->y, r->width, r->height);
#endif

	/* frees image data too */
	if(cached || cache_update)
		/* so that x doesn't free the data */
		resized->data = NULL;
#ifndef NO_X
	XDestroyImage(resized);
#endif	/* ! NO_X */

	if(cache_update) {
		err = tc_resize_add_cache(fb,cache_id, data);
		if(err == TC_ERROR) {
			tclient_msg("Error adding update to cache\n");
			return TC_ERROR;
		}
	}

	tc_stats_add(fb->stats.rawbytes, update_size + T_FB_RUPDATE_SIZE + 
		THINC_HEADERSIZE);

	tclient_dbg(1, ("%s: finished\n", func));
	return TC_OK;
}

/* returns pointer to cached data */
static char* tc_resize_get_cached(struct tc_fb *fb, 
	thinc_rawUpdate *r)
{
	int res;
	char *data;
	unsigned long int cache_id;

	res = tc_handleUpdateCached(fb, &cache_id);
	if(res == TC_ERROR) {
		tclient_msg("Error: Couldn't read cached data message\n");
		return NULL;
	}

	data = thinc_cache_get(fb->raw_cache, cache_id);
	if(!data) {
		tclient_msg("Error: Don't have the data for cached update "
			"%lu\n", cache_id);
		return NULL;
	}

	return data;
}

static int tc_resize_add_cache(struct tc_fb *fb, unsigned long cache_id, 
	char *data)
{
	return thinc_cache_add_hash(fb->raw_cache, cache_id, data);
}

static int tc_resize_read_data(struct tc_fb *fb, thinc_reszData *d,
	thinc_header *h, char *data, unsigned int width,
	unsigned int height, unsigned int pitch)
{
	int err = TC_OK;
	if(h->flags & T_FB_RUPDATE_COMPRESSED)
		err = tc_resize_read_zdata(fb, data, d->size, width, height,
			pitch);
	else {
		int i;
		char *c = data;
		unsigned int len = pix_to_bytes(fb, width);
		for(i=0; i<height; i++) {
			err = readData(fb, c, len, T_TRUE);
			if(err == TC_ERROR || err == TC_CLOSED) 
				return TC_ERROR;
			c += pitch;
		}
	}
	
	if(err == TC_ERROR || err == TC_CLOSED)
		return TC_ERROR;
	return TC_OK;
}

static int tc_resize_read_zdata(struct tc_fb *fb, char *data,
	int total, unsigned int width, unsigned int height, unsigned int pitch)
{
	tc_dbg_name(func)
	struct tc_compress *zfb = &fb->zfb;
	int err;
	thinc_fbZData z;
	png_structp png_ptr;
	png_infop info_ptr;
	char *curr = data;
	unsigned int i;

	tclient_dbg(1, ("%s: total %d\n", func, total));

	err = readData(fb, &z, T_FB_ZDATA_SIZE, T_TRUE);
	if(err == TC_ERROR || err == TC_CLOSED) {
		tclient_dbg(1, ("Error: %s: Couldn't read compressed header\n",
			func));
		return TC_ERROR;
	}

	z.zsize = ntohl(z.zsize);
	z.usize = ntohl(z.usize);

	tclient_dbg(1, ("%s: z %d, u %d\n", func, z.zsize, z.usize));

	err = tc_pngInit(zfb);
	if(err == TC_ERROR) {
		tclient_dbg(1, ("Error; %s: Couldn't initialize png\n", func));
		return TC_ERROR;
	}
	
	png_ptr = zfb->png_ptr;
	info_ptr = zfb->info_ptr;
	png_read_info(png_ptr, info_ptr);

	for(i=0; i<height; ++i) {
		png_read_row(png_ptr, (png_bytep)curr, (png_bytep)NULL);
		curr += pitch;
	}

	tc_pngEnd(zfb);

	tclient_dbg(1, ("%s: finished\n", func));
	return TC_OK;
}

/* called when server sent a raw update for a size which is not the current
 * one. It just reads the data from the socket and discards it. Kind of 
 * crazy due to the cases it needs to handle */
static int tc_resize_handle_old_update(struct tc_fb *fb, thinc_reszData *d,
	thinc_header *h)
{
	int size, err;
	char *data;
	
	size = d->size;

	if(h->flags & T_FB_RUPDATE_COMPRESSED) {
		/* ugh... */
		thinc_fbZData z;
		do {
			err = readData(fb, &z, T_FB_ZDATA_SIZE, T_TRUE);
			if(err == TC_ERROR || err == TC_CLOSED)
				return TC_ERROR;
			z.zsize = ntohl(z.zsize);
			z.usize = ntohl(z.usize);
			data = mmap(NULL, z.zsize, PROT_WRITE, MAP_ANONYMOUS, 
				-1, 0);
			if(!data)
				return TC_ERROR;

			err = readData(fb, data, z.zsize, T_TRUE);
			if(err == TC_ERROR) {
				munmap(data, z.zsize);
				return TC_ERROR;
			}
			munmap(data, z.zsize);
			size -= z.usize;
		} while(size > 0);
	} else {
		/* anonymous mmap() and get it over with.
		 * TODO: I think this needs at least linux 2.4. mmap()'ing
		 * /dev/zero seems to be more portable */
		data = mmap(NULL, size, PROT_WRITE, MAP_ANONYMOUS, -1, 0);
		if(!data) {
			return TC_ERROR;
		}

		err = readData(fb, data, size, T_TRUE);
		if(err == TC_ERROR) {
			munmap(data, size);
			return TC_ERROR;
		}
		munmap(data, size);
	} 

	return TC_OK;
}

/* FOR BITMAPS */
int tc_resize_handle_bitmap(struct tc_fb *fb, thinc_header *h, 
	unsigned int *bitmap_size)
{
	tc_dbg_name(func)
	thinc_reszData d;
	int err;

	err = readData(fb, &d, T_RESZ_DATA_SIZE, T_TRUE);
	if(err == TC_ERROR || err == TC_CLOSED) {
		tclient_msg("Error: Couldn't read resized message header\n");
		return TC_ERROR;
	}
	d.fb_width = ntohs(d.fb_width);
	d.fb_height = ntohs(d.fb_height);
	d.size = ntohl(d.size);

	tclient_dbg(1, ("%s: got resized BITMAP update: w %u, h %u, size %u\n",
		func, d.fb_width, d.fb_height, d.size));

	/* check sizes match */
	if( 	(d.fb_width != fb->curr_width) || 
		(d.fb_height != fb->curr_height) ) {
		/* ignore. need to read the data tho */
		tclient_dbg(1, ("%s: ignoring resized update for old size: "
			"%dx%d\n", func, d.fb_width, d.fb_height));

		/* FIXME: made for RAW but useful anyways (since bitmaps are 
		 * not compressed) */
		err = tc_resize_handle_old_update(fb, &d, h);
		if(err == TC_ERROR) {
			tclient_msg("Error: Couldn't handle outdated resize "
				"update\n");
			return TC_ERROR;
		} else
			return TC_OK;
	}

	*bitmap_size = d.size;
	return TC_OK;
}

/**** Util Functions *****/
/* scale all rectangles in a region */
inline void tc_resize_scale_region(struct tc_fb *fb, Region reg)
{
	tc_region_resize(fb, reg);
}

/* scale a rectangle using current factors. note that the width and or
 * height may be 0 after scaling. it is the caller's responsability to
 * handle this appropiately */
inline void tc_resize_scale_rect(struct tc_fb *fb, XRectangle *rect)
{
	rect->x = tc_resize_scale_x(fb, rect->x);
        rect->y = tc_resize_scale_y(fb, rect->y);
	rect->width = rect->width*fb->resize->factor_x;
	rect->height = rect->height*fb->resize->factor_y;
}

inline void tc_resize_scale_box(struct tc_fb *fb, tc_xlib_box *box)
{
	box->x1 = tc_resize_scale_x(fb, box->x1);
        box->y1 = tc_resize_scale_y(fb, box->y1);
	box->x2 = box->x2 * fb->resize->factor_x;
	box->y2 = box->y2 * fb->resize->factor_y;
}

inline int tc_resize_scale_x(struct tc_fb *fb, int num)
{
	return num*fb->resize->factor_x;
}

inline int tc_resize_scale_y(struct tc_fb *fb, int num)
{
	return num*fb->resize->factor_y;
}

/* useful to send info to the server in unresized terms (e.g. mouse
 * location) */
inline int tc_resize_unscale_numx(struct tc_fb *fb, int num)
{
	return num/fb->resize->factor_x;
}

inline int tc_resize_unscale_numy(struct tc_fb *fb, int num)
{
	return num/fb->resize->factor_y;
}

/* video resizing functions. mostly just get the factor and pass it to
 * video handling functions */
/* IMPORTANT: Caller should check that we're actually doing video */
inline int tc_resize_video_all(struct tc_fb *fb)
{
	struct tc_resize *resize = fb->resize;
	tc_video_client_scale_all(fb, resize->factor_x, resize->factor_y);
	return TC_OK;
}

inline int tc_resize_video(struct tc_fb *fb, int id)
{
	struct tc_resize *resize = fb->resize;
	tc_video_client_scale(fb, id, resize->factor_x, resize->factor_y);
	return TC_OK;
}

inline int tc_resize_move_video_all(struct tc_fb *fb)
{
	struct tc_resize *resize = fb->resize;
	tc_video_client_move_all(fb, resize->factor_x, resize->factor_y);
	return TC_OK;
}

inline int tc_resize_move_video(struct tc_fb *fb, int id)
{
	struct tc_resize *resize = fb->resize;
	tc_video_client_move(fb, id, resize->factor_x, resize->factor_y);
	return TC_OK;
}

#if 0
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

int writeToFile(char *data, int width,
	int height, int pitch)
{
	static int count = 0;
	unsigned int a, i;
	char fn[255];
	int dump; /* fd */
	const char *base = "/tmp/raw/raw";
	register char *curr;
	const int mode = S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH;
	const int Bpp = 2;

	/* Write a header?? */
	const unsigned char write_header = T_TRUE;

	strncpy(fn, base, 12);
	sprintf(fn+12, "%d", count++);

	printf("%s: %d %d %d\n", fn, width, height, pitch);

	dump = creat(fn, mode);

	if(write_header) {
		a = width;
		a = htonl(a);
		write(dump, &a, 4);
		a = height;
		a = htonl(a);
		write(dump, &a, 4);
		a = Bpp;
		a = htonl(a);
		write(dump, &a, 4);
		a= 0;   /* pad */
		write(dump, &a, 4);
	}

	curr = data;
	for(i=0; i<height; ++i) {
		write(dump, curr, width*Bpp);
		curr += pitch;
	}

	close(dump);
	return T_OK;
}
#endif
