/* Copyright 2005, 2007 Ricardo Baratto */
/* Copyright 2005, 2007 Columbia University */
/* This file is part of THINC.

    THINC is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 2 of the License, or
    (at your option) any later version.

    THINC is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with THINC.  If not, see <http://www.gnu.org/licenses/>.
*/
/*
 * tc_sound.c
 *
 * Sound handling functions for THINC client. Makes use of the alsa library.
 *
 * Authors: 
 * Leo Kim
 * Ricardo Baratto
 * Network Computing Lab - Columbia University
 */

/* set tabstop 8 */

#include "thinc_client.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>	/* for strerror */
#include <errno.h>
#include <fcntl.h>	
#include <netinet/tcp.h>
#include <unistd.h>

/* local funcs */
static int tc_openSoundPlayback(struct tc_fb *fb);
static int tc_openSoundCapture(struct tc_fb *fb);
static int tc_closeSoundPlayback(struct tc_fb *fb);
static int tc_closeSoundCapture(struct tc_fb *fb);
static int tc_playSound(struct tc_fb *fb);
static int tc_playSoundNetwork(struct tc_fb *fb);
static int tc_soundCapture(struct tc_fb *fb);
static int tc_soundCaptureSendTCP(struct tc_fb *fb);
static int tc_soundCaptureSendUDP(struct tc_fb *fb);

static int tc_alsaPlay(struct tc_snd *snd, const char *buf, 
	unsigned long bytes);
static int tc_alsaRecover(struct tc_snd *snd, int err);
static int tc_alsaRecoverCap(struct tc_snd *snd, int err);

static int tc_soundReadUDP(struct tc_fb *fb);

/* Important: assummes res and fb are defined in current context */
#define alsa_call(call, msg) 						\
	res = call;							\
	if(res < 0) {							\
		tc_error_msg(msg ": %s\n", snd_strerror(res));	\
		return TC_ERROR;					\
	}

int tc_initSound(struct tc_fb *fb)
{
	struct tc_snd *snd = &fb->snd;

	snd->pcm = NULL;

	tc_stats_zero(fb->stats.soundcount);
 	snd->timestamp = 0;
 	snd->didSync = T_FALSE;
 	snd->instr_stamps.recvstamp = 0;
 	snd->instr_stamps.playstamp = 0;
	snd->usock = 0;
	snd->udp_data = NULL;

	return TC_OK;
}

/* This function connects to the tsd. Note that it is entirely possible that 
 * the local sound card is initialized but no network connection to tsd can be
 * made. 
 * fb->remote_addr has remote address and port contains the remote port that
 * we're supposed to connect to. if port is 0, then this port will be the
 * display port (as found in remote_addr) + TC_SND_PORT_INCR.
 */
int tc_doConnectTsd(struct tc_fb *fb, int port)
{
	struct tc_snd *snd;
	struct sockaddr_in *addr;
	int sock, one;
	const char *cert_path = "/etc/X11/thinc/certs/cacert.pem";
	thinc_sndInit si;

	snd = &fb->snd;
	snd->sock = -1;

	sock = socket(PF_INET, SOCK_STREAM, 0);
	if (sock == -1) {
		return TC_ERROR;
	} 

	memcpy(&snd->addr, &fb->remote_addr, sizeof(snd->addr));
	addr = &snd->addr;
	if(port) {
		addr->sin_port = htons(port);
	} else {
		addr->sin_port = fb->remote_addr.sin_port + 
			htons(TC_SND_PORT_INCR);
	}

	if(connect(sock, (struct sockaddr*)addr, sizeof(*addr)) == -1) {
		return TC_ERROR;
	}

	if(fb->use_ssl) {
		int err;
		snd->ssl = NULL;
again:
		err = thinc_ssl_connect(sock, cert_path, &snd->ssl);
		if(err == T_SSL_WOULDBLOCK) {
			goto again;
		} else if(err == T_SSL_CLOSED || err == T_SSL_ERROR) {
			tc_error_msg("Couldn't initialize secure "
				"connection with tsd\n");
			close(sock);
			return TC_ERROR;
		}

		tclient_msg("Happy: Got secure connection with tsd!\n");
	}

	if(fcntl(sock, F_SETFL, O_NONBLOCK) < 0) {
		tc_error_msg("couldn't set non-blocking flag: %s\n",
			strerror(errno));
		close(sock);
		return TC_ERROR;
	}

	/* disable nagle's algorithm. Non fatal if we can't do it */
	one = 1;
	if(setsockopt(sock, IPPROTO_TCP, TCP_NODELAY, (char*)&one,
		sizeof(one)) < 0) {
		tclient_msg("Warning: Couldn't disable nagle's algorithm. "
			"Your performance may suffer\n");
	}   

	snd->sock = sock;

	/* now read the init message */
	if(readSoundData(fb, &si, T_SND_INIT_SIZE, T_TRUE) == TC_ERROR) {
		tc_error_msg("Couldn't read sound initialization message\n");
		tc_closeConn(fb, T_TRUE);
		return TC_ERROR;
	}
	si.port = ntohs(si.port);
	if(si.port) {
		if(tc_initUDP(fb, snd, si.port) == TC_ERROR) {
			tc_error_msg("Couldn't initialize UDP\n");
			tc_closeConn(fb, T_TRUE);
			return TC_ERROR;
		}
		snd->udp_data = calloc(1, sizeof(*snd->udp_data));
		if(!snd->udp_data) {
			tc_error_msg("Out of memory: %s\n", strerror(errno));
			tc_closeConn(fb, T_TRUE);
			return TC_ERROR;
		}
		/* so we don't have to reset this everytime */
		snd->udp_data->header.type = T_SND_DATA;
	} else {
		snd->usock = 0;
	}

	return TC_OK;
}

int tc_handleSound(struct tc_fb *fb)
{
	thinc_header h;
	int res;
 	tc_dbg_name(func)

	if (tc_hasSoundData(fb)) {
		tclient_dbg(1, ("Has existing sound data.\n"));
		return tc_playSound(fb);
	}
 
 	/* If video sync'ed us, then just return. This prevents us from
 	 * blocking forever if video read all additional sound data while
 	 * sync'ing, and there is no more data to be read. */
 	if (fb->snd.didSync) {
 		/* reset the flag */
 		fb->snd.didSync = T_FALSE;
 		tclient_dbg(1, ("%s: video sync'ed with sound, not attempting "
 			"to read sound header.\n", func));
 		return TC_OK;
 	}

	res = tc_readSoundHeader(fb, &h);
	if (res == TC_ERROR) {
		tc_error_msg("Couldn't read sound message header.\n");
		return TC_ERROR;
	} else if (res == TC_CLOSED) {
		tc_error_msg("Connection closed while reading sound message "
			"header.\n");
		return TC_CLOSED;
	}

	/* now handle the message */
	switch (h.type) {
		case T_SND_DATA:
			res = tc_playSoundNetwork(fb);
			if (res == TC_ERROR) {
				tc_error_msg("Cannot play sound.\n");
				return TC_ERROR;
			}
			break;
		case T_SND_PLAYBACK_OPEN:
			res = tc_openSoundPlayback(fb);
			if (res == TC_ERROR) {
				tc_error_msg("Cannot open playback sound\n");
				return TC_ERROR;
			}
			break;
		case T_SND_PLAYBACK_END:
			res = tc_closeSoundPlayback(fb);
			if (res == TC_ERROR) {
				tc_error_msg("Cannot close playback sound.\n");
				return TC_ERROR;
			}
			break;
		case T_SND_CAPTURE_OPEN:
			res = tc_openSoundCapture(fb);
			if (res == TC_ERROR) {
				tc_error_msg("Cannot open capture sound\n");
				return TC_ERROR;
			}
			break;
		case T_SND_CAPTURE_END:
			res = tc_closeSoundCapture(fb);
			if (res == TC_ERROR) {
				tc_error_msg("Cannot close capture sound.\n");
				return TC_ERROR;
			}
			break;
		default:
			tc_error_msg("Got invalid sound message '%d'.\n", 
				h.type); 
			return TC_ERROR;
	}

	return res;
}

static int tc_openSoundPlayback(struct tc_fb *fb)
{
	tc_dbg_name(func)
	int res;
	struct tc_snd *snd = &fb->snd;
	thinc_sndOpen o;
	unsigned int stamp;
	snd_pcm_hw_params_t *hw_params;
#if 0
	snd_pcm_sw_params_t *sw_params;
	unsigned int samples_per_chunk;
#endif
	double ms_per_sample;

	/* read message */
	res = tc_readSoundMsgNoHdr(fb, &o, T_SND_OPEN_SIZE);

	gettime_us(stamp);
	tc_timestamp(("Got open for soundcard at local timestamp %u\n", stamp));

	if(res == TC_ERROR) {
		tc_error_msg("Couldn't read playback open message\n");
		return TC_ERROR;
	} else if(res == TC_CLOSED) { /* connection closed */ 
		tc_error_msg("Connection closed while reading "
			"playback open message\n");
		return TC_ERROR;                                                             
	}

	snd->pparams.bits = o.bits;
	snd->pparams.channels = o.channels;
	snd->pparams.rate = ntohs(o.rate);
#ifdef TC_USESYNC
	tc_resetVideoTimeStamp(fb);
#endif

	tclient_dbg(1, ("%s: Setting sound playback to %d bits, %d channels, "
		"%d rate, %d (%s) endian\n", func, snd->pparams.bits, 
		snd->pparams.channels, snd->pparams.rate, o.endian, 
		(o.endian == T_BIG) ? "big" : "little"));

#ifndef TC_DROPSOUND

	/* :HACK: Sometimes multiple calls to open the driver are made by an
	 * audio application. We try to catch these calls and make sure that
	 * only the last open is used. Otherwise, this thing craps out. */
	if(snd->pcm) {
		tclient_msg("Warning: Audio application on remote server is "
			"opening playback audio driver more than once.\n"); 
		snd_pcm_close(snd->pcm);
	}

	alsa_call(snd_pcm_open(&snd->pcm, "default", 
		SND_PCM_STREAM_PLAYBACK, SND_PCM_NONBLOCK),
		"cannot open audio device"
	);

	/* set parameters. calling snd_pcm_set_params doesn't seem to work so we
	 * have to go the long way :( */
	alsa_call(snd_pcm_hw_params_malloc(&hw_params),
		"cannot allocate hardware parameters"
	);
	alsa_call(snd_pcm_hw_params_any(snd->pcm, hw_params),
		"cannot initialize hardware parameters"
	);

	alsa_call(snd_pcm_hw_params_set_access(snd->pcm, hw_params,
		SND_PCM_ACCESS_RW_INTERLEAVED), "cannot set access"
	);
	switch(snd->pparams.bits) {
		case 8:
			alsa_call(snd_pcm_hw_params_set_format(snd->pcm,
				hw_params, SND_PCM_FORMAT_U8),
				"cannot set format SND_PCM_FORMAT_U8"
			);
			snd->pparams.sample_size = 
				snd_pcm_format_physical_width(
				SND_PCM_FORMAT_U8) >> 3;
			break;
		case 16:
			/* XXX: Assuming is signed bits as we don't send
			 * that info over the wire */
			if(o.endian == T_BIG) {
				alsa_call(snd_pcm_hw_params_set_format(
					snd->pcm,
					hw_params, SND_PCM_FORMAT_S16_BE),
					"cannot set format "
					"SND_PCM_FORMAT_S16BE"
				);
				snd->pparams.sample_size = 
					snd_pcm_format_physical_width(
					SND_PCM_FORMAT_S16_BE) >> 3;
			} else {
				alsa_call(snd_pcm_hw_params_set_format(
					snd->pcm,
					hw_params, SND_PCM_FORMAT_S16_LE),
					"cannot set format "
					"SND_PCM_FORMAT_S16LE"
				);
				snd->pparams.sample_size = 
					snd_pcm_format_physical_width(
					SND_PCM_FORMAT_S16_LE) >> 3;
			}
			break;
		default:
			tc_error_msg(
				"Don't know the format for %d bits and "
				"%s endian\n", snd->pparams.bits, 
				o.endian == T_BIG ? "big" : "little");
			return TC_ERROR;
	}
	/* alsa needs us to tell it how many samples, not bytes */
	snd->pparams.sample_size *= snd->pparams.channels;
	tclient_dbg(1, ("sample size %u\n", snd->pparams.sample_size));

	alsa_call(snd_pcm_hw_params_set_channels(snd->pcm, hw_params,
		snd->pparams.channels),
		"cannot set number of channels"
	);
	tclient_dbg(1, ("set channels %u\n", snd->pparams.channels));

	alsa_call(snd_pcm_hw_params_set_rate_near(snd->pcm, hw_params,
		&snd->pparams.rate, NULL),
		"cannot set rate"
	);
	tclient_dbg(1, ("set rate %u\n", snd->pparams.rate));

	/* set up buffering */
#if 0
	samples_per_chunk = snd->pparams.rate/TC_SND_CHUNKS_PER_SEC;
	snd->chunk_size = samples_per_chunk * snd->pparams.channels *
		(snd->pparams.bits >> 3);
	if(snd->chunk_size < 1024)
		snd->chunk_size = 1024;

	tclient_dbg(1, ("asking for %lu period size\n", snd->chunk_size));
	alsa_call(snd_pcm_hw_params_set_period_size_near(snd->pcm, 
		hw_params, &snd->chunk_size, NULL), "cannot set period size");

	tclient_dbg(1, ("got %lu period size\n", snd->chunk_size));

	snd->num_chunks = TC_SND_CHUNKS_IN_BUF;
	tclient_dbg(1, ("asking for %u periods\n", snd->num_chunks));
	alsa_call(snd_pcm_hw_params_set_periods_near(snd->pcm, hw_params,
		&snd->num_chunks, NULL), "cannot set number of periods");

	tclient_dbg(1, ("got %u periods\n", snd->num_chunks));

	/* all done, put the params into the pcm struct */
	alsa_call(snd_pcm_hw_params(snd->pcm, hw_params),
		"cannot associate parameters with pcm handle"
	);

	/* get buffer size and allocate a copy of it */
	alsa_call(snd_pcm_hw_params_get_buffer_size(hw_params,
		&snd->buf_size_s), "cannot get buffer size");
	snd->buf_size = snd->buf_size_s * snd->pparams.sample_size * 2;
	tclient_dbg(1, ("Using buffer size %lu, period_size %lu\n", 
		snd->buf_size, snd->chunk_size));

	/* XXX: i'm not sure yet if we need to save this but just in case */
	alsa_call(snd_pcm_hw_params_get_period_size(hw_params, &snd->chunk_size,
		NULL), "cannot get period size");
#else
	/* all done, put the params into the pcm struct */
	alsa_call(snd_pcm_hw_params(snd->pcm, hw_params),
		"cannot associate parameters with pcm handle"
	);

	/* compute how big the buffer must be to keep TC_SND_BUF_TIME ms
	 * of sound data */
	ms_per_sample = 1000.0/snd->pparams.rate;
	snd->threshold = (TC_SND_BUF_TIME/ms_per_sample) * 
		snd->pparams.sample_size;
#if 0
	snd->buf_size = snd->threshold * 1.3;
#else
	snd->buf_size = snd->threshold;
#endif
	if(snd->buf_size < TC_SND_MIN_BUF_SIZE) {
		tclient_msg("Warning: Using minimum buffer size %u instead of %lu\n",
			TC_SND_MIN_BUF_SIZE, snd->buf_size);
		snd->buf_size = TC_SND_MIN_BUF_SIZE;
	}
	tclient_msg("buffer size is %lu, threshold %lu\n", snd->buf_size,
		snd->threshold);
#endif
	
	/* don't need this scheisse anymore */
	snd_pcm_hw_params_free(hw_params);

	snd->play_buf = malloc(snd->buf_size);
	if(!snd->play_buf) {
		tc_error_msg("No memory for sound buffer\n");
		snd_pcm_close(snd->pcm);
		return TC_ERROR;
	}
	snd->play_end = snd->play_buf + snd->buf_size;
	snd->play_add = snd->play_play = snd->play_buf;

	/* now set up software params */
#if 0
	alsa_call(snd_pcm_sw_params_malloc(&sw_params),
		"cannot allocate software parameters"
	);

	/* get current sw parameters */
	alsa_call(snd_pcm_sw_params_current(snd->pcm, sw_params),
		"cannot get software parameters");

	/* idea here is to start playback only when the buffer is almost full.
	 * purpose being to make us resilient to small delays on the server
	 * side */
	alsa_call(snd_pcm_sw_params_set_start_threshold(snd->pcm, sw_params, 
		(snd->buf_size_s/snd->chunk_size)*snd->chunk_size), 
		"cannot set start threshold mode");

	/* allow playback when at least snd->chunk_size samples can be 
	 * processed */
	alsa_call(snd_pcm_sw_params_set_avail_min(snd->pcm, sw_params, 
		snd->chunk_size), 
		"cannot set available minimum for playback");

	/* align all transfers to 1 sample */
	alsa_call(snd_pcm_sw_params_set_xfer_align(snd->pcm, sw_params, 1),
		"cannot set transfer align");

	/* write the parameters */
	alsa_call(snd_pcm_sw_params(snd->pcm, sw_params), 
		"cannot set software parameters");

	snd_pcm_sw_params_free(sw_params);
#else
	snd->curr_size = 0;
#endif

	alsa_call(snd_pcm_prepare(snd->pcm), "cannot prepare audio card");

	/* finally get the sound card descriptor so we can poll on it
	 * later on (we only poll it if we have data to play) */
	/* in theory we 
	 * only have to care about 1, hopefully that's all we need, but
	 * just in case... */
	if(snd_pcm_poll_descriptors_count(snd->pcm) == 1) {
		struct pollfd f;
		alsa_call(snd_pcm_poll_descriptors(snd->pcm, &f, 1),
			"cannot get poll descriptors");
		snd->play_sock = f.fd;
		snd->play_events = f.events;
	} else {
		tclient_msg("BUG: There's more than one sound file descriptor "
			"to poll on for playback :( %d\n", 
			snd_pcm_poll_descriptors_count(snd->pcm)
		);
	}
#endif	/* !TC_DROPSOUND */

#if 0
	/* now compute the jitter buffer size. the idea here is:
	 * - figure out how many samples we need for each chunk
	 * - figure out how big each chunk is (using channel and bits)
	 * - figure out buffer size
	 */
	samples_per_chunk = snd->pparams.rate/TC_SND_CHUNKS_PER_SEC;
	snd->chunk_size = samples_per_chunk * snd->pparams.channels *
		(snd->pparams.bits >> 3);
	snd->buf_size = snd->chunk_size * TC_SND_CHUNKS_IN_BUF;

	snd->play_buf = malloc(snd->buf_size);
	if(!snd->play_buf) {
		tc_error_msg("No memory for sound buffer\n");
		return TC_ERROR;
	}
	snd->play_end = snd->play_buf + snd->buf_size;
	snd->play_add = snd->play_play = snd->play_buf;

	tclient_dbg(1, ("%s: rate %d bits %d channels %d\n"
		"jitter buffer: samples_per_chunk %u, "
		"chunk_size %u buf_size %u\n", func,
		snd->pparams.rate, snd->pparams.bits, snd->pparams.channels,
		samples_per_chunk, snd->chunk_size, snd->buf_size));
#endif

	/* if have UDP, need to start listening on the UDP socket */
	if(snd->usock) {
		add_fd(snd->usock, FD_SND_UDP, FD_READ, tc_soundReadUDP, NULL);
	}

	return TC_OK;
}

static int tc_playSoundNetwork(struct tc_fb *fb)
{
	tc_dbg_name(func)
	int res;
	struct tc_snd *snd = &fb->snd;
	thinc_sndData p;
 	unsigned int stamp;

	/* read message */
 	gettime_us(stamp);
	res = tc_readSoundMsgNoHdr(fb, &p, T_SND_DATA_SIZE);
	if(res == TC_ERROR) {
		tc_error_msg("Couldn't read T_SND_PLAY message\n");
		return TC_ERROR;
	} else if(res == TC_CLOSED) { /* connection closed */ 
		tc_error_msg("Connection closed while reading "
			"T_SND_PLAY message\n");
		return TC_ERROR;
	}

	/* drop the ID for now */
	p.size = ntohl(p.size);
	if(p.size > snd->buf_size) {
		tclient_msg("Warning: realloc'ing the buffer to %u bytes!\n",
			p.size);
		/* aiieee, so dirty, using pointers as deltas */
		snd->play_add = (char*) (snd->play_add - snd->play_buf);
		snd->play_play = (char*) (snd->play_play - snd->play_buf);
		snd->play_buf = realloc(snd->play_buf, p.size);
		if(!snd->play_buf) {
			tc_error_msg("Cannot grow sound buffer\n");
			return TC_ERROR;
		}
		snd->play_add = snd->play_buf + (int) snd->play_add;
		snd->play_play = snd->play_buf + (int) snd->play_play;
		snd->play_end = snd->play_buf + p.size;
		snd->buf_size = p.size;
	}
	snd->timestamp = ntohl(p.timestamp);
	snd->instr_stamps.recvstamp = stamp;

	tclient_dbg(1, ("%s: about to read %d bytes.\n", func, p.size));

	/* see if we need to wrap around the buffer end */
	if(snd->play_add + p.size > snd->play_end) {
		unsigned int space = snd->play_end - snd->play_add;
		tclient_dbg(1, ("%s: only have %u bytes left, have to wrap "
			"around\n", func, space));
 		res = readSoundData(fb, snd->play_add, space, T_TRUE);
		if(res == TC_CLOSED || res == TC_ERROR) {
			tc_error_msg("Can't read audio data\n");
			return TC_ERROR;
		}
		snd->play_add = snd->play_buf;
		space = p.size - space;
		res = readSoundData(fb, snd->play_add, space, T_TRUE);
		if(res == TC_CLOSED || res == TC_ERROR) {
			tc_error_msg("Can't read audio data\n");
			return TC_ERROR;
		}
		snd->play_add += space;
	} else {
		res = readSoundData(fb, snd->play_add, p.size, T_TRUE);
		if(res == TC_CLOSED || res == TC_ERROR) {
			tc_error_msg("Can't read audio data\n");
			return TC_ERROR;
		}
		snd->play_add += p.size;
		if(snd->play_add == snd->play_end)
			snd->play_add = snd->play_buf;
	}
	tclient_dbg(1, ("%s: readSoundData read %d bytes.\n", func, res));

	snd->curr_size += p.size;
	if(snd->curr_size < snd->threshold)
		return TC_OK;
	else {
		snd->threshold = 0;
		return tc_playSound(fb);
	}
}

static int tc_playSound(struct tc_fb *fb)
{
	int err;
	struct tc_snd *snd = &fb->snd;

#if 0	/* #ifdef TC_USESYNC TODO: Don't use just yet */
	res = tc_doSoundSync(fb, snd);
	if (res != TC_OK) {
		if (res > 0) {
			/* waitsync sound */
			if (tc_waitSyncSound(fb, snd, res) != TC_OK) {
				tclient_msg("Error: problem occurred while "
					"pausing/sync'ing sound.\n");
				return TC_ERROR;
			}
		} else {
			/* drop sound */
			tclient_dbg(1,("Dropping sound.\n"));
			snd->offset = 0;
			snd->remaining = 0;
			return TC_OK;
		}
	}
#endif

#ifndef TC_DROPSOUND
	if(snd->play_play < snd->play_add) {
		err = tc_alsaPlay(snd, snd->play_play, 
			snd->play_add - snd->play_play);
		if(err < 0) {
			tc_error_msg("cannot play audio data: %s\n", 
				snd_strerror(err));
			return TC_ERROR;
		}
		snd->play_play += err;
	} else {
		/* two shots */
		err = tc_alsaPlay(snd, snd->play_play, 
			snd->play_end - snd->play_play);
		if(err < 0) {
			tc_error_msg("cannot play audio data: %s\n", 
				snd_strerror(err));
			return TC_ERROR;
		}
		snd->play_play += err;
		/* paranoia */
		if(snd->play_play > snd->play_end) {
			tclient_msg("BUG: wrote sound data past the end "
				"of the buffer\n");
			return TC_ERROR;
		}
		/* check if played all the way to the end */
		if(snd->play_play != snd->play_end) {
			add_fd(snd->play_sock, FD_SND_PLAY,
				FD_READ | FD_WRITE,
				tc_soundCanPlay, NULL);
			return TC_OK;
		}
		/* rewind */
		snd->play_play = snd->play_buf;
		err = tc_alsaPlay(snd, snd->play_play, 
			snd->play_add - snd->play_play);
		if(err < 0) {
			tc_error_msg("cannot play audio data: %s\n", 
				snd_strerror(err));
			return TC_ERROR;
		}
		snd->play_play += err;
	}
	/* paranoia */
	if(snd->play_play > snd->play_add) {
		tclient_msg("BUG: wrote sound data past the head of the "
			"buffer\n");
		return TC_ERROR;
	} else if(snd->play_play != snd->play_add) {
		/* have data to play, let alsa wake us up once we can play
		 * again */
		add_fd(snd->play_sock, FD_SND_PLAY, 
			FD_READ | FD_WRITE,
			tc_soundCanPlay, NULL);
	} else {
		remove_fd(FD_SND_PLAY);
	}
#else	/* !TC_DROPSOUND */
	snd->play_play = snd->play_add;	/* just "consume" the data */
#endif	/* !TC_DROPSOUND */

	return TC_OK;
}

/* called when we get an event on the playback fd that it's ok to play 
 * in theory we should call snd_pcm_poll_descriptors_revents but
 * fuck it, just assume it means we can really play */
int tc_soundCanPlay(struct tc_fb *fb)
{
	/* chk if we have any data to play */
	if(fb->snd.play_play == fb->snd.play_add) {
		/* nothing to play :'( yes, i know is wherefore... */
		tclient_dbg(1, ("Warning: Could be playing data but we have "
			"none. tsd, tsd, where art thou tsd?\n"));
		return TC_OK;
	}

	return tc_playSound(fb);
}

/* returns how much data was succesfully played */
static int tc_alsaPlay(struct tc_snd *snd, const char *buf, 
	unsigned long bytes)
{
	snd_pcm_sframes_t err;
	snd_pcm_uframes_t size = bytes/snd->pparams.sample_size;
#if 0
	static unsigned long last = 0;
	unsigned long w;
#endif

	while (size > 0) {
		err = snd_pcm_writei(snd->pcm, buf, size);
		if (err == -EAGAIN) {
			/* can't play anymore for now */
			break;
		} else if (err < 0) {
			err = tc_alsaRecover(snd, err);
			if(err < 0)
				return err;
			continue;
		}
		if (err > 0) {
			buf += err * snd->pparams.sample_size;
			size -= err;
		}
	}

#if 0
	gettime_ms(w);
	fprintf(stderr, "[%lu] had to %lu, played %lu bytes\n", w-last, bytes, bytes - (size*snd->pparams.sample_size));
	last = w;
#endif

	/* return how much we played */
	return bytes - (size*snd->pparams.sample_size);
}

static int tc_alsaRecover(struct tc_snd *snd, int err)
{
	if (err == -EPIPE) {
		/* buffer underrun */
		err = snd_pcm_prepare(snd->pcm);
		if (err < 0) {
			tc_error_msg("Can't recover from underrun: %s\n",
				snd_strerror(err));
			return err;
		}
		tclient_msg("playback: recovered from buffer underrun\n");
		return TC_OK;
	} else if (err == -ESTRPIPE) {
		while ((err = snd_pcm_resume(snd->pcm)) == -EAGAIN)
			/* have to wait until suspend flag is released :/ */
			sleep(1);
		if (err < 0) {
			err = snd_pcm_prepare(snd->pcm);
			if (err < 0) {
				tc_error_msg("Can't recover from suspend: %s\n",
					snd_strerror(err));
				return err;
			}
		}
		return TC_OK;
	}
	/* can't recover from other errors :( */
	return err;
}

static int tc_closeSoundPlayback(struct tc_fb *fb)
{
	int res;
	thinc_sndClose c;
	/* read message */
	res = tc_readSoundMsgNoHdr(fb, &c, T_SND_CLOSE_SIZE);

	/* just discard the message for now. */
	tclient_msg("Info: Closing playback sound device.\n");

#ifdef TC_USESYNC
	/* for synchronization. if there is a stream, then clear its 
	 * timestamp. */
	/* :HACK: we assume that there's only one stream. */
	tc_resetVideoTimeStamp(fb);
#endif

	if(fb->snd.usock) {
		remove_fd(FD_SND_UDP);
	}
	if(fb->snd.pcm) {
		remove_fd(FD_SND_PLAY);
		snd_pcm_close(fb->snd.pcm);
	}
	fb->snd.pcm = NULL;

	if (fb->snd.play_buf) {
		free(fb->snd.play_buf);
		fb->snd.play_buf = fb->snd.play_end = fb->snd.play_add =
			fb->snd.play_play = NULL;
	}

	fb->snd.timestamp = 0;
	fb->snd.didSync = T_FALSE;
	fb->snd.instr_stamps.recvstamp = 0;
	fb->snd.instr_stamps.playstamp = 0;
	return TC_OK;
}

int tc_hasSoundData(struct tc_fb *fb) 
{
	/* TODO */
#if 0
	return (fb->use_sound && fb->snd.remaining > 0) ? T_TRUE: T_FALSE; 
#else
#warning "tc_hasSoundData is unimplemented"
	return T_FALSE;
#endif
}

void tc_closeConnTsd(struct tc_fb *fb)
{
	close(fb->snd.sock);
	if(fb->snd.usock) {
		close(fb->snd.usock);
		free(fb->snd.udp_data);
	}
	/* make eligible for reconnection. */
	fb->snd.sock = -1;
}

void tc_killSound(struct tc_fb *fb)
{
	if (!fb->use_sound) 
		return;
	tclient_msg("Info: Closing sound card resources.\n");
	if(fb->snd.pcm)
		snd_pcm_close(fb->snd.pcm);
}

static int tc_openSoundCapture(struct tc_fb *fb)
{
	tc_dbg_name(func)
	int res;
	struct tc_snd *snd = &fb->snd;
	thinc_sndOpen o;
	snd_pcm_hw_params_t *hw_params;

	/* read message */
	res = tc_readSoundMsgNoHdr(fb, &o, T_SND_OPEN_SIZE);
	if(res == TC_ERROR || res == TC_CLOSED) {
		tc_error_msg("Couldn't read capture open message\n");
		return TC_ERROR;
	}
	snd->cparams.bits = o.bits;
	snd->cparams.channels = o.channels;
	snd->cparams.rate = ntohs(o.rate);

	tclient_msg("%s: bits %d, channels %d, rate %d\n", __func__, 
		snd->cparams.bits, snd->cparams.channels, snd->cparams.rate);

	tclient_dbg(1, ("%s: Setting sound capture to %d bits, %d channels, "
		"%d rate, %d (%s) endian\n", func, snd->pparams.bits, 
		snd->pparams.channels, snd->pparams.rate, o.endian, 
		(o.endian == T_BIG) ? "big" : "little"));

	/* unsure if this happens but let's be on the safe side */
	if(snd->pcm_cap) {
		tclient_msg("Warning: Audio application on remote server is "
			"opening capture audio driver more than once.\n"); 
#if 0
		snd_pcm_close(snd->pcm_cap);
#else
		return TC_OK;
#endif
	}

	alsa_call(snd_pcm_open(&snd->pcm_cap, "default", 
		SND_PCM_STREAM_CAPTURE, SND_PCM_NONBLOCK),
		"cannot open audio device"
	);
	/* set parameters. calling snd_pcm_set_params doesn't seem to work so we
	 * have to go the long way :( */
	alsa_call(snd_pcm_hw_params_malloc(&hw_params),
		"cannot allocate hardware parameters"
	);
	alsa_call(snd_pcm_hw_params_any(snd->pcm_cap, hw_params),
		"cannot initialize hardware parameters"
	);

	alsa_call(snd_pcm_hw_params_set_access(snd->pcm_cap, hw_params,
		SND_PCM_ACCESS_RW_INTERLEAVED), "cannot set access"
	);
	switch(snd->cparams.bits) {
		case 8:
			alsa_call(snd_pcm_hw_params_set_format(snd->pcm_cap,
				hw_params, SND_PCM_FORMAT_U8),
				"cannot set format SND_PCM_FORMAT_U8"
			);
			snd->cparams.sample_size = 
				snd_pcm_format_physical_width(
				SND_PCM_FORMAT_U8) >> 3;
			break;
		case 16:
			/* XXX: Assuming is signed bits as we don't send
			 * that info over the wire */
			if(o.endian == T_BIG) {
				alsa_call(snd_pcm_hw_params_set_format(
					snd->pcm_cap,
					hw_params, SND_PCM_FORMAT_S16_BE),
					"cannot set format SND_PCM_FORMAT_S16BE"
				);
				snd->cparams.sample_size = 
					snd_pcm_format_physical_width(
					SND_PCM_FORMAT_S16_BE) >> 3;
			} else {
				alsa_call(snd_pcm_hw_params_set_format(
					snd->pcm_cap,
					hw_params, SND_PCM_FORMAT_S16_LE),
					"cannot set format SND_PCM_FORMAT_S16LE"
				);
				snd->cparams.sample_size = 
					snd_pcm_format_physical_width(
					SND_PCM_FORMAT_S16_LE) >> 3;
			}
			break;
		default:
			tc_error_msg(
				"Don't know the format for %d bits and "
				"%s endian\n", snd->cparams.bits, 
				o.endian == T_BIG ? "big" : "little");
			return TC_ERROR;
	}
	/* alsa needs us to tell it how many samples, not bytes */
	snd->cparams.sample_size *= snd->cparams.channels;
	tclient_dbg(1, ("sample size %u\n", snd->cparams.sample_size));

	alsa_call(snd_pcm_hw_params_set_channels(snd->pcm_cap, hw_params,
		snd->cparams.channels),
		"cannot set number of channels"
	);
	tclient_dbg(1, ("set channels %u\n", snd->cparams.channels));

	alsa_call(snd_pcm_hw_params_set_rate_near(snd->pcm_cap, hw_params,
		&snd->cparams.rate, NULL),
		"cannot set rate"
	);
	tclient_dbg(1, ("set rate %u\n", snd->cparams.rate));
	/* all done, put the params into the pcm struct */
	alsa_call(snd_pcm_hw_params(snd->pcm_cap, hw_params),
		"cannot associate parameters with pcm handle"
	);

	/* set up the buffer */
	snd->cap_bufsize = TC_SND_CAP_BUF;
	snd->cap_buf = malloc(snd->cap_bufsize);
	if(!snd->cap_buf) {
		tc_error_msg("No memory for sound capture buffer\n");
		snd_pcm_close(snd->pcm_cap);
		return TC_ERROR;
	}
	snd->cap_end = snd->cap_buf + snd->cap_bufsize;
	snd->cap_add = snd->cap_send = snd->cap_buf;

	alsa_call(snd_pcm_prepare(snd->pcm_cap), "cannot prepare audio card");
	/* finally get the sound card descriptor so we can poll on it
	 * later on. in theory we only have to care about 1, hopefully that's 
	 * all we need, but just in case... */
	if(snd_pcm_poll_descriptors_count(snd->pcm_cap) == 1) {
		struct pollfd f;
		alsa_call(snd_pcm_poll_descriptors(snd->pcm_cap, &f, 1),
			"cannot get poll descriptors");
		snd->cap_sock = f.fd;
		snd->cap_events = f.events;
	} else {
		tclient_msg("BUG: There's more than one sound file descriptor "
			"to poll on for capture :( %d\n", 
			snd_pcm_poll_descriptors_count(snd->pcm_cap)
		);
	}

	/* now add file descriptors so we know when we have data to capture */
	add_fd(snd->cap_sock, FD_SND_CAPTURE, FD_READ | FD_WRITE, 
		tc_soundCapture, NULL);

	tclient_dbg(1, ("%s: all set, trying to capture first bit of data\n",
		func));

	/* and try to capture some data */
	return tc_soundCapture(fb);
}

/* called when we get an event on the capture fd that we have data.
 * in theory we should call snd_pcm_poll_descriptors_revents but
 * fuck it, just assume it means we have some data */
static int tc_soundCapture(struct tc_fb *fb)
{
	tc_dbg_name(func)
	struct tc_snd *snd = &fb->snd;
	snd_pcm_sframes_t err;
	snd_pcm_uframes_t size;

#if 0
	do {
		if(snd->cap_end - snd->cap_add < TC_SND_CAP_SIZE) {
			size = snd->cap_end - snd->cap_add;
			size /= snd->cparams.sample_size;
		} else
			size = TC_SND_CAP_SIZE/snd->cparams.sample_size;

		tclient_dbg(1, ("%s: gonna capture me some frames (%lu)\n", 
			func, size));

	again:
		err = snd_pcm_readi(snd->pcm_cap, snd->cap_add, size);
		if(err == -EAGAIN) {
			tclient_dbg(1, ("%s: no more data available\n", func));
			break;
		} else if(err < 0) {
			err = tc_alsaRecoverCap(snd, err);
			if(err < 0)
				return TC_ERROR;
			goto again;
		}
		tclient_dbg(1, ("%s: read %ld frames\n", func, err));
		snd->cap_add += (err * snd->cparams.sample_size);
		if(snd->cap_add == snd->cap_end)
			snd->cap_add = snd->cap_buf;
	} while(1);
#else
	if(snd->cap_end - snd->cap_add < TC_SND_CAP_SIZE) {
		size = snd->cap_end - snd->cap_add;
		size /= snd->cparams.sample_size;
	} else
		size = TC_SND_CAP_SIZE/snd->cparams.sample_size;

	tclient_dbg(1, ("%s: gonna capture me some frames (%lu)\n", 
		func, size));

again:
	err = snd_pcm_readi(snd->pcm_cap, snd->cap_add, size);
	if(err == -EAGAIN) {
		tclient_dbg(1, ("%s: no more data available\n", func));
		return TC_OK;
	} else if(err < 0) {
		err = tc_alsaRecoverCap(snd, err);
		if(err < 0)
			return TC_ERROR;
		goto again;
	}
	tclient_dbg(1, ("%s: read %ld frames\n", func, err));
	snd->cap_add += (err * snd->cparams.sample_size);
	if(snd->cap_add == snd->cap_end)
		snd->cap_add = snd->cap_buf;
#endif

	/* now send it to tsd */
	if(snd->usock) {
		if(tc_soundCaptureSendUDP(fb) == TC_ERROR)
			return TC_ERROR;
	} else {
		if(tc_soundCaptureSendTCP(fb) == TC_ERROR)
			return TC_ERROR;
	}

	tclient_dbg(1, ("%s: finished\n", func));
	return TC_OK;
}

/* send captured data to tsd using TCP */
static int tc_soundCaptureSendTCP(struct tc_fb *fb)
{
	tc_dbg_name(func)
	struct tc_snd *snd = &fb->snd;
	int err, avail;
	thinc_sndData d;

	if(snd->cap_send < snd->cap_add) {
		avail = snd->cap_add - snd->cap_send;
		tclient_dbg(1, ("%s: sending %d bytes\n", func, avail));

		d.size = htonl(avail);
		d.timestamp = 0;
		d.id = 0;

		err = tc_writeSoundMsg(fb, T_SND_DATA, &d, T_SND_DATA_SIZE);
		if(err == TC_ERROR) {
			tc_error_msg("Couldn't send captured sound data\n");
			return TC_ERROR;
		}

		err = writeSoundData(fb, snd->cap_send, avail);
		if(err == TC_ERROR) {
			tc_error_msg("Couldn't send captured sound data\n");
			return TC_ERROR;
		}
	} else {
		d.size = snd->cap_end - snd->cap_send;
		d.size += snd->cap_add - snd->cap_buf;
		tclient_dbg(1, ("%s: sending %d bytes\n", func, d.size));

		d.size = htonl(d.size);
		d.timestamp = 0;
		d.id = 0;

		err = tc_writeSoundMsg(fb, T_SND_DATA, &d, T_SND_DATA_SIZE);
		if(err == TC_ERROR) {
			tc_error_msg("Couldn't send captured sound data\n");
			return TC_ERROR;
		}

		avail = snd->cap_end - snd->cap_send;
		tclient_dbg(1, ("%s: sending first %d bytes\n", func, avail));

		err = writeSoundData(fb, snd->cap_send, avail);
		if(err == TC_ERROR) {
			tc_error_msg("Couldn't send captured sound data\n");
			return TC_ERROR;
		}

		avail = snd->cap_add - snd->cap_buf;
		tclient_dbg(1, ("%s: sending remaining %d bytes\n", func, 
			avail));

		err = writeSoundData(fb, snd->cap_buf, avail);
		if(err == TC_ERROR) {
			tc_error_msg("Couldn't send captured sound data\n");
			return TC_ERROR;
		}
	}
	snd->cap_send = snd->cap_add;
	return TC_OK;
}

static int tc_closeSoundCapture(struct tc_fb *fb)
{
	int res;
	thinc_sndClose c;
	/* read message */
	res = tc_readSoundMsgNoHdr(fb, &c, T_SND_CLOSE_SIZE);
	if(res == TC_ERROR || res == TC_CLOSED) {
		tc_error_msg("Couldn't read sound capture close message\n");
		return TC_ERROR;
	}

	/* just discard the message for now. */
	tclient_msg("Info: Closing capture sound device.\n");

	if(fb->snd.pcm_cap) {
		remove_fd(FD_SND_CAPTURE);
		snd_pcm_close(fb->snd.pcm_cap);
	}
	fb->snd.pcm_cap = NULL;

	if (fb->snd.cap_buf) {
		free(fb->snd.cap_buf);
		fb->snd.cap_buf = fb->snd.cap_end = fb->snd.cap_add =
			fb->snd.cap_send = NULL;
	}

	return TC_OK;
}

static int tc_alsaRecoverCap(struct tc_snd *snd, int err)
{
	if (err == -EPIPE) {
		/* buffer overrun */
		err = snd_pcm_prepare(snd->pcm_cap);
		if (err < 0) {
			tc_error_msg("Can't recover from overrun: %s\n",
				snd_strerror(err));
			return err;
		}
		tclient_msg("capture: recovered from buffer overrun\n");
		return TC_OK;
	} else if (err == -ESTRPIPE) {
		while ((err = snd_pcm_resume(snd->pcm_cap)) == -EAGAIN)
			/* have to wait until suspend flag is released :/ */
			sleep(1);
		if (err < 0) {
			err = snd_pcm_prepare(snd->pcm_cap);
			if (err < 0) {
				tc_error_msg("Can't recover from suspend: %s\n",
					snd_strerror(err));
				return err;
			}
		}
		return TC_OK;
	}
	/* can't recover from other errors :( */
	return err;
}

/* UDP */
#if 0
#include <sys/ioctl.h>
#endif

static int tc_soundReadUDP(struct tc_fb *fb)
{
	tc_dbg_name(func)
	struct tc_snd *snd = &fb->snd;
	thinc_sndUDPData *d;
	int err;
 	unsigned int stamp;

 	gettime_us(stamp);

	/* TODO: Can we use this somehow? */
#if 0
	if(ioctl(snd->usock, FIONREAD, &err) == -1) {
		tc_error_msg("Couldn't get size of next datagram: %s\n", strerror(errno));
	} else {
		tclient_msg("next datagram has size %d\n", err);
	}
#endif

	d = snd->udp_data;
	err = read(snd->usock, d, T_SND_UDP_DATA_SIZE);
	if(err <= 0) {
		tc_error_msg("Couldn't read UDP audio data: %s\n", 
			strerror(errno));
		return TC_ERROR;
	}

	d->size = ntohl(d->size);
	snd->timestamp = ntohl(d->timestamp);
	snd->instr_stamps.recvstamp = stamp;
	tclient_dbg(1, ("%s: read %d bytes.\n", func, d->size));

	/* see if we need to wrap around the buffer end */
	if(snd->play_add + d->size > snd->play_end) {
		unsigned int space = snd->play_end - snd->play_add;
		char *src;
		tclient_dbg(1, ("%s: only have %u bytes left, have to wrap "
			"around\n", func, space));
		src = d->data;
		memcpy(snd->play_add, src, space);
		snd->play_add = snd->play_buf;
		src += space;
		space = d->size - space;
		memcpy(snd->play_add, src, space);
		snd->play_add += space;
	} else {
		memcpy(snd->play_add, d->data, d->size);
		snd->play_add += d->size;
		if(snd->play_add == snd->play_end)
			snd->play_add = snd->play_buf;
	}

	snd->curr_size += d->size;
	if(snd->curr_size < snd->threshold)
		return TC_OK;
	else {
		snd->threshold = 0;
		return tc_playSound(fb);
	}

	return TC_OK;
}

/* send captured data to tsd using UDP */
#if 0
static int tc_soundCaptureSendUDP(struct tc_fb *fb)
{
	tc_dbg_name(func)
	struct tc_snd *snd = &fb->snd;
	int err, avail;
	/* d->header is initialized when we first allocate it */
	thinc_sndUDPData *d = snd->udp_data;

	/* see if it's all contiguous or need to wrap around */
	if(snd->cap_send < snd->cap_add) {
		/* all contiguous */
		avail = snd->cap_add - snd->cap_send;
		tclient_dbg(1, ("%s: sending %d bytes\n", func, avail));

		d->size = htonl(avail);
		d->timestamp = 0;
		d->id = 0;
		memcpy(d->data, snd->cap_send, avail);

		err = write(snd->usock, d, T_SND_UDP_DATA_SIZE);
		if(err == -1) {
			tc_error_msg("Couldn't send captured sound data\n");
			return TC_ERROR;
		}
	} else {
		char *to;

		d->size = snd->cap_end - snd->cap_send;
		d->size += snd->cap_add - snd->cap_buf;
		tclient_dbg(1, ("%s: sending %d bytes\n", func, d.size));

		d->size = htonl(d->size);
		d->timestamp = 0;
		d->id = 0;

		avail = snd->cap_end - snd->cap_send;
		memcpy(d->data, snd->cap_send, avail);
		to = d->data + avail;
		avail = snd->cap_add - snd->cap_buf;
		memcpy(to, snd->cap_buf, avail);

		err = write(snd->usock, d, T_SND_UDP_DATA_SIZE);
		if(err == TC_ERROR) {
			tc_error_msg("Couldn't send captured sound data\n");
			return TC_ERROR;
		}
	}
	snd->cap_send = snd->cap_add;
	return TC_OK;
}
#else

#define _udp_send_it(len) do { \
	err = write(snd->usock, d, len);			\
	if(err == -1) {							\
		if(errno != EAGAIN && errno != EWOULDBLOCK && errno != EINTR) {			\
			tc_error_msg("Couldn't send captured sound data %s\n", strerror(errno));	\
			return TC_ERROR;					\
		}							\
	}								\
} while(0)

static int tc_soundCaptureSendUDP(struct tc_fb *fb)
{
	struct tc_snd *snd = &fb->snd;
	int err, avail;
	char *from;
	/* d->header is initialized when we first allocate it */
	thinc_sndUDPData *d = snd->udp_data;

	d->id = 0;
	d->timestamp = 0;
	d->size = htonl(T_SND_UDP_PKT_SIZE);

	from = snd->cap_send;

	if(snd->cap_send < snd->cap_add) {
		avail = snd->cap_add - snd->cap_send;
		while(avail >= T_SND_UDP_PKT_SIZE) {
			memcpy(d->data, from, T_SND_UDP_PKT_SIZE);

			_udp_send_it(T_SND_UDP_DATA_SIZE);
			avail -= T_SND_UDP_PKT_SIZE;
			from += T_SND_UDP_PKT_SIZE;
		}
		if(avail) {
			d->size = htonl(avail);
			memcpy(d->data, from, avail);
			_udp_send_it(T_SND_UDP_DATA_HSIZE + avail);
		}
	} else {
		/* send first chunk (from cap_send to the end of the 
		 * buffer) */
		char *to;
		int left;
		avail = snd->cap_end - snd->cap_send;
		while(avail >= T_SND_UDP_PKT_SIZE) {
			memcpy(d->data, from, T_SND_UDP_PKT_SIZE);

			_udp_send_it(T_SND_UDP_DATA_SIZE);
			avail -= T_SND_UDP_PKT_SIZE;
			from += T_SND_UDP_PKT_SIZE;
		}

		/* horrendous hack to deal with some data leftover from first
		 * chunk, and unknown amount of data remaining in the
		 * second chunk */
		if(avail) {
			memcpy(d->data, from, avail);
			to = d->data + avail;
			left = T_SND_UDP_PKT_SIZE - avail;
		} else {
			left = T_SND_UDP_PKT_SIZE;
			to = d->data;
		}
		from = snd->cap_buf;
		avail = snd->cap_add - snd->cap_buf;
		if(left >= avail) {
			/* all done */
			int len = to - d->data + avail;
			memcpy(to, from, avail);
			d->size = htonl(len);
			_udp_send_it(len + T_SND_UDP_DATA_HSIZE);
		} else {
			/* still have some data, send full transition packet
			 * then go back to original loop */
			memcpy(to, from, left);
			_udp_send_it(T_SND_UDP_DATA_SIZE);

			/* back to normal operation */
			avail -= left;
			from += left;
			while(avail >= T_SND_UDP_PKT_SIZE) {
				memcpy(d->data, from, T_SND_UDP_PKT_SIZE);

				_udp_send_it(T_SND_UDP_DATA_SIZE);
				avail -= T_SND_UDP_PKT_SIZE;
				from += T_SND_UDP_PKT_SIZE;
			}
			if(avail) {
				d->size = htonl(avail);
				memcpy(d->data, from, avail);
				_udp_send_it(T_SND_UDP_DATA_HSIZE + avail);
			}
		}
	}
	snd->cap_send = snd->cap_add;
	return TC_OK;
}
#endif
