/* Copyright 2002, 2003, 2004, 2005, 2007 Ricardo Baratto */
/* Copyright 2002, 2003, 2004, 2005, 2007 Columbia University */
/* This file is part of THINC.

    THINC is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 2 of the License, or
    (at your option) any later version.

    THINC is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with THINC.  If not, see <http://www.gnu.org/licenses/>.
*/
/* 
 * Util functions for thinc client
 *
 * Author:
 * Ricardo Baratto
 * Network Computing Lab - Columbia University
 *
 * Copyright 2004
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <errno.h>
#include <netdb.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <stdarg.h>
#include <fcntl.h>
#include <netinet/in.h>
#include <netinet/tcp.h>
#include <time.h>
#include <sys/poll.h>

#include "thinc_client.h"

/* XXX: Nasty hack */
const char *cert_path = "/etc/X11/thinc/certs/cacert.pem";

//#define BUF_SIZE 655360
#define BUF_SIZE 65536

/* buffering inspired on vnc. Make it global so that we can test wether
 * or not it's empty */
static char readBuffer[BUF_SIZE], *rBufStart;
static int rBufSize=0;

static int readDataSocket(union tc_io io, void *data, int count);
static int writeDataSocket(union tc_io io, const void *buf, int count);
static int readDataSecure(union tc_io io, void *data, int count);
static int writeDataSecure(union tc_io io, const void *buf, int count);
static char* byte_to_string(int bytes);
static int tc_setNonBlocking(int sock);
static int tc_udpSocket(int port);

static int tc_readDataFromNetwork(struct tc_fb *fb, 
	char *buffer, int need, int howmuch, int block);

inline int buffer_empty() 
{
	return (rBufSize == 0);
}

inline int buffer_notempty(struct tc_fb *fb) 
{
	return (rBufSize != 0);
}

inline int buffer_size() 
{
	return rBufSize;
}

void initIO(struct tc_fb *fb)
{
	/* initially we don't know if we'll use ssl or not. When the handshake
	 * gets underway this will get reassigned as needed */
	fb->read = readDataSocket;
	fb->write = writeDataSocket;

	fb->io.sock = fb->sock;
}

void setSecureIO(struct tc_fb *fb)
{
	fb->read = readDataSecure;
	fb->write = writeDataSecure;
	fb->io.ssl = fb->ssl;
}

int doConnect(struct tc_fb *fb, const char *host, const char *port)
{
	int sock, p, one;
	struct sockaddr_in *addr;
	struct hostent *h;
	char *end;

	sock = socket(PF_INET, SOCK_STREAM, 0);
	if(sock == -1) {
		tc_error_msg("Couldn't open socket connection: %s\n", 
			strerror(errno));
		return TC_ERROR;
	}

	addr = &fb->remote_addr;

	memset(addr, 0, sizeof(*addr));
	addr->sin_family = AF_INET;

	/* host */
	h = gethostbyname(host);
	if(!h) {
		tc_error_msg("Couldn't resolve destination hostname %s\n",
			host);
		return TC_ERROR;
	}
	memmove(&addr->sin_addr, h->h_addr, sizeof(struct in_addr));

	/* port */
	p = (int) strtol(port, &end, 0);
	if(*end == '\0') /* numeric passed, we're all set */
		addr->sin_port = htons(p);
	else {
		struct servent *s = getservbyname(port, "tcp");
		if(!s) { /* can't find it.. oh, well */
			tc_error_msg("Couldn't resolve destination "
				"port %s: %s\n", port, strerror(errno));
			return TC_ERROR;
		}
		addr->sin_port = s->s_port;
	}

	/* see if we should set tcp window size */
	/* TODO: Should also set RFC1323 options? */
	if(fb->tcp_window_size) {
		/* we recheck after connect to see what it got set to 
		 * (from iperf source) */
		if(setsockopt(sock, SOL_SOCKET, SO_RCVBUF, 
			&fb->tcp_window_size, sizeof(fb->tcp_window_size)))
		{
			tclient_msg("Warning: Couldn't set tcp window "
				"size: %s\n", strerror(errno));
			fb->tcp_window_size = 0;
		}
	}

	/* connect */
	if(connect(sock, (struct sockaddr*)addr, sizeof(*addr)) == -1) {
		tc_error_msg("Couldn't connect to server: %s\n",
			strerror(errno));
		return TC_ERROR;
	}

	/* make it non-blocking. After doing SSL connection so we don't need
	 * to worry too much */
	if(tc_setNonBlocking(sock) == TC_ERROR) {
		/* TODO: Call SSL shutdown function */
		tc_error_msg("Couldn't set non-blocking flag: %s\n",
			strerror(errno));
		close(sock);
		return TC_ERROR;
	}

	/* disable nagle's algorithm. Non fatal if we can't do it */
	one = 1;
	if(setsockopt(sock, IPPROTO_TCP, TCP_NODELAY, (char*)&one, 
		sizeof(one)) < 0) {
		tclient_msg("Warning: Couldn't disable nagle's algorithm. "
			"Your performance may suffer\n");
	}

	if(fb->tcp_window_size) {
		/* we recheck after connect to see what it got set to 
		 * (from iperf source) */
		unsigned int window, len;
		len = sizeof(window);
		if(getsockopt(sock, SOL_SOCKET, SO_RCVBUF, 
			&window, &len))
		{
			tclient_msg("Warning: Couldn't get TCP window "
				"size: %s\n", strerror(errno));
		} else {
			tclient_msg("Info: TCP Window Size: %s", 
				byte_to_string(window));
			if(window != fb->tcp_window_size) {
				tclient_msg(" (Warning: You requested %s)\n",
					byte_to_string(fb->tcp_window_size));
				fb->tcp_window_size = window;
			} else
				tclient_msg("\n");
		}
	}

	fb->sock = sock;
	return sock;
}

/**************** UTIL FUNCTIONS *******************/
/* If block is T_TRUE this function won't return until count bytes are
 * returned. Otherwise, it will return as some data is read, and return
 * how many bytes it returned. 
 * In the blocking case, it won't return until count bytes have been read or 
 * something bad happens in which case we return TC_ERROR for error, or 
 * TC_CLOSED if the connection was closed. On success we return how many
 * bytes were read.
 * Finally, if block is false, and no data at all could be read, 
 * TC_WOULDBLOCK is returned.
 */     

#define make_buffer_emtpy(buffer, start, size) \
	start = buffer; \
size = 0;

int readData(struct tc_fb *fb, void *data, int count, int block)
{       
	int res, need, got;
	char *curr = (char *) data;
	tc_dbg_name(func)
	int read_to_buffer, howmuch;
	char *where;

	instr_vars(sread, eread, sany, eany);

	instr_start(sread);
	time_add_extra("READ", TI_START);

	tclient_dbg(3, ("%s: reading %d bytes\n", func, count));

	/* chk if we have enough buffered */
	if(count <= rBufSize) { 
		instr_start(sany); 
		memcpy(data, rBufStart, count);
		instr_end(fb, sany, eany, read_memcpy);
		rBufSize -= count;
		rBufStart += count;
		tclient_dbg(3, ("%s: returned %d, still have %d buffered\n",
			func, count, rBufSize));
		tc_pass_or_die(rBufStart <= readBuffer+BUF_SIZE,
			"BUG: %s:%d:%s(): just overflowed our read buffer\n",
			__FILE__, __LINE__, func);
		instr_end(fb, sread, eread, read);
		time_add_extra("READ", TI_END);
		return count;
	}

	/* ok, not enough, copy what we have and get the rest. curr is pointer
	 * into data where we put whatever we get */
	tclient_dbg(3, ("%s: didn't have enough (only %d), need to read "
		"%d more\n", func, rBufSize, count-rBufSize));
	instr_start(sany);
	memcpy(data, rBufStart, rBufSize);
	instr_end(fb, sany, eany, read_memcpy);
	curr = data + rBufSize;
	need = count - rBufSize;
	/* for now, have at least one buffer worth of data. let's see how
	 * much else we can get */
	got = rBufSize;
	make_buffer_emtpy(readBuffer, rBufStart, rBufSize);

	tc_pass_or_die(curr < ((char*)data)+count,
		"BUG:%s:%d:%s(): Just overflowed caller's pointer\n",
		__FILE__, __LINE__, func);
	tc_pass_or_die(need > 0,
		"BUG: %s:%d:%s(): Why are we trying to read data when we "
		"already have all we need?\n",
		__FILE__, __LINE__, func);


 	/* howmuch is what we want. need is what we really need to get it
 	 * over with. in the case where we're reading directly to the 
 	 * caller's buffer, these two are equal */
	if(need <= BUF_SIZE) {
		read_to_buffer = T_TRUE;
		howmuch = BUF_SIZE;
		where = rBufStart;
	} else {
		read_to_buffer = T_FALSE;
		howmuch = need;
		where = curr;
	}

	res = tc_readDataFromNetwork(fb, where, need, howmuch,
		block);
	if(res == TC_ERROR || res == TC_CLOSED) {
		if(errno) {
			tclient_dbg(1, ("%s: An error occurred while "
				"reading data: %s\n", func, strerror(errno)));
		} else {
			tclient_dbg(1, ("%s: An error occurred while "
				"reading data\n", func));
		}
		instr_end(fb, sread, eread, read);
		time_add_extra("READ", TI_END);
		return TC_ERROR;
	} else if(res == TC_WOULDBLOCK) {
		/* agh, didn't get anything (block is false) */
		if(got) {
			/* at least have something from before */
			tclient_dbg(3, ("%s: couldn't read from the network "
				"but had %d bytes buffered\n", func, got));
			instr_end(fb, sread, eread, read);
			time_add_extra("READ", TI_END);
			return got;
		} else {
			/* got nothing at all. tell the caller about this */
			tclient_dbg(3, ("%s: Couldn't read any data at all\n",
				func));
			instr_end(fb, sread, eread, read);
			time_add_extra("READ", TI_END);
			return TC_WOULDBLOCK;
		}
	} else {
		tclient_dbg(3, ("%s: read %d bytes\n", func, res));
		/* all good */
		if(read_to_buffer) {
			/* copy to caller, and leave the rest in the buffer */
			if(need <= res) {
				tclient_dbg(3, ("%s: reading to buffer, "
					"and got enough (%d bytes) to make "
					"caller happy\n", func, res));
				/* got enough for caller */
				instr_start(sany);
				memcpy(curr, readBuffer, need);
				instr_end(fb, sany, eany, read_memcpy);
				rBufSize = res - need;
				rBufStart = readBuffer + need;
				got += need;
			} else {
				tclient_dbg(3, ("%s: reading to buffer, "
					"but couldn't read enough. got only "
					"%d bytes, but needed %d\n", func,
					res, need));
				/* didn't read enough (block is T_FALSE) */
				instr_start(sany);
				memcpy(curr, readBuffer, res);
				instr_end(fb, sany, eany, read_memcpy);
				make_buffer_emtpy(readBuffer, rBufStart,
					rBufSize);
				got += res;
			}
		} else {
			got += res;
		}
		tclient_dbg(3, ("%s: finished, have %d buffered\n", func,
			rBufSize));

		instr_end(fb, sread, eread, read);
		time_add_extra("READ", TI_END);
		return got;
	}
}

/* reads enough data to fill the buffer. called when what we need to
 * read from the socket is less than the buffer size.
 * need is what we actually need, howmuch is what we would like to
 * read */
static int tc_readDataFromNetwork(struct tc_fb *fb, 
	char *buffer, int need, int howmuch, int block)
{   
	int have, done, pollEv, res;
	tc_dbg_name(func)
	instr_vars(sany, eany);

	tclient_dbg(3, ("%s: trying to read %d bytes, but only need %d "
		"bytes\n", func, howmuch, need));
	done = T_FALSE;
	have = 0;

	while(!done) {
		time_add_extra("READ", TI_NET_START);
		instr_start(sany);
		res = fb->read(fb->io, buffer, howmuch-have);
		instr_end(fb, sany, eany, io_read);
		time_add_extra("READ", TI_NET_END);
		if(res == TC_ERROR || res == TC_CLOSED) {
			return res;
		} else if(res == TC_WOULDBLOCK) {
			tclient_dbg(3, ("%s: would block...\n", func));
			if(!block) {
				tclient_dbg(3, ("%s: was going to block "
					"but caller doesn't want that "
					"so bailing out. right now "
					"have %d bytes\n", func, have));
				/* if caller doesn't want us blocking,
				 * return now */
				done = T_TRUE;
				if(!have)
					/* tell caller we didn't read
					 * anything */
					have = TC_WOULDBLOCK;
				continue;
			}
			/* caller is ok with us blocking so block waiting 
			 * for more... we can block safely because we
			 * know we haven't read enough data yet */
			pollEv = POLLIN;
			instr_start(sany);
			res = doPoll(fb->sock, pollEv, NULL);
			instr_end(fb, sany, eany, read_poll);
			if(res == TC_ERROR)
				return TC_ERROR;
			else
				continue;
		}
		/* good, good, got something */
		have += res;
		tclient_dbg(3, ("%s: have already %d\n", func, have));
		if(have >= need) { /* we're done */
			done = T_TRUE;
			continue;
		} else { /* ok, update everything and keep reading */
			buffer += res;
		}
	}

	tclient_dbg(3, ("%s: finished, read %d bytes\n", func, have));

	return have;
}

/* This function is used purely for sound data. If we wish this to block, we do
 * not poll for data, but return TC_WOULDBLOCK instead.
 */
int readSoundData(struct tc_fb *fb, void *data, int count, int block)
{
	int res, need, have;
	have = 0;
	need = count;

	tc_dbg_name(func)
	union tc_io io;

	struct tc_snd *snd = &fb->snd;

	tclient_dbg(2, ("%s: reading %d bytes\n", func, count));

	if(fb->use_ssl)
		io.ssl = snd->ssl;
	else
		io.sock = snd->sock;

try:
	res = fb->read(io, data + have, need);
	if(res == TC_CLOSED) {
		/* connection closed, somebody screwed us 
		 * :( */
		tclient_dbg(3, ("%s: connection was closed "
			"while reading from tsd.\n", func));
		return TC_CLOSED;
	} else if(res == TC_ERROR) {
		goto error;
	} else if(res == TC_WOULDBLOCK) {
		if (!block)
			return TC_WOULDBLOCK;
		res = doPoll(snd->sock, POLLIN, NULL);
		if (res == TC_ERROR)
			return TC_ERROR;
		else
			goto try;
	}

	have += res;
	need -= res;
	/* :HACK: Need to rewrite this networking crap!!!!!! */
	if (have < count) {
		res = doPoll(snd->sock, POLLIN, NULL);
		if (res == TC_ERROR)
			return TC_ERROR;
		else
			goto try;
	}

	/* note the important difference here -- we return res instead of 
	 * TC_OK */
	return res;
error:
	if(errno) {
		tclient_dbg(3, ("%s: An error occurred while reading data: "
			"%s\n", func, strerror(errno)));
	} else {
		tclient_dbg(3, ("%s: An error occurred while reading data\n",
			func));
	}
	return TC_ERROR;
}
/* Virtually equivalent to writeData, except it writes to the sound socket. 
 * Will go away when we have a centralized manager daemon. */
int writeSoundData(struct tc_fb *fb, const void *buf, int count)
{   
	struct tc_snd *snd = &fb->snd;
	int left = count, wrote = 0;
	const char *curr = (char *) buf;
	union tc_io io;
	instr_vars(sio_write, eio_write, swrite, ewrite);

	instr_start(swrite);

	if(fb->use_ssl)
		io.ssl = snd->ssl;
	else
		io.sock = snd->sock;

	while (left > 0) {
		instr_start(sio_write);
		wrote = fb->write(io, curr, left);
		instr_end(fb, sio_write, eio_write, io_write);
		if(wrote == TC_WOULDBLOCK) {
			fd_set fds;
			FD_ZERO (&fds);
			FD_SET (snd->sock, &fds);
			wrote = select(snd->sock + 1, NULL, &fds, NULL, NULL);
			if (wrote == -1) {
				instr_end(fb, swrite, ewrite, write);
				return TC_ERROR;
			}
			continue;
		} else if(wrote == TC_ERROR) {
			tclient_msg("Error writing data: %s\n",
				strerror(errno));
			instr_end(fb, swrite, ewrite, write);
			return TC_ERROR;
		}
		left -= wrote;
		curr += wrote;
	}
	instr_end(fb, swrite, ewrite, write);
	return TC_OK;
}

static int readDataSocket(union tc_io io, void *data, int count)
{
	int res;
	int sock = io.sock;
try:
	res = read(sock, data, count);
	if(res == -1) {
		if(errno == EINTR)
			goto try;
		else if(errno == EWOULDBLOCK || errno == EAGAIN)
			return TC_WOULDBLOCK;
		else /* out out out */
			return TC_ERROR;
	} else if(res == 0) {
		return TC_CLOSED;
	}

	return res;
}

static int readDataSecure(union tc_io io, void *data, int count)
{
	SSL *ssl = io.ssl;
	int err  = thinc_ssl_read(ssl, data, count);
	if(err == T_SSL_CLOSED) 
		return TC_CLOSED;
	else if(err == T_SSL_WOULDBLOCK)
		return TC_WOULDBLOCK;
	else if(err == T_SSL_ERROR)
		return TC_ERROR;
	else
		return err;
}

/* writes everything. May block if not possible to write everything now.
 * Returns TC_ERROR on error, TC_OK on success */
int writeData(struct tc_fb *fb, const void *buf, int count)
{   
	int left = count, wrote = 0;
	const char *curr = (char *) buf;
	instr_vars(sio_write, eio_write, swrite, ewrite);

	instr_start(swrite);

	while (left > 0) {
		instr_start(sio_write);
		wrote = fb->write(fb->io, curr, left);
		instr_end(fb, sio_write, eio_write, io_write);
		if(wrote == TC_WOULDBLOCK) {
			fd_set fds;
			FD_ZERO (&fds);
			FD_SET (fb->sock, &fds);
			wrote = select(fb->sock + 1, NULL, &fds, NULL, NULL);
			if (wrote == -1) {
				instr_end(fb, swrite, ewrite, write);
				return TC_ERROR;
			}
			continue;
		} else if(wrote == TC_ERROR) {
			tclient_msg("Error writing data: %s\n",
				strerror(errno));
			instr_end(fb, swrite, ewrite, write);
			return TC_ERROR;
		}
		left -= wrote;
		curr += wrote;
	}
	instr_end(fb, swrite, ewrite, write);
	return TC_OK;
}

static int writeDataSocket(union tc_io io, const void *buf, int count)
{
	int wrote;
	int sock = io.sock;
try:
	wrote = send (sock, buf, count, 0);
	if (wrote <= 0) {
		if (errno == EINTR)
			goto try;
		else if (errno == EWOULDBLOCK || errno == EAGAIN) {
			return TC_WOULDBLOCK;
		} else {
			return TC_ERROR;
		}
	}
	return wrote;
}

static int writeDataSecure(union tc_io io, const void *buf, int count)
{
	int wrote;
	SSL *ssl = io.ssl;
	wrote = thinc_ssl_write(ssl, buf, count);
	if(wrote == T_SSL_CLOSED) 
		return TC_CLOSED;
	else if(wrote == T_SSL_WOULDBLOCK)
		return TC_WOULDBLOCK;
	else if(wrote == T_SSL_ERROR)
		return TC_ERROR;
	else
		return wrote;
}

/* poll() indefinitely the socket waiting for "events" to happen. Returns 
 * TC_ERROR or TC_OK. If revents is not NULL, it's filled with the revents 
 * returned from poll. */
int doPoll(int sock, short events, short *revents)
{
	struct pollfd fds[1];
	int res;
	tc_dbg_name(func)

again:
	fds[0].fd = sock;
	fds[0].events = events;
	fds[0].revents = 0;

	res = poll(fds, 1, -1);
	if(res == -1) {
		if(errno == EINTR)
			goto again;
		/* TODO: have various verbosity levels, and this error message
		 * be printed only at the highest level */
		tclient_dbg(1, ("%s: Error polling socket %d\n", func, sock));
		return TC_ERROR;
	} else if(fds[0].revents & (POLLERR | POLLHUP | POLLNVAL)) {
		tclient_dbg(1, ("%s: Error condition on socket\n", func));
		if(revents)
			*revents = fds[0].revents;
		return TC_ERROR;
	}

	if(revents)
		*revents = fds[0].revents;

	return TC_OK;
}

void tclient_msg(const char *fmt, ...)
{
	va_list v;
	time_t t;
	struct tm b;

	time(&t);
	localtime_r(&t, &b);

	va_start(v, fmt);
	vfprintf(stdout, fmt, v);
	va_end(v);
}

void tclient_verb(const char *fmt, ...)
{
	va_list v;
	time_t t;
	struct tm b;

	time(&t);
	localtime_r(&t, &b);

	va_start(v, fmt);
	vfprintf(stderr, fmt, v);
	va_end(v);
}


void _tclient_dbg(const char *fmt, ...)
{
	va_list v;
	time_t t;
	struct tm b;

	time(&t);
	localtime_r(&t, &b);

	va_start(v, fmt);
	fprintf(stderr, "(%02d:%02d:%02d) ", b.tm_hour, b.tm_min, b.tm_sec);
	vfprintf(stderr, fmt, v);
	va_end(v);
}

#define TC_BYTE_STRING_LEN	32
#define MEGA		1024*1024
#define MEGA_SHIFT	20
#define KILO		1024
#define KILO_SHIFT	10

static char* byte_to_string(int bytes)
{
	static char *string = NULL;
	int len = TC_BYTE_STRING_LEN;

	if(!string) {
		string = malloc(len);
		if(!string)
			return NULL;
	}

	if(bytes >= MEGA) {
		bytes >>= MEGA_SHIFT;
		snprintf(string, len, "%d MB", bytes);
	} else if(bytes >= KILO) {
		bytes >>= KILO_SHIFT;
		snprintf(string, len, "%d KB", bytes);
	} else {
		snprintf(string, len, "%d bytes", bytes);
	}

	return string;
}

static int tc_setNonBlocking(int sock)
{
	int opts;

	opts = fcntl(sock,F_GETFL);
	if (opts < 0) {
		perror("fcntl(F_GETFL)");
		return TC_ERROR;
	}
	opts = (opts | O_NONBLOCK);
	if (fcntl(sock,F_SETFL,opts) < 0) {
		return TC_ERROR;
	}

	return TC_OK;
}

/* UDP functions */
int tc_initUDP(struct tc_fb *fb, struct tc_snd *snd, int tsd_port)
{
	struct sockaddr_in addr;
	socklen_t addrlen;
	int lport;

	addrlen = sizeof(addr);

	if(getsockname(snd->sock, (struct sockaddr*)&addr, &addrlen) == -1) {
		tc_error_msg("Couldn't get address of local socket: %s\n",
			strerror(errno));
		return TC_ERROR;
	}

	/* set up a UDP socket bound to the same port as our TCP socket */
	lport = ntohs(addr.sin_port);
	snd->usock = tc_udpSocket(lport);
	if(snd->usock == TC_ERROR) {
		tc_error_msg("Couldn't create UDP socket\n");
		return TC_ERROR;
	}

	/* now connect to tsd's UDP socket */
	memcpy(&addr, &snd->addr, sizeof(addr));
	addr.sin_port = htons(tsd_port);

	if(connect(snd->usock, (struct sockaddr*)&addr, addrlen) == -1) {
		tc_error_msg("Couldn't connect to client's UDP "
			"socket: %s\n", strerror(errno));
		return TC_ERROR;
	}

	/* TODO:
	 * if wanted to traverse a client firewall, do something like this
	 */
#if 0
	{
		int dummy = 0;
		write(snd->usock, &dummy, sizeof(dummy));
	}
#endif

	tclient_msg("Info: Set up UDP port, bound to local port %d and "
		"connected to remote port %d\n", lport, tsd_port);

	return TC_OK;
}

/* port is the port to bind to */
static int tc_udpSocket(int port)
{
	struct sockaddr_in addr_s;
	socklen_t addrlen;
	int sock, reuse;

	if ( (sock = socket(PF_INET, SOCK_DGRAM, 0)) == -1) {
		return TC_ERROR;
	}

	reuse = 1;
	if(setsockopt(sock, SOL_SOCKET, SO_REUSEADDR, &reuse, 
		sizeof(reuse)) == -1) 
	{
		close(sock);
		return TC_ERROR;
	}

	if (tc_setNonBlocking(sock) == TC_ERROR) {
		close(sock);
		return TC_ERROR;
	}

	addr_s.sin_family = AF_INET;
	addr_s.sin_port = htons(port);

	addr_s.sin_addr.s_addr = htonl(INADDR_ANY);
	addrlen = sizeof(addr_s);

	if(bind(sock, (struct sockaddr*) &addr_s, addrlen) == -1) {
		tc_error_msg("Couldn't bind listening socket: %s\n",
			strerror(errno));
		close(sock);
		return TC_ERROR;
	}

	return sock;
}
