/* Copyright 2002, 2003, 2004, 2005, 2006, 2007 Ricardo Baratto */
/* Copyright 2002, 2003, 2004, 2005, 2006, 2007 Columbia University */
/* This file is part of THINC.

    THINC is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 2 of the License, or
    (at your option) any later version.

    THINC is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with THINC.  If not, see <http://www.gnu.org/licenses/>.
*/
/*
 * X11 related functions for thinc client
 *
 * Author:
 * Ricardo Baratto
 * Network Computing Lab - Columbia University
 *
 * Copyright 2004
 */

/* set tabstop 8 */

#include "thinc_client.h"
#include "tc_x11.h"
#include <stdlib.h>
#include <stdio.h>
#include <netinet/in.h>

/* for argb cursors */
#include <X11/Xcursor/Xcursor.h>

/* GC's used to handle the messages. Very primitive "GC manager" :)
 * The idea is that each function that needs a GC can take one from here
 * and use it continuously. 
 * Use tc_findScratchGC() if possible and you don't have to worry about
 * any of this.
 * Initially the GCs need to be created, so the function should use num_gcs to 
 * figure out the next available slot (note that since GCs will only be 
 * destroyed when the program exits, things are _a lot_ simpler).
 * If it finds a free slot, it creates it, saves the index, updates num_gcs,
 * marks it as valid (put 1 in validGCS, see below for story on validity) 
 * and starts using the GC.
 * If it doesn't find a free slot, then it needs to reuse one. That's
 * when validGCS comes into play: If a function has to take a GC away from
 * somebody else, it should mark it as invalid, so that the GC is not reused
 * blindly (in case values are cached, etc...). So it should put -1
 * at the index it decides to reuse and then use the GC.
 * NOTE:
 * It's not clear at this moment if it should be better to look for any GCs
 * that are invalid and keep making those invalid (since its users already
 * have to check, the penalty seems to be lower) or if the "invalidity"
 * should be spread over all the GCs. If you use tc_findScratchGC, it will
 * look for invalid GCs and use the first one it finds
 * TODO:
 * Would be nice if instead of just making the GC invalid, a function could
 * say "I was the last one to use it". This way, if it uses it again before
 * anybody else, it won't have to "revalidate" it. Maybe using some sort of
 * ids?
 * TODO:
 * Destroy the GCs on shutdown
 */
#define MAX_GCS		5
static int num_gcs=0;
static GC tc_GCS[MAX_GCS];
static char validGCS[MAX_GCS];

#ifdef TC_INSTRUMENT
#if 0
#define tc_sync(disp) \
	XSync((disp), False)
#else
#define tc_sync(disp)
#endif
#else
#define tc_sync(disp)
#endif	/* TC_INSTRUMENT */

#define tc_findScratchGC(idx) { \
	int i;	\
	if(num_gcs < MAX_GCS) { \
		/* space available, grab next one */ \
		idx = num_gcs++; \
		validGCS[idx] = 1; \
	} else {		\
		/* nothing available, must reuse one. Find one that's */  \
		/* already invalid */	\
		for(i=0; i<num_gcs; i++) {	\
			if(validGCS[i] == -1) {	\
				idx = i;	\
				break;		\
			}			\
		}	\
		/* none invalid, get first one */	\
		if(i == num_gcs) {	\
			idx = 0;	\
			validGCS[idx] = -1;	\
		}	\
	}	\
}

#ifdef TC_DONT_UPDATE_X
static struct tc_fillHandle empty = {(Pixmap)0, (GC)0};
#endif

/* hack so keys don't get stuck */
struct a_key_down {
	int key;
	struct a_key_down *next;
};

struct a_key_down* first_key_down;

static int release_all_keys(struct tc_fb *fb, struct a_key_down *head);

/* local funcs */
static int getVisual(struct tc_fb *fb, int class, int depth);
static int createWindow(struct tc_fb *fb, char *name, int width, int height, 
	int depth, int argc, char **argv);
static void createGC(struct tc_fb *fb);
static int doCreateImage(struct tc_fb *fb, int format);

/** cursor functions **/
static int doCreateCursor(struct tc_fb *fb);
static void doDestroyCursor(struct tc_fb *fb);
static void doUninitCursor(struct tc_x11 *x11);
static int doCreateCursor_bitmap(struct tc_fb *fb);
static int doCreateCursor_argb(struct tc_fb *fb);
static XcursorImage* tc_createCursorImage(char *pixels,
	unsigned int width, unsigned int height, unsigned int xhot,
	unsigned int yhot);

static void tc_ExposeEv(struct tc_fb *fb, XEvent* e);
static int tc_ButtonEv(struct tc_fb *fb, XEvent* e, int updown);
static int tc_MotionEv(struct tc_fb *fb, XEvent* e);
static int tc_KeyEv(struct tc_fb *fb, XEvent* e, int updown);
static int tc_createShmImage(struct tc_fb *fb, int format);
static void tc_shmDestroy(struct tc_fb *fb);
static int tc_shmError(Display *disp, XErrorEvent *error);

static int tc_rgn_init(struct tc_fb *fb);
static void tc_rgn_end(struct tc_fb *fb);
static inline int tc_rgn_add_normal_coords(struct tc_fb *fb, int x, int y, 
	int w, int h, int update);
static int tc_rgn_add_normal(struct tc_fb *fb, XRectangle *rect, int update);
static int tc_rgn_update_win(struct tc_fb *fb);
static void tc_rgn_update_win_normal(struct tc_fb *fb);
static void tc_rgn_update_expose(struct tc_fb *fb);
static int tc_x11copyData_normal(struct tc_fb *fb, XRectangle *src, 
	XRectangle *dst);
static int tc_x11copyData_resized(struct tc_fb *fb, XRectangle *src, 
	XRectangle *dst);
static int tc_ResizeEv(struct tc_fb *fb, XEvent* e);
static int tc_rgn_update_resized(struct tc_fb *fb);
static inline int tc_rgn_add_resized_coords(struct tc_fb *fb, int rx, int ry, 
	int rw, int rh);
static int tc_rgn_add_resized(struct tc_fb *fb, XRectangle *rszd);
static inline int tc_rgn_add_resized_coords_pixmap(struct tc_fb *fb, int rx, 
	int ry, int rw, int rh);
static int tc_rgn_add_resized_pixmap(struct tc_fb *fb, XRectangle *rszd);


/****************
 * Init/Destroy *
 ****************/

/* creates everything necessary to display the desktop image. To change
 * the contents of the image just change the passed pointer's contents
 * and call t_updateImage
 * Returns TC_OK on sucess, TC_ERROR on error */
int tc_x11Init(struct tc_fb *fb, int argc, char **argv)
{
	char *display_name = NULL;
	int class = TrueColor;
	int format = ZPixmap;
	char *title = TCLIENT_TITLE;
	struct tc_x11 *x11;

	fb->x11 = malloc(sizeof(*fb->x11));
	if(!fb->x11) {
		tclient_msg("Error: Couldn't allocate memory for X11 "
			"data\n");
		return TC_ERROR;
	}
	/* init so we can un-init */
	x11 = fb->x11;
	x11->disp = NULL;
	x11->Main = (Window)-1;
	x11->theGC = (GC)-1;
	x11->DesktopP = (Pixmap)-1;
	x11->visual = NULL;
	x11->img = NULL;
	x11->w_depth = -1;
	x11->use_shm = T_FALSE;
	x11->updated = T_FALSE;
	x11->exposed = T_FALSE;
	x11->damageRgn = NULL;
	x11->exposeRgn = NULL;
	x11->resizedRgn = NULL;
	x11->resized = T_FALSE;
	x11->resizedP = (Pixmap) -1;
	x11->focused = T_FALSE;
	x11->mouse_in = T_FALSE;

#ifndef NO_X
	/* open the connection to the x server */
	if((x11->disp = XOpenDisplay(display_name)) == NULL) {
		tclient_msg("Error: can't open display %s\n", 
			XDisplayName(display_name));
		free(x11);
		return TC_ERROR;
	}
#endif

	tc_rgn_init(fb);

#ifndef NO_X
	/* get the visual */
	if(getVisual(fb, class, fb->depth) == TC_ERROR) {
		tclient_msg("Error: Couldn't find a suitable visual\n");
		goto error;
	}

	/* create the window */
	if(createWindow(fb, title, fb->width, fb->height, fb->depth, 
		argc, argv) == TC_ERROR) {
		tclient_msg("Error: couldn't create main window\n");
		goto error;
	}

	/* create the gc */
	createGC(fb);

	/* create the image */
	if(doCreateImage(fb, format) == TC_ERROR) {
		tclient_msg("Error: couldn't create desktop image\n");
		goto error;
	}

	/* Copy the image to the pixmap */
	/* TODO: How about using ShmPixmaps instead of a Shm Image?? */
	if(x11->use_shm)
		XShmPutImage(x11->disp, x11->DesktopP, x11->theGC, 
			x11->img, 0, 0, 0, 0, x11->img->width, 
			x11->img->height, False);
	else
		XPutImage(x11->disp, x11->DesktopP, x11->theGC, x11->img, 
			0, 0, 0, 0, x11->img->width, x11->img->height);

	/* Create the cursor */
	if(fb->cursorInfo.cursor) {
		if(doCreateCursor(fb) == TC_ERROR) {
			tclient_msg("Error: Couldn't create cursor\n");
			goto error;
		}
	}

	/* hack so keys don't get stuck */
	first_key_down = NULL;

	/* ok, map the window and we're all set */
	XMapWindow(x11->disp, x11->Main);

	XFlush(x11->disp);
#endif /* !NO_X */

	return TC_OK;

error:
	tc_rgn_end(fb);
#ifndef NO_X
	if(x11->theGC != (GC)-1) XFreeGC(x11->disp, x11->theGC);
	if(x11->img) XDestroyImage(x11->img);
	if(x11->DesktopP != (Pixmap)-1) XFreePixmap(x11->disp, x11->DesktopP);
	if(x11->Main != (Window)-1) XDestroyWindow(x11->disp, x11->Main);
	
	XCloseDisplay(x11->disp);
#endif	/* ! NO_X */
	
	free(fb->x11);
	return TC_ERROR;
}

/* shutdown X */
void tc_x11shutdown(struct tc_fb *fb)
{
	struct tc_x11 *x11 = fb->x11;

	tc_rgn_end(fb);
#ifndef NO_X
	doDestroyCursor(fb);
	if(x11->theGC != (GC)0) XFreeGC(x11->disp, x11->theGC);
	if(x11->img) {
		if(x11->use_shm)
			tc_shmDestroy(fb);
		else
			XDestroyImage(x11->img); 
	}
	if(x11->DesktopP != (Pixmap)-1) XFreePixmap(x11->disp, x11->DesktopP);
	if(x11->Main != (Window)0) XDestroyWindow(x11->disp, x11->Main);

	if(x11->disp) XCloseDisplay(x11->disp);
	free(x11);
#endif	/* ! NO_X */
}

inline int tc_x11GetFD(struct tc_fb *fb)
{
#ifndef NO_X
	return XConnectionNumber(fb->x11->disp);
#else
	return 0;
#endif	/* ! NO_X */
}

static int getVisual(struct tc_fb *fb, int class, int depth)
{
	struct tc_x11 *x11 = fb->x11;
	const char *func = __func__;
	XVisualInfo tmpl, *vinfo;
	int nvis, mask;

	tmpl.screen = DefaultScreen(x11->disp);
	tmpl.class = class;
	tmpl.depth = depth;
	mask = VisualScreenMask | VisualClassMask | VisualDepthMask;
	vinfo = XGetVisualInfo(x11->disp, mask, &tmpl, &nvis);
	if(!vinfo) {
		tclient_msg("%s: Couldn't find suitable visual class %d, depth %d\n", func, class, depth);
		return TC_ERROR;
	}
	x11->visual = vinfo[0].visual;
	XFree(vinfo);

	return TC_OK;
}

static void createGC(struct tc_fb *fb)
{
	struct tc_x11 *x11 = fb->x11;
	unsigned long bgpix, fgpix;
	XGCValues gcv;
	
	bgpix = BlackPixel(x11->disp, DefaultScreen(x11->disp));
	fgpix = WhitePixel(x11->disp, DefaultScreen(x11->disp));

	gcv.foreground = fgpix;
	gcv.background = bgpix;

	x11->theGC = XCreateGC(x11->disp, x11->Main, 
		(GCForeground | GCBackground), &gcv);
}

/* creates window on x11->display x11->disp with passed parameters. Returns
 * (Window) 0 on error */
static int createWindow(struct tc_fb *fb, char *name, int width, 
	int height, int depth, int argc, char **argv)
{
	struct tc_x11 *x11 = fb->x11;
	const char *func = __func__;
	XSizeHints *xsh;
	XClassHint *xch;
	XWMHints *xwmh;
	XTextProperty wname, iname;
	unsigned long fgpix, bgpix, valuemask;
	XSetWindowAttributes xswa;

	bgpix = BlackPixel(x11->disp, DefaultScreen(x11->disp));
	fgpix = WhitePixel(x11->disp, DefaultScreen(x11->disp));

	if((xsh = XAllocSizeHints()) == NULL) {
		tclient_msg("%s: Error allocating size hints\n", func);
		return TC_ERROR;
	}

	/* FIXME: Need to implement scrollbar so we don't have min size.
	 * Right now we don't allow resizing */
	if(fb->do_resize) {
		xsh->flags = PSize | PMinSize;
		xsh->width = width;
		xsh->height = height;
		/* FIXME: maximizing beyond original size is broken, chk 
		   XGetSubImage and the way we create our resized image */
		xsh->max_width = width;
		xsh->max_height = height;
		xsh->flags |= PMaxSize;
		xsh->min_width = 10;
		xsh->min_height = 10;
	} else {
		xsh->flags = PSize | PMaxSize | PMinSize;
		xsh->min_width = xsh->max_width = xsh->width = width;
		xsh->min_height = xsh->max_height = xsh->height = height;
	}

	/***************** create the window ****************/
	x11->Main = XCreateSimpleWindow(x11->disp, DefaultRootWindow(x11->disp),
		0, 0, /* initial position */
		xsh->width, xsh->height, 
		0, /* border width */
		fgpix, bgpix);

	/**************** create the offscreen pixmap *******/
	/* TODO: Handle a pixmap creation error (asynchronous notification) */
	x11->DesktopP = XCreatePixmap(x11->disp, x11->Main, xsh->width, 
		xsh->height, depth);

	/************* wm properties **********************/
	if((xch = XAllocClassHint()) == NULL) {
		tclient_msg("%s: Error allocating class hint\n", func);
		return TC_ERROR;
	}
	xch->res_name = name;
	xch->res_class = name;

	/* window and icon name */
	/* FIXME: Leaking memory here: how to free this? */
	if( XStringListToTextProperty(&name, 1, &wname) == 0 ||
		XStringListToTextProperty(&name, 1, &iname) == 0) {
		tclient_msg("%s: Error creating XTextProperty\n", func);
		return TC_ERROR;
	}

	/* XXX: i think this is deprecated in the x world... one day we'll
	 * fix it */
	if((xwmh = XAllocWMHints()) == NULL) {
		tclient_msg("%s: Error allocating WM hints\n", func);
		return TC_ERROR;
	}

	xwmh->flags = (InputHint|StateHint);
	xwmh->input = True;
	xwmh->initial_state = NormalState;
	XSetWMProperties(x11->disp, x11->Main, &wname, &iname, argv, argc,
		xsh, xwmh, xch);
	xswa.colormap = DefaultColormap(x11->disp, DefaultScreen(x11->disp));
	xswa.bit_gravity = CenterGravity;
	xswa.event_mask = 	ExposureMask 
				| PointerMotionMask
				| ButtonPressMask | ButtonReleaseMask
				| KeyPressMask | KeyReleaseMask
				| FocusChangeMask 
				| EnterWindowMask | LeaveWindowMask
				;
	if(fb->do_resize)
		xswa.event_mask |= StructureNotifyMask;

	valuemask = (CWColormap|CWBitGravity|CWEventMask);
	if(fb->fullscreen) {
		xswa.event_mask |= EnterWindowMask;
		xswa.override_redirect = True;
		valuemask |= CWOverrideRedirect;
	}

	XChangeWindowAttributes(x11->disp, x11->Main, valuemask, &xswa);


	/* save the window's depth for later use */
	x11->w_depth = depth;

	/* cleanup */
	XFree(xch);
	XFree(xsh);
	XFree(xwmh);

	/* voila! */
	return TC_OK;
}

/* creates an image on x11->display x11->disp with passed parameters. Returns
 * NULL on error */
static int doCreateImage(struct tc_fb *fb, int format)
{
	struct tc_x11 *x11 = fb->x11;
	tc_dbg_name(func)

	tclient_dbg(1, ("%s: creating image %dx%dx%d\n", func, fb->width, 
		fb->height, fb->depth));

	if(tc_createShmImage(fb, format) == TC_ERROR) {
		x11->use_shm = T_FALSE;
	} else {
		x11->use_shm = T_TRUE;
	}
	if(!x11->img) {
		x11->use_shm = T_FALSE;
		x11->img = thinc_createImage(fb, fb->width, fb->height,
			fb->fbdata);
		if(!x11->img) {
			tclient_msg("Error: Couldn't create desktop image\n");
			return TC_ERROR;
		}
	}
	return TC_OK;
}

/* Returns void so header files don't need to include X headers. Cast to
 * XImage*. If data is NULL, memory is allocated before creating the
 * image (you can access it through img->data */
void* thinc_createImage(struct tc_fb *fb, unsigned int width, 
	unsigned int height, char *data)
{
	struct tc_x11 *x11 = fb->x11;
	int offset = 0;
	int bitmap_pad = 32;
	int bytes_per_line = 0; /* # of bytes between start of successive scan
				   lines. 0 -> lines are contiguous */
	char *img_data;
	XImage *img;

	if(!data) {
		/* TODO: We definitely need to store Bpp (bpp/8) */
		img_data = malloc(width*height*fb->Bpp);
		if(!img_data)
			return NULL;
	} else
		img_data = data;
	img =  XCreateImage(x11->disp, x11->visual, fb->depth, ZPixmap,
		offset, (char*)img_data, width, height, bitmap_pad, 
		bytes_per_line);
	if(!img) {
		if(!data)
			free(img_data);
		return NULL;
	}

	if(fb->byte_order == T_LITTLE)
		img->byte_order = LSBFirst;
	else
		img->byte_order = MSBFirst;

	return img;
}

static Bool shmFailed = False;

static int tc_createShmImage(struct tc_fb *fb, int format)
{
	struct tc_x11 *x11 = fb->x11;
	int ourwidth;
  	XErrorHandler prev;

	/* do we have SHM? */
	if(!XShmQueryExtension(x11->disp)) {
		tclient_msg("Warning: SHM extension not found in X server\n");
		return TC_ERROR;
	}
	
	/* create shm x11->img */
	x11->img = XShmCreateImage(x11->disp, x11->visual, fb->depth, format, 
		NULL, &x11->shminfo, fb->width, fb->height);
	if(!x11->img) {
		tclient_msg("Error: Couldn't create SHM image\n");
		return TC_ERROR;
	}

	/* FIXME: This image will have bytes_per_line bytes per line, which 
	 * could be different from width*bpp/8 (what we normally use).
	 * Need to let everybody else know about this
	 */
	ourwidth = fb->width * (fb->Bpp);
	if(ourwidth != x11->img->bytes_per_line) {
		tclient_msg("Error: ShmImage has different width than us. "
			"Tell author to fix this problem. Sorry!\n");
		XDestroyImage(x11->img);
		return TC_ERROR;
	}

	/* allocate memory for it */
	x11->shminfo.shmid = shmget (IPC_PRIVATE, 
		x11->img->bytes_per_line * x11->img->height, IPC_CREAT|0777);
	if(x11->shminfo.shmid == -1) {
		tclient_msg("Error: Couldn't allocate SHM data\n");
		XDestroyImage(x11->img);
		return TC_ERROR;
	}
	/* attach memory to us */
	x11->shminfo.shmaddr = x11->img->data = shmat(x11->shminfo.shmid, 0, 0);
	if(x11->shminfo.shmaddr == (char*)-1) {
		tclient_msg("Error: Couldn't attach SHM segment\n");
		shmctl(x11->shminfo.shmid, IPC_RMID, 0);
		XFree(x11->img);
		return TC_ERROR;
	}
	/* if we set it to True then GetImage's won't work.. do we
	 * care? */
	x11->shminfo.readOnly = True;

	/* catch errors (for example, when being tunneled over ssh) and
	 * revert */
  	prev = XSetErrorHandler(tc_shmError);
	/* tell server to attach the memory segment */
	if(XShmAttach(x11->disp, &x11->shminfo)  == 0) {
		tclient_msg("Error: Couldn't make the server attach "
			"the shared memory\n");
		shmdt(x11->shminfo.shmaddr);
		shmctl(x11->shminfo.shmid, IPC_RMID, 0);
		XFree(x11->img);
		return TC_ERROR;
	}
	/* sync to make sure we ok with this shm stuff */
	XSync(x11->disp, False);
	/* put back old handler and check if we failed */
	XSetErrorHandler(prev);
	if(shmFailed) {
		tclient_msg("Warning: Can't use Shared Memory, reverting to "
			"ye old XImage\n");
		shmdt(x11->shminfo.shmaddr);
		shmctl(x11->shminfo.shmid, IPC_RMID, 0);
		XFree(x11->img);
		x11->img = NULL;
		return TC_ERROR;
	}

	/* now copy the data to the image */
	memcpy(x11->img->data, fb->fbdata, 
		x11->img->bytes_per_line * x11->img->height);
	free(fb->fbdata);
	fb->fbdata = (char*) x11->img->data;

	return TC_OK;
}

static int tc_shmError(Display *disp, XErrorEvent *error)
{
	/* gotta use da global vars */
	shmFailed = True;
	return 0;
}

static void tc_shmDestroy(struct tc_fb *fb)
{
	struct tc_x11 *x11 = fb->x11;
	if(x11->use_shm) {
		XShmDetach(x11->disp, &x11->shminfo);
		shmdt(x11->shminfo.shmaddr);
		shmctl(x11->shminfo.shmid, IPC_RMID, 0);
		XFree(x11->img);
		fb->fbdata = NULL;
	}
}

static int doCreateCursor(struct tc_fb *fb)
{
	if(fb->cursorInfo.argb)
		return doCreateCursor_argb(fb);
	else
		return doCreateCursor_bitmap(fb);
}

static void doUninitCursor(struct tc_x11 *x11)
{
	if(x11->theCursor == (Cursor)0)
		return;

	XUndefineCursor(x11->disp, x11->Main);
	XFreeCursor(x11->disp, x11->theCursor);
}

static void doDestroyCursor(struct tc_fb *fb)
{
	struct tc_x11 *x11 = fb->x11;
	struct tc_cursorInfo *cursor = &fb->cursorInfo;

	doUninitCursor(x11);
	free(cursor->cursor);
	cursor->cursor = cursor->mask = cursor->source = NULL;
}

static int doCreateCursor_bitmap(struct tc_fb *fb)
{
	struct tc_x11 *x11 = fb->x11;
	tc_dbg_name(func)
	Pixmap source;
	Pixmap mask;
	XColor fg, bg;
	unsigned long fgv, bgv;

	tclient_dbg(1, ("%s: cursor size %d\n", func, fb->cursorInfo.size));

	source = XCreateBitmapFromData(x11->disp, x11->Main, 
		(char*) fb->cursorInfo.source, fb->cursorInfo.width, 
		fb->cursorInfo.height);
	if(source == None) {
		tclient_msg("Error: Couldn't create cursor\n");
		return TC_ERROR;
	}
	mask = XCreateBitmapFromData(x11->disp, x11->Main, 
		(char*) fb->cursorInfo.mask, fb->cursorInfo.width, 
		fb->cursorInfo.height);
	if(mask == None) {
		tclient_msg("Error: Couldn't create cursor\n");
		return TC_ERROR;
	}

	fgv = fb->cursorInfo.fg;
	bgv = fb->cursorInfo.bg;

	tclient_dbg(1, ("%s: fg %lu, bg %lu\n", func, fgv, bgv));

	/* The color is sent to us in 32 bits (even when depth is 8) so
	 * this is ok to assume */
	/* TODO: Make sure this actually works on 8bpp x11->displays */
	fg.pixel = 0;
	fg.red = (fgv & 0xFF0000) >> 16;
	fg.green = (fgv & 0xFF00) >> 8;
	fg.blue = (fgv & 0xFF);

	/* The color is sent to us in RGB, so we only get 8 bits per
	 * color. However, XCreatePixmapCursor (or XColor) has each channel 
	 * being 16 bits (why????), so we need to convert it to 0-65535 
	 * scale */
	fg.red = fg.red * 65535 / 255;
	fg.green = fg.green * 65535 / 255;
	fg.blue = fg.blue * 65535 / 255;

	bg.pixel = 0;
	bg.red = (bgv & 0xFF0000) >> 16;
	bg.green = (bgv & 0xFF00) >> 8;
	bg.blue = (bgv & 0xFF);

	bg.red = bg.red * 65535 / 255;
	bg.green = bg.green * 65535 / 255;
	bg.blue = bg.blue * 65535 / 255;

	x11->theCursor = XCreatePixmapCursor(x11->disp, source, mask,
		&fg, &bg, fb->cursorInfo.xhot, fb->cursorInfo.yhot);

	XFreePixmap(x11->disp, source);
	XFreePixmap(x11->disp, mask);

	XDefineCursor(x11->disp, x11->Main, x11->theCursor);

	return TC_OK;
}

static int doCreateCursor_argb(struct tc_fb *fb)
{
	tc_dbg_name(func)
	XcursorImage *image;
	struct tc_x11 *x11 = fb->x11;
	struct tc_cursorInfo *cursor = &fb->cursorInfo;

	tclient_dbg(1, ("%s: creating argb cursor\n", func));

	image = tc_createCursorImage(cursor->cursor, cursor->width,
		cursor->height, cursor->xhot, cursor->yhot);
	if(!image) {
		tclient_dbg(1, ("Error: %s: Couldn't allocate memory for "
			"cursor\n", func));
		return TC_ERROR;
	}

	x11->theCursor = XcursorImageLoadCursor(x11->disp, image);
	if(x11->theCursor == (Cursor)0) {
		tclient_dbg(1, ("%s: Couldn't create cursor\n", func));
		XcursorImageDestroy(image);
		return TC_ERROR;
	}
	XcursorImageDestroy(image);

	XDefineCursor(x11->disp, x11->Main, x11->theCursor);
	tclient_dbg(1, ("%s: finished\n", func));
	return TC_OK;
}

/* Xcursor's api is kind of lame, mostly tailored to loading cursors from
 * files. lamer so here's our own wrapper struct. as the api or internal
 * details of the structs change, make sure to keep this up to date.
 * Important: If you create a cursor image using this function, 
 * XCursorImageDestroy() won't free the pixel data */
static XcursorImage* tc_createCursorImage(char *pixels,
	unsigned int width, unsigned int height, unsigned int xhot,
	unsigned int yhot)
{
	XcursorImage *image;

	image = malloc (sizeof (*image));
	if (!image)
		return NULL;
	image->version = XCURSOR_IMAGE_VERSION;
	image->pixels = (XcursorPixel *) pixels;
	image->size = width > height ? width : height;
	image->width = width;
	image->height = height;
	image->delay = 0;
	image->xhot = xhot;
	image->yhot = yhot;
	return image;
}
	
/*************
 * Main loop *
 *************/

/* called when we're about to sleep waiting for events or data from the
 * server. It updates the window contents with all the changes done to the
 * pixmap */
inline void tc_x11_going_to_bed(struct tc_fb *fb)
{
#ifndef NO_X
	/* copy new updates to visible window */
	time_add("X_UP", TI_START);
	tc_rgn_update_win(fb);
	time_add("X_UP", TI_END);
#endif
}

/* copies new display content from pixmap to main window */
static int tc_rgn_update_win(struct tc_fb *fb)
{
	tc_dbg_name(func)
	struct tc_x11 *x11 = fb->x11;
	int did_something = T_FALSE;

	instr_vars(sany, eany);

	tclient_dbg(3, ("%s: updating main window. have updates? %d\n", func,
		x11->updated));
	/* if x11->updated is true, then we've had new updates from the server 
	 * since last time we were here. In the unresized case, we just
	 * copy the contents of the damage region from the pixmap to the
	 * window */ 
	if(x11->updated) {
		time_add("X_UP_UP", TI_START);
		tclient_dbg(3, ("%s: got new commands, copying to window\n",
			func));
		if(x11->resized) {
			/* not too easy.. */
			instr_start(sany);
			tc_rgn_update_resized(fb);
			instr_end(fb, sany, eany, x_update);
		} else {
			/* easy */
			instr_start(sany);
			tc_rgn_update_win_normal(fb);
			instr_end(fb, sany, eany, x_update);

		}
		x11->updated = T_FALSE;
		did_something = T_TRUE;
		time_add("X_UP_UP", TI_END);
	}
	tclient_dbg(3, ("%s: finished with updates, checking exposed: %d\n", 
		func, x11->exposed));

	/* if we got any expose events, we need to copy things back from 
	 * the pixmap to the window */
	if(x11->exposed) {
		time_add("X_UP_EXP", TI_START);
		tclient_dbg(3, ("%s: handling exposed regions\n", func));
		instr_start(sany);
		tc_rgn_update_expose(fb);
		instr_end(fb, sany, eany, x_expose);
		x11->exposed = T_FALSE;
		did_something = T_TRUE;
		time_add("X_UP_EXP", TI_END);
	}

	if(did_something) {
		time_add("X_UP_FLUSH", TI_START);
		XFlush(x11->disp);
		time_add("X_UP_FLUSH", TI_END);
	}

	tclient_dbg(3, ("%s: finished\n", func));
	return TC_OK;
}

static void tc_rgn_update_win_normal(struct tc_fb *fb)
{
	struct tc_x11 *x11 = fb->x11;
	XRectangle rect;
	instr_vars(s, e);

	XClipBox(x11->damageRgn, &rect);
	instr_start(s);
	XCopyArea(x11->disp, x11->DesktopP, x11->Main, x11->theGC,
		rect.x, rect.y, rect.width, rect.height, rect.x, rect.y);
	instr_end(fb, s, e, win_copy);

	/* empty region */
	XShrinkRegion(x11->damageRgn, rect.width, rect.height);

	instr_inc(fb, num_updates);

	instr_start(s);
	tc_sync(x11->disp);
	instr_end(fb, s, e, win_sync);
}

static void tc_rgn_update_expose(struct tc_fb *fb)
{
	struct tc_x11 *x11 = fb->x11;
	XRectangle rect;
	Pixmap p;

	if(XEmptyRegion(x11->exposeRgn)) {
		tclient_dbg(3, ("%s: expose region empty. why did you call "
			"me??\n", __func__));
		return;
	}

	if(x11->resized)
		p = x11->resizedP;
	else
		p = x11->DesktopP;

	XClipBox(x11->exposeRgn, &rect);
	tclient_dbg(3, ("%s: setting clip region: (%d,%d):(%dx%d)\n",
		__func__, rect.x, rect.y, rect.width, rect.height));

	XCopyArea(x11->disp, p, x11->Main, x11->theGC,
		rect.x, rect.y, rect.width, rect.height, rect.x, rect.y);
	/* empty region */
	XShrinkRegion(x11->exposeRgn, rect.width, rect.height);

	tclient_dbg(3, ("%s: finished\n", __func__));
}

/* this is a hack */
static int tc_rgn_update_resized(struct tc_fb *fb)
{
	tc_dbg_name(func)
	Region tmp, tmp_rsz;
	struct tc_x11 *x11 = fb->x11;
	XRectangle rect;

	tclient_dbg(1, ("%s: updating while being resized\n", func));
	
	tmp = XCreateRegion();
	tmp_rsz = XCreateRegion();
	if(!tmp || !tmp_rsz) {
		tclient_msg("Error: Couldn't allocate memory for temporary "
			"region\n");
		return TC_ERROR;
	}

	if(XEmptyRegion(x11->damageRgn)) {
		tclient_dbg(1, ("%s: damage region empty, just have resize "
			"updates\n", func));
		goto resized;
	}

	/* interesting way to copy a region :/ */
	XUnionRegion(x11->damageRgn, tmp, tmp);
	XUnionRegion(tmp, tmp_rsz, tmp_rsz);

	/* copy damaged part (after resizing) to resized pixmap */
	tc_resize_scale_region(fb, tmp_rsz);
	tc_resize_update_region_rects(fb, tmp, tmp_rsz);

resized:
	if(XEmptyRegion(x11->resizedRgn)) {
		tclient_dbg(1, ("%s: resized region is empty\n", func));
	} else
		/* now copy everything to the window */
		XUnionRegion(x11->resizedRgn, tmp_rsz, tmp_rsz);

	XClipBox(tmp_rsz, &rect);

	tc_x11_setclip_region(fb, tmp_rsz);

	tclient_dbg(1, ("%s: clip box: (%d,%d):(%dx%d)\n", func, 
		rect.x, rect.y, rect.width, rect.height));
	XCopyArea(x11->disp, x11->resizedP, x11->Main, x11->theGC,
		rect.x, rect.y, rect.width, rect.height, rect.x, rect.y);
	tc_x11_clearclip(fb);

	/* cleanup */
	XDestroyRegion(tmp_rsz);
	XDestroyRegion(tmp);
	XClipBox(x11->damageRgn, &rect);
	XShrinkRegion(x11->damageRgn, rect.width, rect.height);
	XClipBox(x11->resizedRgn, &rect);
	XShrinkRegion(x11->resizedRgn, rect.width, rect.height);

	return TC_OK;
}

/*************************
 * Handle Server Updates *
 *************************/

/* call when the desktop's data has changed. Returns TC_OK on success, 
 * TC_ERROR on error */
int tc_x11imgUpdate(struct tc_fb *fb, int x, int y, int width, int height)
{
	struct tc_x11 *x11 = fb->x11;
	instr_vars(s, e);

	time_add("RAW", TI_DRAW_START);

	instr_add(fb, raw_pixels, width*height);
	/* Copy to the Pixmap, and from there to the window */
	if(x11->use_shm)
		XShmPutImage(x11->disp, x11->DesktopP, x11->theGC, x11->img, 
			x, y, x, y, width, height, False);
	else
		XPutImage(x11->disp, x11->DesktopP, x11->theGC, x11->img, x, 
			y, x, y, width, height);

	tc_rgn_add_normal_coords(fb, x, y, width, height, T_TRUE);

	instr_start(s);
	tc_sync(x11->disp);
	instr_end(fb, s, e, raw_sync);

	time_add("RAW", TI_DRAW_END);
	return TC_OK;
}

/* updates resize regions */
int tc_x11update_resized(struct tc_fb *fb, int rsz_x, int rsz_y, int rsz_w,
	int rsz_h)
{
	XRectangle two;

	two.x = rsz_x;
	two.y = rsz_y;
	two.width = rsz_w;
	two.height = rsz_h;

	tc_rgn_add_resized_pixmap(fb, &two);
	return TC_OK;
}

/* copies the rectangle of size width*height from sx,sy to dx, dy. */
/* 
 * if resized:
 * if(updated) {
 * 	tmp = src - outdated
 * 	tmp' = resize(tmp)
 * 	copy [tmp] => [tmp']
 * }
 * src-rsz = resize(src)
 * dst-rsz = resize(dst)
 * copy [src-rsz] => [dst-rsz]
 *
 * add dst to outdated region
 * => TODO: Could copy non-outdated parts in desktop pixmap and only add
 * outdated parts to outdated region
 */
int tc_x11copyData(struct tc_fb *fb, int sx, int sy, int dx, int dy, 
	int width, int height)
{
	XRectangle src, dst;

	time_add("COPY", TI_DRAW_START);

	src.x = sx;
	src.y = sy;
	src.width = dst.width = width;
	src.height = dst.height = height;
	dst.x = dx;
	dst.y = dy;

	if(fb->x11->resized)
		tc_x11copyData_resized(fb, &src, &dst);
	else
		tc_x11copyData_normal(fb, &src, &dst);

	tc_sync(fb->x11->disp);

	time_add("COPY", TI_DRAW_END);
	return TC_OK;
}

static int tc_x11copyData_normal(struct tc_fb *fb, XRectangle *src, 
	XRectangle *dst)
{
	struct tc_x11 *x11 = fb->x11;

	tc_rgn_add_normal(fb, dst, T_TRUE);

	XCopyArea(x11->disp, x11->DesktopP, x11->DesktopP, x11->theGC, src->x,
		src->y, src->width, src->height, dst->x, dst->y);

	return TC_OK;
}
 
/* what to do:
 * -> intersect dmg and src
 * -> copy intersection to resized pixmap
 * -> issue copy for whole region in resized pixmap
 * -> take out intersect from dmg
 * -> add intersect and dst to resizedRgn
 *
 * but since i'm lazy:
 * -> copy dmg to resized (tc_resize_update_region()
 * -> issue copy and empty dmg
 * -> add dmg and dst to resizedRgn
 */
static int tc_x11copyData_resized(struct tc_fb *fb, XRectangle *src, 
	XRectangle *dst)
{
	tc_dbg_name(func)
	struct tc_x11 *x11 = fb->x11;
	Region tmp_rsz;
	XRectangle rect, zsrc, zdst;

	tclient_dbg(1, ("%s: copying data while resized. coords: "
		"(%d,%d)=>(%d,%d)::(%dx%d)\n", func, src->x, src->y,
		dst->x, dst->y, dst->width, dst->height));

	if(XEmptyRegion(x11->damageRgn)) {
		tclient_dbg(1, ("%s: no outstanding updates, issuing copy "
			"directly to resized pixmap\n", func));
		goto resized;
	}

	tmp_rsz = XCreateRegion();
	if(!tmp_rsz) {
		tclient_msg("Error: Couldn't allocate memory for temporary "
			"region\n");
		return TC_ERROR;
	}

	XUnionRegion(x11->damageRgn, tmp_rsz, tmp_rsz);
	tc_resize_scale_region(fb, tmp_rsz);
	tc_resize_update_region_rects(fb, x11->damageRgn, tmp_rsz);
	/* add to resizedRgn */
	XClipBox(x11->damageRgn, &rect);
	tc_rgn_add_resized(fb, &rect);

	tclient_dbg(1, ("%s: copied outstanding requests to resized pixmap. "
		"cleaning up\n", func));
	XShrinkRegion(x11->damageRgn, rect.width, rect.height);
	XDestroyRegion(tmp_rsz);

resized:

	zsrc = *src;
	zdst = *dst;
	tc_resize_scale_rect(fb, &zsrc);
	tc_resize_scale_rect(fb, &zdst);

	tclient_dbg(1, ("%s: resized copy: "
		"(%d,%d)=>(%d,%d)::(%dx%d)\n", func, zsrc.x, zsrc.y,
		zdst.x, zdst.y, zdst.width, zdst.height));

	XCopyArea(x11->disp, x11->resizedP, x11->resizedP, x11->theGC,
		zsrc.x, zsrc.y, zdst.width, zdst.height, zdst.x, zdst.y);

	tc_rgn_add_resized_pixmap(fb, &zdst);

	tclient_dbg(1, ("%s: finished\n", func));
	return TC_OK;
}

/* fills the rectangle (x,y)x(width,height) with the color in pixel.
 * One of the most horrendous functions ever written. tries to be all so
 * smart... */
int tc_x11solidFill(struct tc_fb *fb, int x, int y, int width, int height, 
	unsigned long pixel)
{
	struct tc_x11 *x11 = fb->x11;
	static int idx = -1;
	static unsigned long prevpixel = 0, prevprevpixel = 0;
	XGCValues gcv;
	tc_dbg_name(func)
	XRectangle rect;
	instr_vars(sany, eany);
	static GC blackGC;
	static GC whiteGC;
	static unsigned long black = 0;
	static unsigned long white = 0xFFFFFF;
	static GC fillGC;
	time_add("SFILL", TI_DRAW_START);

	tclient_dbg(1, ("%s: pixel %ld, prevpixel %ld\n", func,
		pixel, prevpixel));

	/* GC stuff */
	if(idx == -1) {
		instr_start(sany);
		/* don't have a GC, get an index and create it */
		tc_findScratchGC(idx);
		tclient_dbg(1, ("%s: got GC index %d\n", func, idx));
		/* create the GC */
		gcv.foreground = pixel;
		gcv.fill_style = FillSolid;
		tc_GCS[idx] = XCreateGC(x11->disp, x11->Main, 
			(GCForeground | GCFillStyle), &gcv);
		instr_end(fb, sany, eany, sfill_creategc);
		instr_start(sany);
		tc_sync(x11->disp);
		instr_end(fb, sany, eany, sfill_sync_prepare_create);
		instr_inc(fb, sfill_create_gc);
		gcv.foreground = black;
		blackGC = XCreateGC(x11->disp, x11->Main, 
			(GCForeground | GCFillStyle), &gcv);
		gcv.foreground = white;
		whiteGC = XCreateGC(x11->disp, x11->Main, 
			(GCForeground | GCFillStyle), &gcv);
		fillGC = tc_GCS[idx];
	} else if(pixel == black) {
		fillGC = blackGC;
	} else if(pixel == white) {
		fillGC = whiteGC;
	} else {
		/* have a GC. chk if it's still valid and if the previous
		 * color is the same as the new one. If it's invalid it
		 * has to stay that way */
		if((pixel != prevpixel) && (pixel != prevprevpixel)) {
			instr_inc(fb, sfill_change_color);
		}
		if(validGCS[idx] == -1 || pixel != prevpixel) {
			/* need to change the values in the GC */
			tclient_dbg(1, ("%s: gc not valid, refreshing\n",
				func));
			instr_start(sany);
			gcv.foreground = pixel;
			gcv.fill_style = FillSolid;
			XChangeGC(x11->disp, tc_GCS[idx], 
				(GCForeground | GCFillStyle), &gcv); 
			instr_end(fb, sany, eany, sfill_changegc);
			instr_start(sany);
			tc_sync(x11->disp);
			instr_end(fb, sany, eany, sfill_sync_prepare_change);
			instr_inc(fb, sfill_change_gc);
			fillGC = tc_GCS[idx];
		}
	}

	prevprevpixel = prevpixel;
	prevpixel = pixel;

	/* ok, draw the rectangle */
	tclient_dbg(1, ("%s: filling rectangle x %d y %d w %d h %d\n",
		func, x, y, width, height));

	if(x11->resized) {
		rect.x = x;
		rect.y = y;
		rect.width = width;
		rect.height = height;
		tc_resize_scale_rect(fb, &rect);
		if(rect.width == 0) rect.width = 1;
		if(rect.height == 0) rect.height = 1;
		XFillRectangle(x11->disp, x11->resizedP, fillGC, 
			rect.x, rect.y, rect.width, rect.height);
		tc_rgn_add_resized_coords_pixmap(fb, rect.x, rect.y, 
			rect.width, rect.height);
	} else {
		instr_add(fb, sfill_pixels, width*height);
		instr_start(sany);
		XFillRectangle(x11->disp, x11->DesktopP, fillGC, x, y, 
			width, height);
		instr_end(fb, sany, eany, sfill_fill);

		instr_start(sany);
		tc_rgn_add_normal_coords(fb, x, y, width, height, T_TRUE);
		instr_end(fb, sany, eany, sfill_addcoords);
	}
	instr_start(sany);
	tc_sync(x11->disp);
	instr_end(fb, sany, eany, sfill_sync_draw);

	tclient_dbg(1, ("%s: finished\n", func));

	time_add("SFILL", TI_DRAW_END);
	return TC_OK;
}

/* here's the way it works:
 * - call tc_x11pfillNew, with the pixmap data. It will return a pixmap
 *   and GC (as an opaque handle).
 * - if you want to cache the pixmap (according to the message flags),
 *   add the returned handle to the cache.
 * - to do the actual drawing, call tc_x11pfillFill() with the opaque handle 
 *   and the rect you want to fill
 */
#ifdef TC_DONT_UPDATE_X
void* tc_x11pfillNew(struct tc_fb *fb, char *data, int x, int y,
	int width, int height)
{
	return &empty;
}
#else
void* tc_x11pfillNew(struct tc_fb *fb, char *data, int x, int y,
	int width, int height)
{
	tc_dbg_name(func)
	struct tc_x11 *x11 = fb->x11;
	struct tc_fillHandle *cache;
	XGCValues gcv;
	XImage *i;

	time_add("PFILL", TI_DRAW_START);
	tclient_dbg(1, ("%s: creating new pixmap fill update\n", func));

	cache = malloc(sizeof(*cache));
	if(!cache) {
		tclient_msg("Error: no memory for new pfill update: %s\n",
			strerror(errno));
		return NULL;
	}

	tclient_dbg(1, ("%s: creating XImage\n", func));
	i = XCreateImage(x11->disp, x11->visual, x11->w_depth, ZPixmap,
		0, (char*)data, width, height, 32, 0);
	if(!i) {
		tclient_msg("Error: Couldn't create new pfill update\n");
		free(cache);
		return NULL;
	}
	i->byte_order = x11->img->byte_order;

	tclient_dbg(1, ("%s: creating pixmap\n", func));
	cache->pix = XCreatePixmap(x11->disp, x11->Main, width, height,
		x11->w_depth);
	XPutImage(x11->disp, cache->pix, x11->theGC, i, 0, 0, 0, 0, 
		width, height);

	tclient_dbg(1, ("%s: creating gc\n", func));
	gcv.fill_style = FillTiled;
	gcv.tile = cache->pix;
	gcv.ts_x_origin = x;
	gcv.ts_y_origin = y;
	cache->gc = XCreateGC(x11->disp, x11->Main, 
		(GCFillStyle | GCTile | GCTileStipXOrigin | GCTileStipYOrigin),
		&gcv);

	tclient_dbg(1, ("%s: almost there, cleaning up\n", func)); 
	/* don't free the data here to avoid problems with caller */
	i->data = NULL;	
	XDestroyImage(i);

	tclient_dbg(1, ("%s: finished\n", func));
	time_add("PFILL", TI_DRAW_END);
	return cache;
}
#endif /* TC_DONT_UPDATE_X */

int tc_x11pfillFill(struct tc_fb *fb, void *handle, thinc_rectRegion *r,
	unsigned char resized)
{
	tc_dbg_name(func)
	struct tc_x11 *x11 = fb->x11;
	struct tc_fillHandle *cache = handle;
	XRectangle rszd;
	instr_vars(s, e);

	time_add("PFILL", TI_DRAW_START);

	tclient_dbg(1, ("%s: filling rectangle\n", func));

	if(resized) {
		rszd.x = r->x;
		rszd.y = r->y;
		rszd.width = r->width;
		rszd.height = r->height;
		tc_resize_scale_rect(fb, &rszd);
		if(rszd.height == 0) rszd.height = 1;
		tclient_dbg(1, ("%s: filling rectangle x:%d y:%d w:%d h:%d\n", 
			func, rszd.x, rszd.y, rszd.width, rszd.height));
		XFillRectangle(x11->disp, x11->resizedP, cache->gc, rszd.x, 
			rszd.y, rszd.width, rszd.height);
		tc_rgn_add_resized_coords_pixmap(fb, rszd.x, rszd.y, 
			rszd.width, rszd.height);
	} else {
		tclient_dbg(1, ("%s: filling rectangle x:%d y:%d w:%d h:%d\n", 
			func, r->x, r->y, r->width, r->height));

		instr_add(fb, pfill_pixels, r->width*r->height);
		XFillRectangle(x11->disp, x11->DesktopP, cache->gc, r->x, 
			r->y, r->width, r->height);
		tc_rgn_add_normal_coords(fb, r->x, r->y, r->width, r->height, 
			T_TRUE);

	}

	tclient_dbg(1, ("%s: finished\n", func));

	time_add("PFILL", TI_DRAW_END);
	return TC_OK;
}

void tc_x11fillDestroyWrap(void *f, void *p)
{
	if(p) {
		struct tc_fb *fb = f;
		struct tc_fillHandle *cache = p;

		XFreePixmap(fb->x11->disp, cache->pix);
		XFreeGC(fb->x11->disp, cache->gc);
		free(cache);
	}
}

void* tc_x11bilevelNew(struct tc_fb *fb, char *bitmap, int x, int y,
	int width, int height, unsigned long fg, unsigned long bg)
{
	tc_dbg_name(func)
	struct tc_x11 *x11 = fb->x11;
	struct tc_fillHandle *handle;
	XGCValues gcv;

	time_add("BILEVEL", TI_DRAW_START);

	tclient_dbg(1, ("%s: creating new bilevel update\n", func));

	handle = malloc(sizeof(*handle));
	if(!handle) {
		tclient_msg("Error: no memory for new bilevel update: %s\n",
			strerror(errno));
		return NULL;
	}

	tclient_dbg(1, ("%s: creating pixmap\n", func));
	handle->pix = XCreateBitmapFromData(x11->disp, x11->Main, 
		(char*)bitmap, width, height);

	tclient_dbg(1, ("%s: creating gc\n", func));
	gcv.fill_style = FillOpaqueStippled;
	gcv.foreground = fg;
	gcv.background = bg;
	gcv.stipple = handle->pix;
	gcv.ts_x_origin = x;
	gcv.ts_y_origin = y;
	handle->gc = XCreateGC(x11->disp, x11->Main, (GCFillStyle | GCStipple 
		| GCForeground | GCBackground | GCTileStipXOrigin 
		| GCTileStipYOrigin), &gcv);

	tclient_dbg(1, ("%s: finished\n", func));
	time_add("BILEVEL", TI_DRAW_END);
	return handle;
}

void* tc_x11glyphNew(struct tc_fb *fb, char *bitmap, int x, int y,
	int width, int height, unsigned long color)
{
	tc_dbg_name(func)
	struct tc_x11 *x11 = fb->x11;
	struct tc_fillHandle *handle;
	XGCValues gcv;

	time_add("GLYPH", TI_DRAW_START);

	tclient_dbg(1, ("%s: creating new glyph update\n", func));

	handle = malloc(sizeof(*handle));
	if(!handle) {
		tclient_msg("Error: no memory for new glyph update: %s\n",
			strerror(errno));
		return NULL;
	}

	tclient_dbg(1, ("%s: creating pixmap\n", func));
	handle->pix = XCreateBitmapFromData(x11->disp, x11->Main, 
		(char*)bitmap, width, height);

	tclient_dbg(1, ("%s: creating gc\n", func));
	gcv.fill_style = FillStippled;
	gcv.foreground = color;
	gcv.stipple = handle->pix;
	gcv.ts_x_origin = x;
	gcv.ts_y_origin = y;
	handle->gc = XCreateGC(x11->disp, x11->Main, (GCFillStyle | GCStipple 
		| GCForeground | GCTileStipXOrigin | GCTileStipYOrigin), 
		&gcv);

	tclient_dbg(1, ("%s: finished\n", func));
	time_add("GLYPH", TI_DRAW_END);
	return handle;
}

int tc_x11bitmapFill(struct tc_fb *fb, void *p, thinc_rectRegion *r)
{
	struct tc_x11 *x11 = fb->x11;
	struct tc_fillHandle *handle = p;

	time_add("BITMAP", TI_DRAW_START);
	XFillRectangle(x11->disp, x11->DesktopP, handle->gc, r->x, r->y, 
		r->width, r->height);
	tc_rgn_add_normal_coords(fb, r->x, r->y, r->width, 
		r->height, T_FALSE);

	time_add("BITMAP", TI_DRAW_END);
	return TC_OK;
}

int tc_x11cursorChange(struct tc_fb *fb)
{
	doUninitCursor(fb->x11);

	return doCreateCursor(fb);
}

int tc_x11cursorMove(struct tc_fb *fb, int x, int y)
{
	tc_dbg_name(func)

	/* send it */
	tclient_dbg(1, ("%s: Moving mouse to x %d y %d\n", func, x, y));

	/* FIXME: Need to resize the coordinates */
#if 0
	XWarpPointer(fb->x11->disp, 0, fb->x11->Main, 0, 0, 0, 0, x, y);
#endif
	/* XWarpPointer issues a mouse event each time it is invoked.
	 * Since we do not wish to send events issued by the server
	 * back to the server again, we track these events by 
	 * incrementing this counter. See also tc_MotionEv in this
	 * file on how this variable is tested. */
#if 0
	fb->move_count++;
#endif
	return TC_OK;
}

void tc_x11cursorRecolor(struct tc_fb *fb, int fgv, int bgv)
{
	struct tc_x11 *x11 = fb->x11;
	tc_dbg_name(func)
	XColor fg, bg;

	tclient_dbg(3, ("%s: fg %d, bg %d\n", func, fgv, bgv));

	/* The color is sent to us in RGB, so we only get 8 bits per
	 * color. However, XCreatePixmapCursor (or XColor) has each channel 
	 * being 16 bits (why????), so we need to convert it to 0-65535 
	 * scale */
	/* TODO: Make sure this actually works on 8bpp x11->displays */
	fg.pixel = 0;
	fg.red = (fgv & 0xFF0000) >> 16;
	fg.green = (fgv & 0xFF00) >> 8;
	fg.blue = (fgv & 0xFF);

	fg.red = fg.red * 65535 / 255;
	fg.green = fg.green * 65535 / 255;
	fg.blue = fg.blue * 65535 / 255;

	bg.pixel = 0;
	bg.red = (bgv & 0xFF0000) >> 16;
	bg.green = (bgv & 0xFF00) >> 8;
	bg.blue = (bgv & 0xFF);

	bg.red = bg.red * 65535 / 255;
	bg.green = bg.green * 65535 / 255;
	bg.blue = bg.blue * 65535 / 255;

	XRecolorCursor(x11->disp, x11->theCursor, &fg, &bg);

	tclient_dbg(3, ("%s: finished\n", func));
}

/*******************
 * Handle X Events *
 *******************/

int tc_hasevents(struct tc_fb *fb)
{
	struct tc_x11 *x11 = fb->x11;
#ifndef NO_X
	int events = XPending(x11->disp);
	if(events == 0)
		return T_FALSE;
	else
		return T_TRUE;
#else
	return T_FALSE;
#endif	/* ! NO_X */
}

/* in charge of managing X events. Mostly calls the appropiate function
 * for the event received. events_pending is how many events should
 * be processed. Normally should be equal to the amount of events
 * already in the queue when the function is called. Returns TC_OK on success, 
 * TC_ERROR on error */
int tc_handleEvents(struct tc_fb *fb)
{
	struct tc_x11 *x11 = fb->x11;
	int res=TC_OK;
	XEvent e;
	instr_vars(start, end);
	time_add_extra("X", TI_START);

	do {
		XNextEvent(x11->disp, &e);
		switch(e.type) {
			case MappingNotify: /* keyb remapped */
				XRefreshKeyboardMapping((XMappingEvent*)&e);
				res = TC_OK;
				break;
			case Expose:
				time_add("EXPOSE", TI_START);
				instr_start(start);
				tc_ExposeEv(fb, &e);
				instr_end(fb, start, end, x_expose);
				time_add("EXPOSE", TI_END);
				res = TC_OK;
				break;
			case ButtonPress:
				time_add("INPUT", TI_START);
				res = tc_ButtonEv(fb, &e, T_DOWN);
				time_add("INPUT", TI_END);
				break;
			case ButtonRelease:
				time_add("INPUT", TI_START);
				res = tc_ButtonEv(fb, &e, T_UP);
				time_add("INPUT", TI_END);
				break;
			case MotionNotify:
				res = tc_MotionEv(fb, &e);
				if(res == TC_ERROR) break;
				break;
			case KeyPress:
				time_add("KEY_DOWN", TI_START);
				res = tc_KeyEv(fb, &e, T_DOWN);
				time_add("KEY_DOWN", TI_END);
				break;
			case KeyRelease:
				time_add("KEY_UP", TI_START);
				res = tc_KeyEv(fb, &e, T_UP);
				time_add("KEY_UP", TI_END);
				break;
			case ConfigureNotify:
				res = tc_ResizeEv(fb, &e);
				if(res == TC_ERROR) break;
				break;
			case EnterNotify:
				x11->mouse_in = T_TRUE;
				if(fb->fullscreen) {
					XSetInputFocus(x11->disp, x11->Main,
						RevertToPointerRoot, 
						CurrentTime);
				}
				if(x11->focused) {
					XGrabKeyboard(x11->disp, x11->Main, 
						True, GrabModeAsync, 
						GrabModeAsync, CurrentTime);
				}
				break;
			case LeaveNotify:
				x11->mouse_in = T_FALSE;
				XUngrabKeyboard(x11->disp, CurrentTime);
				break;
			case FocusIn:
				if (e.xfocus.mode == NotifyGrab)
					break;
				tclient_msg("have focus, grabbing keyb and "
					"not letting go\n");
				x11->focused = T_TRUE;
				if(x11->mouse_in) {
					XGrabKeyboard(x11->disp, x11->Main, 
						True, GrabModeAsync, 
						GrabModeAsync, CurrentTime);
				}
				break;
			case FocusOut:
				if (e.xfocus.mode == NotifyUngrab)
					break;
				tclient_msg("lost focus, ungrabbing keyb. "
					"good bye old friend\n");
				x11->focused = T_FALSE;
				if (e.xfocus.mode == NotifyWhileGrabbed)
					XUngrabKeyboard(x11->disp, CurrentTime);
				/* hack so keys don't get stuck */
				/* TODO: check return value */
				release_all_keys(fb, first_key_down);
				first_key_down = NULL;
				break;
			default:
				tclient_dbg(3, ("Unknonw event received: %d\n",
					e.type));
				res = TC_OK;
		}
		if(res == TC_ERROR) {
			tclient_msg("Error handling event\n");
			return TC_ERROR;
		}
	} while (XPending(x11->disp));

	time_add_extra("X", TI_END);

	return TC_OK;
}

/* handles expose events. */
static void tc_ExposeEv(struct tc_fb *fb, XEvent *e)
{
	struct tc_x11 *x11 = fb->x11;
	XRectangle rect;
	rect.x = e->xexpose.x;
	rect.y = e->xexpose.y;
	rect.width = e->xexpose.width;
	rect.height = e->xexpose.height;
	XUnionRectWithRegion(&rect, x11->exposeRgn, x11->exposeRgn);
	x11->exposed = T_TRUE;

	tclient_dbg(1, ("%s: added rectangle (%d,%d):(%dx%d)\n", __func__,
		rect.x, rect.y, rect.width, rect.height));
}

/* processes all motion events in the queue and sends them coallesced as
 * one to the server. Returns TC_ERROR on error */
static int tc_MotionEv(struct tc_fb *fb, XEvent *e)
{
	struct tc_x11 *x11 = fb->x11;
	tc_dbg_name(func)
	int res, x, y;

	/* Checking to see if the event was issued by the server or the 
	 * client. If move_count is greater than zero, then it must have 
	 * been issued by the server. Discard the event and move on. 
	 * See also tc_x11cursorMove. */
#if 0
	if (fb->move_count > 0) {
		tclient_dbg(1, ("%s: Mouse event for coordinates %d,%d "
			"issued by server. Not moving.\n", func, e->xmotion.x, 
			e->xmotion.y));
		fb->move_count--;
		return TC_OK;
	}
#endif

	/* collapse all the motion events into one */
	res = XEventsQueued(x11->disp, QueuedAfterFlush); 
	while(XCheckTypedWindowEvent(x11->disp, x11->Main, MotionNotify, e))
		;

	tclient_dbg(3, ("%s: Mouse moved to %d,%d\n", func, e->xmotion.x,
		e->xmotion.y));

	x = e->xmotion.x;
	y = e->xmotion.y;
	if(x11->resized) {
		x = tc_resize_unscale_numx(fb, x);
		y = tc_resize_unscale_numy(fb, y);
	} 

#ifdef TC_RECORD
	if(fb->do_record)
		tc_record_mouse_move(&fb->record, x, y);
#endif

	if(tc_sendMotionEv(fb, x, y) == TC_ERROR) {
		tclient_msg("Error: Couldn't send motion event message\n");
		return TC_ERROR;
	}

	return TC_OK;
}

int tc_sendMotionEv(struct tc_fb *fb, int x, int y)
{
	thinc_evMotion m;

	m.x = htons(x);
	m.y = htons(y);

	if(tc_writeMsg(fb, T_EV_MOTION, &m, T_EV_MOTION_SIZE) == TC_ERROR)
		return TC_ERROR;

	return TC_OK;
}

/* processes a button press/release event and sends it to the server.
 * Returns TC_ERROR on error */
static int tc_ButtonEv(struct tc_fb *fb, XEvent *e, int updown)
{
	tc_dbg_name(func)

	tclient_dbg(3, ("%s: ButtonPress at (%d,%d) button %u\n", 
		func, e->xbutton.x,e->xbutton.y, e->xbutton.button));

#ifdef TC_RECORD
	if(fb->do_record)
		tc_record_mouse_click(&fb->record, e->xbutton.button, updown);
#endif

	if(tc_sendButtonEv(fb, e->xbutton.button, updown) == TC_ERROR) {
		tclient_msg("Error: Couldn't send button event message\n");
		return TC_ERROR;
	}

	return TC_OK;
}

int tc_sendButtonEv(struct tc_fb *fb, int button, int updown)
{
	thinc_evButton b;

	b.buttonMask = 127 & button;
	if(updown == T_DOWN)
		b.buttonMask |= 128;

	/* send the event */
	if(tc_writeMsg(fb, T_EV_BUTTON, &b, T_EV_BUTTON_SIZE) == TC_ERROR)
		return TC_ERROR;

	return TC_OK;
}

static int tc_KeyEv(struct tc_fb *fb, XEvent *e, int updown)
{
	struct tc_x11 *x11 = fb->x11;
	tc_dbg_name(func)
	int key;

	key = XKeycodeToKeysym(x11->disp, e->xkey.keycode, 0);

#ifdef TC_RECORD
	if(fb->do_record)
		tc_record_keyb(&fb->record, key, updown);
#endif
	tclient_dbg(3, ("%s: Got key event, key %d %s\n", 
		func,key, (updown == T_UP ? "up" : "down")));

	/* hack so keys don't get stuck */
	if (updown == T_DOWN) {
		/* add to list of pressed keys */
		struct a_key_down *n = malloc(sizeof(*n));
		if (!n) {
			/* not really fatal :P */
			tclient_msg("Error: No memory for new key down\n");
			goto send;
		}
		n->key = key;
		n->next = first_key_down;
		first_key_down = n;
	} else {
		/* key up, remove from list */
		struct a_key_down *k = first_key_down, *p = NULL;

		while(k) {
			if (k->key == key) {
				if (p) {
					p->next = k->next;
				} else {
					first_key_down = k->next;
				}
				free(k);
				break;
			}
			p = k;
			k = k->next;
		}
	}

send:
	if(tc_sendKeyEv(fb, key, updown) == TC_ERROR) {
		tclient_msg("Error: Couldn't send keyboard event message\n");
		return TC_ERROR;
	}

	return TC_OK;
}

int tc_sendKeyEv(struct tc_fb *fb, int key, int updown)
{
	thinc_evKeyb k;

	k.key = htonl(key);
	k.down = updown;

	if(tc_writeMsg(fb, T_EV_KEYB, &k, T_EV_KEYB_SIZE) == TC_ERROR) 
		return TC_ERROR;

	return TC_OK;
}

static int tc_ResizeEv(struct tc_fb *fb, XEvent *e)
{
	struct tc_x11 *x11 = fb->x11;

	while(XCheckTypedWindowEvent(x11->disp, x11->Main,ConfigureNotify,e))
		;
	tc_resize_resize(fb, e->xconfigure.width, e->xconfigure.height);

	return TC_OK;
}

/***********
 * Regions *
 ***********/
static int tc_rgn_init(struct tc_fb *fb)
{
	struct tc_x11 *x11 = fb->x11;

	x11->damageRgn = XCreateRegion();
	x11->exposeRgn = XCreateRegion();
	x11->resizedRgn = XCreateRegion();
	if(	!x11->damageRgn || !x11->exposeRgn
		|| !x11->resizedRgn
		)
		return TC_ERROR;

	x11->exposed = T_FALSE;
	x11->updated = T_FALSE;

	return TC_OK;
}

static void tc_rgn_end(struct tc_fb *fb)
{
	struct tc_x11 *x11 = fb->x11;

	x11->damageRgn ? XDestroyRegion(x11->damageRgn) : 0;
	x11->exposeRgn ? XDestroyRegion(x11->exposeRgn) : 0;
	x11->resizedRgn ? XDestroyRegion(x11->resizedRgn) : 0;
}

/* 
 * add to damageRgn
 * substract from outdatedRgn
 */
/* if update is T_TRUE the rectangle is taken out from the outdated region.
 * If your update has transparent parts (e.g. glyph) pass T_FALSE */
static int tc_rgn_add_normal(struct tc_fb *fb, XRectangle *rect, int update)
{
	struct tc_x11 *x11 = fb->x11;
	XUnionRectWithRegion(rect, x11->damageRgn, x11->damageRgn);

	x11->updated = T_TRUE;
	return TC_OK;
}

static inline int tc_rgn_add_normal_coords(struct tc_fb *fb, int x, int y, 
	int w, int h, int update)
{
	XRectangle rect;
	rect.x = x;
	rect.y = y;
	rect.width = w;
	rect.height = h;
	return tc_rgn_add_normal(fb, &rect, update);
}

/*
 * add rszd to resizedRgn
 */
/* call this function when you put resized content directly into the resized
 * pixmap. call tc_rgn_add_resized when you put the resized content into
 * the Ximage */
static int tc_rgn_add_resized_pixmap(struct tc_fb *fb, XRectangle *rszd) 
{
	/* tell resize layer it shouldn't trust its image anymore */
	tc_resize_img_outdated(fb);

	return tc_rgn_add_resized(fb, rszd);
}

static int tc_rgn_add_resized(struct tc_fb *fb, XRectangle *rszd)
{
	struct tc_x11 *x11 = fb->x11;
	XUnionRectWithRegion(rszd, x11->resizedRgn, x11->resizedRgn);
	x11->updated = T_TRUE;

	return TC_OK;
}

static inline int tc_rgn_add_resized_coords_pixmap(struct tc_fb *fb, int rx, 
	int ry, int rw, int rh)
{
	/* tell resize layer it shouldn't trust its image anymore */
	tc_resize_img_outdated(fb);

	return tc_rgn_add_resized_coords(fb, rx, ry, rw, rh);
}

static inline int tc_rgn_add_resized_coords(struct tc_fb *fb, int rx, int ry, 
	int rw, int rh)
{
	XRectangle rszd;
	rszd.x = rx;
	rszd.y = ry;
	rszd.width = rw;
	rszd.height = rh;
	return tc_rgn_add_resized(fb, &rszd);
}

/*****************************************/
inline void tc_x11_setclip_region(struct tc_fb *fb, Region reg)
{
	tc_xlib_rgn *rgn = (tc_xlib_rgn*)reg;
	tc_x11_setclip(fb, rgn);
}

void tc_x11_setclip(struct tc_fb *fb, tc_xlib_rgn *rgn)
{
	int total;
	XRectangle *rects;

	total = tc_rgn_numrects(rgn);

	tclient_dbg(1, ("%s: have %d rectangles. current size: %dx%d\n", 
		__func__, total, fb->curr_width, fb->curr_height));

	/* TODO: alloca() is supposed to be unsafe */
	rects = alloca(sizeof(XRectangle)*total);
	if(!rects)
		return;

	tc_rgn_convert(tc_rgn_rects(rgn), rects, total);

	tclient_dbg(1, ("%s: converted them\n", __func__));

	XSetClipRectangles(fb->x11->disp, fb->x11->theGC, 
		0, 0, rects, total, Unsorted);

	tclient_dbg(1, ("%s: finished\n", __func__));
}

inline void tc_x11_clearclip(struct tc_fb *fb)
{
	XGCValues gcv;
	tclient_dbg(1, ("%s: clearing clip mask\n", __func__));

	gcv.clip_mask = None;
	XChangeGC(fb->x11->disp, fb->x11->theGC, (GCClipMask), &gcv);
}

/* hack so keys don't get stuck */
static int release_all_keys(struct tc_fb *fb, struct a_key_down *head)
{
	struct a_key_down *p;

	if(!head)
		return TC_OK;

	do {
		if (tc_sendKeyEv(fb, head->key, T_UP) == TC_ERROR)
			return TC_ERROR;
		p = head;
		head = head->next;
		free(p);
	} while(head);

	return TC_OK;
}
