/* Copyright 2003, 2004, 2005, 2006, 2007 Ricardo Baratto */
/* Copyright 2003, 2004, 2005, 2006, 2007 Columbia University */
/* This file is part of THINC.

    THINC is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 2 of the License, or
    (at your option) any later version.

    THINC is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with THINC.  If not, see <http://www.gnu.org/licenses/>.
*/
/*
 * XVideo support for thinc client
 *
 * Author:
 * Ricardo Baratto
 * Network Computing Lab - Columbia University
 *
 * Copyright 2004
 */
/* set tabstop = 8 */

/* How all the video sizes play together:
 * -> width, height: the size of the video as seen by the application. The
 * data sent from the application to the server has this dimensions.
 * -> dst_width, dst_height: the size of the video as the application wants
 * it displayed. This relates to application resize (e.g. full screen)
 * -> curr_width, curr_height: the size of the video as its being displayed
 * on the client. If the client is resized, the video won't be displayed
 * in the same dimensions as the application wanted (dst_*)
 *
 * Notes on operation:
 * -> all the sizes are specified in the initial START message.
 * -> dst_* is changed with a SCALE message
 * -> curr_* is changed when the client's window size changes.
 * -> a RESIZE message changes normal width,height
 *
 * How the video frames are interpreted:
 * -> Data sent by the server will have size width*height. Therefore,
 * every time they change we'll have to re-create the XVideo Image.
 * -> To display, we use width,height as original dimensions, and
 * curr_width, curr_height as the destination dimensions.
 *
 * Why do we need dst_width/dst_height??
 *
 * We also have 2 pairs of coordinates: the application requested ones,
 * and those used by the client when resizing happens. The server doesn't
 * change the coordinates.
 */

/* FIXME TODO:
 * - don't use tclient_msg for errors. use tc_error_msg. 
 * - chk for error and closed in one single if()
 */

#include "thinc_client.h"

#include "tc_x11.h"
#include <X11/Xlib.h>
#include <X11/extensions/XShm.h>
#include <X11/extensions/Xvlib.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>
#include <stdio.h>
#include <netinet/in.h>
#include <sys/types.h>
#include <sys/ipc.h>
#include <sys/shm.h>
#include <sys/poll.h>

static int tc_video_chkserver(struct tc_fb *fb);
static int tc_get_server_fmts(struct tc_fb *fb);
static int tc_send_fmts(struct tc_fb *fb);
static int tc_read_fmts(struct tc_fb *fb, thinc_vidSInit *s);
static int tc_xvid_init(struct tc_video *vid, struct tc_x11 *x11);
static int tc_x11vid_find_img_format(struct tc_video *vid, struct tc_x11 *x11,
		XvAdaptorInfo *adapt, XvImageFormatValues *img);
static int tc_x11vid_find_port(struct tc_video *vid, struct tc_x11 *x11,
		XvAdaptorInfo *adapt, XvImageFormatValues *imgs, int num_imgs);
static int tc_xvid_initport(struct tc_video *vid, struct tc_x11 *x11);
static int tc_xvid_getcolorkey(struct tc_video *vid, struct tc_x11 *x11);
static int tc_video_setup_stream(struct tc_fb *fb, thinc_vidStart *vs);
static void tc_video_destroy_stream(struct tc_fb *fb, 
		struct tc_video_stream *stream);
static void tc_video_destroy_stream_id(struct tc_fb *fb, int id);
static int tc_video_update_stream(struct tc_fb *fb, thinc_header *h,
		thinc_vidNext *next);
#ifndef NO_X
static int tc_video_get_video_normal(struct tc_fb *fb, 
		struct tc_video_stream *stream);
#else
static int tc_video_get_video_normal(struct tc_fb *fb, 
		struct tc_video_stream *stream, int size);
#endif	/* ! NO_X */
#warning "Our PNG code doesn't support compressed video yet. However, compressed video is not really used so feel free to disregard this warning (unless the client crashes :)"
static int tc_video_create_xvimage(struct tc_fb *fb, struct tc_video *vid,
		struct tc_video_stream *stream);
static void tc_video_destroy_xvimage(struct tc_fb *fb, 
		struct tc_video_stream *stream);

/* format dependent functions */
/* YV12 */
static int tc_video_readinfo_yv12(struct tc_fb *fb, 
	struct tc_video_stream *stream);
static int tc_video_init_image_yv12(struct tc_fb *fb, struct tc_video *vid,
	struct tc_video_stream *stream);
static void tc_video_destroy_image_yv12(struct tc_fb *fb, 
	struct tc_video_stream *stream);
/* YUY2 */
static int tc_video_readinfo_yuy2(struct tc_fb *fb, 
	struct tc_video_stream *stream);
static int tc_video_init_image_yuy2(struct tc_fb *fb, struct tc_video *vid,
	struct tc_video_stream *stream);
static void tc_video_destroy_image_yuy2(struct tc_fb *fb, 
	struct tc_video_stream *stream);
/* UYVY */
static int tc_video_readinfo_uyvy(struct tc_fb *fb, 
	struct tc_video_stream *stream);
static int tc_video_init_image_uyvy(struct tc_fb *fb, struct tc_video *vid,
	struct tc_video_stream *stream);
static void tc_video_destroy_image_uyvy(struct tc_fb *fb, 
	struct tc_video_stream *stream);

int tc_video_init(struct tc_fb *fb)
{
	tc_dbg_name(func)
	int err;

	tclient_dbg(1, ("%s: initializing video support\n", func));

	/* first thing, check that the server supports it */
	err = tc_video_chkserver(fb);
	if(err == TC_ERROR) {
		tclient_dbg(1, ("Error: %s: Couldn't check if server supports "
			"video\n", func));
		return TC_ERROR;
	} else if(err == T_NO) {
		tclient_msg("Info: Server does not support video\n");
		return TC_NOVIDEO;
	}

	/* calloc should make everything 0 so no need to initialize
	 * each stream's struct */
	fb->vid.streams = calloc(TC_MAX_STREAMS, 
		sizeof(struct tc_video_stream));
	if(!fb->vid.streams) {
		tclient_dbg(1, ("Error: %s: Couldn't allocate memory for video "
			"streams\n", func));
		return TC_NOVIDEO;
	}
	fb->vid.num_streams = TC_MAX_STREAMS;

	err = tc_get_server_fmts(fb);
	if(err == TC_ERROR) {
		tclient_dbg(1, ("Error: %s: Couldn't read video formats from "
			"server\n", func));
		free(fb->vid.streams);
		return TC_NOVIDEO;
	}

	tclient_dbg(1, ("%s: got %d formats from server\n", func,
		fb->vid.num_fmts));
	/* initialize video formats */
	if(tc_xvid_init(&fb->vid, fb->x11) == TC_ERROR) {
		tclient_dbg(1, ("Error: %s: Couldn't initialize XVideo\n", 
			func));
		fb->vid.have_vid = T_FALSE;
		free(fb->vid.fmts);
		free(fb->vid.streams);
		return TC_NOVIDEO;
	}

	/* nice, tell the server which formats we support */
	tclient_dbg(1, ("%s: we now support %d video formats\n", func,
		fb->vid.num_fmts));
	err = tc_send_fmts(fb);
	if(err == TC_ERROR) {
		tclient_dbg(1, ("Error: %s: Couldn't send video formats\n", 
			func));
		fb->vid.have_vid = T_FALSE;
		free(fb->vid.fmts);
		free(fb->vid.streams);
		return TC_NOVIDEO;
	}

	fb->vid.have_vid = T_TRUE;
	tclient_dbg(1, ("%s: finished\n", func));
	return TC_OK;
}

int tc_novideo(struct tc_fb *fb)
{
	tc_dbg_name(func)
	int err;
	thinc_reqNoVideo v;

	tclient_dbg(2, ("%s: telling server we don't support video\n", func));

	t_req_set_type(&v, T_REQ_NOVIDEO);

	v2_sendReq(err, fb, &v, t_req_size(T_REQ_NOVIDEO));

	tclient_dbg(2, ("%s: finished\n", func));
	return TC_OK;
}

/* returns T_YES or T_NO (or T_ERROR) if server supports video or not */
static int tc_video_chkserver(struct tc_fb *fb)
{
	tc_dbg_name(func)
	int err, reply;
	thinc_header h;
	thinc_reqVideo v;
	thinc_replyGen g;

	tclient_dbg(2, ("%s: sending request for video\n", func));

	t_req_set_type(&v, T_REQ_VIDEO);

	v2_sendReq(err, fb, &v, t_req_size(T_REQ_VIDEO));

	tclient_dbg(2, ("%s: reading server response\n", func));

	v2_readHeader(err, fb, h, T_SERV_REPLY);
	v2_readMsg(err, fb, &g, h.length);

	reply = t_reply_read_type(&g);
	if(reply == T_REPLY_UNKNOWN || reply == T_REPLY_NOTOK) {
		tclient_dbg(2, ("%s: server does not support video\n",
			func));
		err = T_NO;
	} else if(reply == T_REPLY_OK) {
		tclient_dbg(2, ("%s: server supports video, initializing\n",
			func));
		err = T_YES;
	} else {
		tclient_dbg(1, ("Error: %s: Unknown reply %d received from "
			"server\n", func, reply));
		return TC_ERROR;
	}

	tclient_dbg(2, ("%s: finished\n", func));
	return err;
}

static int tc_get_server_fmts(struct tc_fb *fb)
{
	tc_dbg_name(func)
	int err, reply;
	thinc_header h;
	thinc_reqVideoServFmts fr;
	thinc_replyVideoServFmts fy;
	thinc_replyGen g;

	tclient_dbg(2, ("%s: sending request for video formats\n", func));

	t_req_set_type(&fr, T_REQ_VIDEO_SERV_FMTS);
	v2_sendReq(err, fb, &fr, t_req_size(T_REQ_VIDEO_SERV_FMTS));

	tclient_dbg(2, ("%s: reading video formats\n", func));

	v2_readHeader(err, fb, h, T_SERV_REPLY);
	v2_readMsg(err, fb, &g, h.length);

	v2_checkReply(reply, &g, T_REQ_VIDEO_SERV_FMTS);
	fy = v2_getReply(g, video_serv_fmts);

	err = tc_read_fmts(fb, &fy.fmts);
	if(err == TC_ERROR) {
		tclient_dbg(1, ("Error: %s: Couldn't read video formats\n",
			func));
		return TC_ERROR;
	}

	tclient_dbg(2, ("%s: finished\n", func));
	return TC_OK;
}

static int tc_send_fmts(struct tc_fb *fb)
{
	tc_dbg_name(func)
	thinc_reqVideoClientFmts fr;
	thinc_header h;
	thinc_replyGen g;
	int err, reply;

	t_req_set_type(&fr, T_REQ_VIDEO_CLIENT_FMTS);

	fr.fmts.numformats = htonl(fb->vid.num_fmts);
	v2_sendReq(err, fb, &fr, t_req_size(T_REQ_VIDEO_CLIENT_FMTS));

	err = writeData(fb, fb->vid.fmts, 
			fb->vid.num_fmts * sizeof(thinc_vidFormat));
	if(err == TC_ERROR) {
		tclient_dbg(1, ("Error: %s: Couldn't send VIDEO response "
			"formats to server\n", func));
		return TC_ERROR;
	}
	tclient_dbg(2, ("%s: reading server response\n", func));

	v2_readHeader(err, fb, h, T_SERV_REPLY);
	v2_readMsg(err, fb, &g, h.length);

	reply = t_reply_read_type(&g);
	if(reply == T_REPLY_UNKNOWN || reply == T_REPLY_NOTOK) {
		tclient_dbg(2, ("%s: server does not support video\n",
			func));
		return TC_ERROR;
	} else if(reply == T_REPLY_OK) {
		tclient_dbg(2, ("%s: all good, server likes our formats\n",
			func));
		return TC_OK;
	} else {
		tclient_dbg(1, ("Error: %s: Unknown reply %d received from "
			"server\n", func, reply));
		return TC_ERROR;
	}
}

static int tc_read_fmts(struct tc_fb *fb, thinc_vidSInit *s)
{
	tc_dbg_name(func)
	struct tc_video* vid = &fb->vid;
	int err, i;

	tclient_dbg(1, ("%s: reading video formats\n", func));

	vid->num_fmts = ntohl(s->numformats);
	vid->fmts = calloc(sizeof(thinc_vidFormat), vid->num_fmts);
	if(!vid->fmts) {
		tclient_dbg(1, ("Error: %s: Couldn't allocate memory for "
			"video formats: %s\n", func, strerror(errno)));
		return TC_ERROR;
	}

	err = readData(fb, vid->fmts, vid->num_fmts * sizeof(thinc_vidFormat), 
			T_TRUE);
	if(err == TC_ERROR) {
		tclient_dbg(1, ("Error: %s: Couldn't read video formats\n", 
			func));
		free(vid->fmts);
		return TC_ERROR;
	}
	for(i=0; i<vid->num_fmts; ++i) {
		vid->fmts[i].id = ntohl(vid->fmts[i].id);
	}

	tclient_dbg(1, ("%s: finished, read %d formats\n",func,vid->num_fmts));
	return TC_OK;
}

void tc_video_shutdown(struct tc_fb *fb)
{
	if(!fb->vid.have_vid)
		return;

	free(fb->vid.fmts);
	free(fb->vid.streams);
}

/* TODO: Should check all video formats, and see which ones we can
 * support. Right now if we find one we can use we return happy */
static int tc_xvid_init(struct tc_video *vid, struct tc_x11 *x11)
{
	tc_dbg_name(func)
	unsigned int xv_ver, xv_rev, xv_req_base, xv_event_base, xv_error_base;
	int port_id;

	tclient_dbg(1, ("%s: initializing XVideo extension\n", func));

	/* see if we have XVideo */
#ifndef NO_X
	if(XvQueryExtension(x11->disp, &xv_ver, &xv_rev, &xv_req_base, 
		&xv_event_base, &xv_error_base) != Success) {
		tclient_dbg(1, ("Errror: %s: Couldn't find XV Extension\n", 
			func));
		return TC_ERROR;
	} else
		tclient_dbg(1, ("%s: Found XV Extension version %d.%d\n", 
			func, xv_ver, xv_rev));

	/* try to find an xvideo port */
	port_id = tc_xvid_initport(vid, x11);
	if(port_id == TC_ERROR) {
		tclient_dbg(1, ("Error: %s: Couldn't initialize Xv Port\n",
			func));
		return TC_ERROR;
	}

	vid->port_id = port_id;

	/* set color key */
	if(tc_xvid_getcolorkey(vid, x11) == TC_ERROR) {
		tclient_dbg(1, ("Error: %s: Couldn't set color key\n",
			func));
		return TC_ERROR;
	}
#endif	/* ! NO_X */
	tclient_dbg(1, ("%s: finished\n", func));
	return TC_OK;
}

static int tc_xvid_initport(struct tc_video *vid, struct tc_x11 *x11) 
{
	tc_dbg_name(func)
	int i, num_imgs, port_num;
	unsigned int num_adapt;
	XvAdaptorInfo *adapt;
	XvImageFormatValues *imgs;

	if(XvQueryAdaptors(x11->disp, DefaultRootWindow(x11->disp), 
		&num_adapt, &adapt) != Success) {
		tclient_dbg(1, ("Error: %s: Couldn't query XV Adaptors\n",
			func));
		return TC_ERROR;
	}

	tclient_dbg(1, ("%s: Got %d adaptors\n", func, num_adapt));
	port_num = -1;

	/* find one we can use */
	for(i=0; i< num_adapt; ++i) {
		if(adapt[i].type & XvImageMask) {
			/* this one has image support, is it the right 
			 * format?  */
			imgs = XvListImageFormats(x11->disp, adapt[i].base_id,
				&num_imgs);
			tclient_dbg(1, ("%s: Got %d image formats\n", func,
				num_imgs));
			port_num = tc_x11vid_find_port(vid, x11, &adapt[i],
				imgs, num_imgs);
			XFree(imgs);
		}
		if(port_num != -1) break;
	}
	/* FIXME: This doesn't really free all the memory. see valgrind */
	XvFreeAdaptorInfo(adapt);

	tclient_dbg(1, ("%s: finished, got port number %d\n", func, port_num));
	return port_num;
}

static int tc_x11vid_find_port(struct tc_video *vid, struct tc_x11 *x11,
	XvAdaptorInfo *adapt, XvImageFormatValues *imgs, int num_imgs)
{
	int j;
	int port_num = -1;

	for(j=0; j<num_imgs; ++j) {
		port_num = tc_x11vid_find_img_format(vid,x11,adapt,&imgs[j]); 
		if(port_num != -1) break;
	}

	return port_num;
}

/* thanks to mplayer dudes for inspiration for this code. it will set
 * vid->paint_key and vid->ck_gc if we need to paint color key by hand. 
 * returns TC_OK if all goes well */
static int tc_xvid_getcolorkey(struct tc_video *vid, struct tc_x11 *x11) 
{
	tc_dbg_name(func)
	Atom color = None;
	int count, i, ck;
	XvAttribute *attrs;
	XGCValues gcv;

	tclient_dbg(1, ("%s: figuring out color key info\n", func));

	vid->ck_paint = T_FALSE;
	vid->ck_gc = (GC) -1;

	attrs = XvQueryPortAttributes(x11->disp, vid->port_id, &count);
	if(!attrs) {
		tclient_dbg(1, ("%s: no xv attributes, nothing else to do\n",
			func));
		return TC_OK;
	}

	for(i=0; i<count; ++i) {
		if(!strcmp(attrs[i].name, "XV_COLORKEY")) {
			color = XInternAtom(x11->disp, "XV_COLORKEY", False);
			break;
		}
	}

	if(color == None) {
		tclient_dbg(1, ("%s: card doesn't care about colorkey, done\n",
			func));
		XFree(attrs);
		return TC_OK;
	}

	/* key autopainting produces very suboptimal results so we
	 * just ignore it */
	XFree(attrs);
	tclient_dbg(1, ("%s: painting the key by hand\n", func));
	if(XvGetPortAttribute(x11->disp, vid->port_id, color, &ck) != Success) {
		tclient_dbg(1, ("Error: Could not get Xv color key\n"));
		return TC_ERROR;
	}
	vid->ck_paint = T_TRUE;
	tclient_dbg(1, ("%s: creating GC for painting color key\n",
		func));
	gcv.foreground = ck;
	gcv.function = GXcopy;
	vid->ck_gc = XCreateGC(x11->disp, x11->Main, 
		(GCFunction | GCForeground), &gcv);

	tclient_dbg(1, ("%s: finished\n", func));
	return TC_OK;
}

static int tc_x11vid_find_img_format(struct tc_video *vid, struct tc_x11 *x11,
	XvAdaptorInfo *adapt, XvImageFormatValues *img)
{
	tc_dbg_name(func)
	int i,k, port_num = -1;

	tclient_dbg(1, ("%s: checking if we support image format %#x. "
		"We support %d formats\n", func, img->id, vid->num_fmts));

	for(i=0; i<vid->num_fmts; ++i) {
		if(img->id == vid->fmts[i].id) {
			tclient_dbg(1, ("%s: found image format we support: "
				"%#x. format %d\n", func, img->id, i));
			/* nice, our format */
			for(k=0; k< adapt->num_ports; ++k) {
				/* try to grab a port */
				if(XvGrabPort(x11->disp, adapt->base_id + k,
					CurrentTime) ==	Success) {
					port_num = adapt->base_id + k;
					break;
				}
			}
		}
		if(port_num != -1) break;
	}

	tclient_dbg(1, ("%s: finished\n", func));
	return port_num;
}

int tc_handleVideoStart(struct tc_fb *fb, thinc_header *h)
{
	const char *func = __func__;
	thinc_vidStart vs;
	thinc_vidStartOK ok;
	int res;

	tclient_dbg(1, ("%s: reading video start message\n", func));

	res = tc_readMsgNoHdr(fb, T_VIDEO_START, &vs, T_VIDEO_START_SIZE);
	if(res == TC_ERROR) {
		tclient_msg("Error: %s: Couldn't read Video Start message\n",
			func);
		return TC_ERROR;
	} else if(res == TC_CLOSED) {
		tclient_msg("Error: %s: Connection closed while waiting for "
			"video message\n", func);
		return TC_ERROR;
	}

	/* convert to proper byte order */
	vs.id = ntohl(vs.id);
	vs.fmt_id = ntohl(vs.fmt_id);
	vs.x = ntohs(vs.x);
	vs.y = ntohs(vs.y);
	vs.width = ntohs(vs.width);
	vs.height = ntohs(vs.height);
	vs.dst_width = ntohs(vs.dst_width);
	vs.dst_height = ntohs(vs.dst_height);

	tclient_dbg(1, ("%s: got video start message: id %u, fmt %#x "
		"x %u, y %u, width %u, height %u, dst_width %u, "
		"dst_height %u\n", 
		func, vs.id, vs.fmt_id, vs.x, vs.y, vs.width, vs.height,
		vs.dst_width, vs.dst_height));

	/* Set up the Image for the video */
	res = tc_video_setup_stream(fb, &vs);
	if(res == TC_ERROR) {
		tclient_msg("Error: %s: Couldn't set up video stream\n",
			func);
		ok.id = htonl(vs.id);
		ok.ok = T_VIDEO_NOTOK;
		tc_writeMsg(fb, T_VIDEO_STARTOK, &ok, T_VIDEO_STARTOK_SIZE);
		return TC_ERROR;
	}

	tclient_dbg(1, ("%s: got video stream, sending reply %d to server\n",
		func, T_VIDEO_NOTOK));

	ok.id = htonl(vs.id);
	ok.ok = T_VIDEO_OK;
	res = tc_writeMsg(fb, T_VIDEO_STARTOK, &ok, T_VIDEO_STARTOK_SIZE);
	if(res == TC_ERROR) {
		tclient_msg("Error: %s: Couldn't send video reply to "
			"server\n", func);
		tc_video_destroy_stream_id(fb, vs.id);
		return TC_ERROR;
	}

	tclient_dbg(1, ("%s: finished\n", func));
	return TC_OK;
}

static int tc_video_setup_stream(struct tc_fb *fb, thinc_vidStart *vs)
{
	const char *func = __func__;
	struct tc_video *vid = &fb->vid;
	struct tc_video_stream *stream;
	int i, found = T_FALSE, err;

	tclient_dbg(1, ("%s: setting up video stream %d\n", func, vs->id));

	for(i=0; i<vid->num_fmts; ++i) {
		if(vid->fmts[i].id == vs->fmt_id) {
			found = T_TRUE;
			break;
		}
	}
	if(!found) {
		tclient_msg("Error: %s: Unsupported format %#x for new video "
			"stream\n", func, vs->fmt_id);
		return TC_ERROR;
	}
	/* cool, cool */
	stream = &vid->streams[vs->id];
	stream->id = vs->id;
	stream->curr_x = stream->x = vs->x;
	stream->curr_y = stream->y = vs->y;
	stream->width = vs->width;
	stream->height = vs->height;
	stream->format = vs->fmt_id;
	stream->curr_width = stream->dst_width = vs->dst_width;
	stream->curr_height = stream->dst_height = vs->dst_height;

	switch(stream->format) {
		case T_FOURCC_YV12:
			err = tc_video_readinfo_yv12(fb, stream);
			break;
		case T_FOURCC_YUY2:
			err = tc_video_readinfo_yuy2(fb, stream);
			break;
		case T_FOURCC_UYVY:
			err = tc_video_readinfo_uyvy(fb, stream);
			break;
		default:
			tc_error_msg("Unknown video format %#x received\n",
				stream->format);
			return TC_ERROR;
	}
	if(err == TC_ERROR) {
		tclient_dbg(1, ("Error: %s: Couldn't initialize format "
			"dependent video data\n", func));
		return TC_ERROR;
	}

	if(tc_video_create_xvimage(fb, vid, stream) == TC_ERROR) {
		tclient_msg("Error: %s: Couldn't create XV Image\n", func);
		return TC_ERROR;
	}
#ifndef NO_X
	tclient_dbg(1, ("%s: got XvImage!! its size is %d bytes\n",
		func, stream->image->data_size));
#endif	/* ! NO_X */

	if(fb->x11->resized) {
		tc_resize_video(fb, stream->id);
		tc_resize_move_video(fb, stream->id);
	}

	tclient_dbg(1, ("%s: finished\n", func));
	return TC_OK;
}

static int tc_video_create_xvimage(struct tc_fb *fb, struct tc_video *vid,
	struct tc_video_stream *stream)
{
	const char *func = __func__;
	int err;

	tclient_dbg(1, ("%s: creating xvimage of size %ux%u\n", func,
		stream->width, stream->height));
#ifndef NO_X
	if(fb->x11->use_shm) {
		stream->image = XvShmCreateImage(fb->x11->disp, vid->port_id, 
			stream->format, NULL, stream->width, stream->height, 
			&stream->shminfo);
	} else {
		stream->image = XvCreateImage(fb->x11->disp, vid->port_id, 
			stream->format, NULL, stream->width, stream->height);
	}
	if(!stream->image) {
		tclient_msg("Error: %s: Couldn't create XvImage for video "
			"stream\n", func);
		return TC_ERROR;
	}

	if(fb->x11->use_shm) {
		stream->shminfo.shmid = shmget(IPC_PRIVATE, 
			stream->image->data_size, IPC_CREAT|0777);
		stream->shminfo.shmaddr = stream->image->data = 
			shmat(stream->shminfo.shmid, 0, 0);
		stream->shminfo.readOnly = False;
		if(!XShmAttach(fb->x11->disp, &stream->shminfo)) {
			tclient_msg("Error: %s: Couldn't attach ShmImage for "
				"video stream\n", func);
			shmdt(stream->shminfo.shmaddr);
			shmctl(stream->shminfo.shmid, IPC_RMID, 0);
			XFree(stream->image);
			return TC_ERROR;
		}
	} else {
		stream->image->data = malloc(stream->image->data_size);
		if(!stream->image->data) {
			tclient_msg("Error: Cannot allocate memory for "
				"Xv Image\n");
			XFree(stream->image);
			return TC_ERROR;
		}
	}
#endif	/* ! NO_X */

	/* HACK ALERT HACK ALERT
	 * IMPORTANT: In NO_X, the video_init_* functions are the ones
	 * allocating the image data since they're the ones that know the
	 * size according to the format of the video */
	switch(stream->format) {
		case T_FOURCC_YV12:
			err = tc_video_init_image_yv12(fb, vid, stream);
			break;
		case T_FOURCC_YUY2:
			err = tc_video_init_image_yuy2(fb, vid, stream);
			break;
		case T_FOURCC_UYVY:
			err = tc_video_init_image_uyvy(fb, vid, stream);
			break;
		default:
			tc_error_msg("Stream has unknown format %#x\n",
				stream->format);
			return TC_ERROR;
	}
	if(err == TC_ERROR) {
		tc_error_msg("Cannot initialize video stream\n");
#ifndef NO_X
		if(fb->x11->use_shm) {
			shmdt(stream->shminfo.shmaddr);
			shmctl(stream->shminfo.shmid, IPC_RMID, 0);
		}
		XFree(stream->image);
#endif	/* ! NO_X */
		return TC_ERROR;
	}

	tclient_dbg(1, ("%s: finished\n", func));
	return TC_OK;
}

static void tc_video_destroy_stream(struct tc_fb *fb, 
		struct tc_video_stream *stream)
{
	tc_video_destroy_xvimage(fb, stream);
}

static void tc_video_destroy_xvimage(struct tc_fb *fb, 
		struct tc_video_stream *stream)
{
#ifndef NO_X
	if(fb->x11->use_shm) {
		shmdt(stream->shminfo.shmaddr);
		shmctl(stream->shminfo.shmid, IPC_RMID, 0);
	}
	XFree(stream->image);
#else
	free(stream->image);
#endif	/* ! NO_X */

	stream->image = NULL;

	switch(stream->format) {
		case T_FOURCC_YV12:
			tc_video_destroy_image_yv12(fb, stream);
			break;
		case T_FOURCC_YUY2:
			tc_video_destroy_image_yuy2(fb, stream);
			break;
		case T_FOURCC_UYVY:
			tc_video_destroy_image_uyvy(fb, stream);
			break;
		default:
			return;
	}
}

static void tc_video_destroy_stream_id(struct tc_fb *fb, int id)
{
	struct tc_video_stream *stream;
	if(id >= fb->vid.num_streams)
		return;
	stream = &fb->vid.streams[id];
	if(!stream->image)
		return;
	tc_video_destroy_stream(fb, stream);
}

int tc_handleVideoEnd(struct tc_fb *fb, thinc_header *h)
{
	const char *func = __func__;
	thinc_vidEnd end;
	int res;
	struct tc_video_stream *stream;
	struct tc_video *vid = &fb->vid;

	tclient_dbg(1, ("%s: reading video end message\n", func));
	res = tc_readMsgNoHdr(fb, T_VIDEO_END, &end, T_VIDEO_END_SIZE);
	if(res == TC_ERROR) {
		tclient_msg("Error: %s: Couldn't read Video End message\n",
			func);
		return TC_ERROR;
	} else if(res == TC_CLOSED) {
		tclient_msg("Error: %s: Connection closed while waiting for "
			"video message\n", func);
		return TC_ERROR;
	}

	end.id = ntohl(end.id);
	if(end.id >= vid->num_streams) {
		tclient_msg("Error: %s: Invalid stream %d received\n",
			func, end.id);
		return TC_ERROR;
	}
	stream = &vid->streams[end.id];
	if(!stream->image) {
		tclient_msg("Error: %s: stream %d hasn't been initialized\n",
			func, end.id);
		return TC_ERROR;
	}

	tclient_dbg(1, ("%s: ending video stream %d\n", func, stream->id));
	tc_video_destroy_stream(fb, stream);

	/* :HACK: For synchronization purposes. If this stream is closed, then
	 * we want to refresh the timestamp and prevent sound from dropping when
	 * a new stream is played again. */
	stream->timestamp = 0;
	stream->doWaitSync = T_FALSE;

	tclient_dbg(1, ("%s: finished\n", func));
	return TC_OK;
}

int tc_handleVideoNext(struct tc_fb *fb, thinc_header *h)
{
	const char *func = __func__;
	thinc_vidNext next;
	struct tc_video *vid = &fb->vid;
	struct tc_video_stream *stream;
	unsigned int stamp;
	int res;
	tclient_dbg(1, ("%s: reading video next message\n", func));

	res = tc_readMsgNoHdr(fb, T_VIDEO_NEXT, &next, T_VIDEO_NEXT_SIZE);

	gettime_us(stamp);

	if(res == TC_ERROR) {
		tclient_msg("Error: %s: Couldn't read Video Next message\n",
			func);
		return TC_ERROR;
	} else if(res == TC_CLOSED) {
		tclient_msg("Error: %s: Connection closed while waiting for "
			"video message\n", func);
		return TC_ERROR;
	}

	next.id = ntohl(next.id);
	next.size = ntohl(next.size);
	stream = &vid->streams[next.id];
	stream->instr_stamps.recvstamp = stamp;
	stream->timestamp = ntohl(next.timestamp);

	res = tc_video_update_stream(fb, h, &next);
	if(res == TC_ERROR) {
		tclient_msg("Error: %s: Couldn't update video stream\n",
			func);
		return TC_ERROR;
	}

	tclient_dbg(1, ("%s: finished\n", func));
	return TC_OK;
}

static int tc_video_update_stream(struct tc_fb *fb, thinc_header *h,
		thinc_vidNext *next)
{
	const char *func = __func__;
	struct tc_video *vid = &fb->vid;
	struct tc_video_stream *stream;
	int res, id, size;

	id = next->id;
	size = next->size;

	tclient_dbg(1, ("%s: reading video frame for stream %d with size %d "
		"bytes\n", func, id, size));

	stream = &vid->streams[id];
#ifndef NO_X
	if(!stream->image) {
		tclient_msg("Error: %s: Stream hasn't been initialized\n",
			func);
		return TC_ERROR;
	}

	if(stream->image->data_size != size) {
		tclient_msg("%s: AGGGGGGGGGHHH, local size %d doesn't "
			"correspond to message size %d what's up??\n", func,
			stream->image->data_size, size);
		return TC_ERROR;
	}
#endif	/* ! NO_X */

	/* read the data */
	/* FIXME: If video is compressed there will be a flag
	 * in the START message, not on every update */
	if(h->flags & T_FB_RUPDATE_COMPRESSED) {
		/* reads compressed video and uncompresses it into 
		 * stream->image */
		tclient_msg("Error: Got compressed video, we "
			"have no idea how to handle it :'(\n");
		return TC_ERROR;
	} else {
#ifndef NO_X
		res = tc_video_get_video_normal(fb, stream);
#else
		res = tc_video_get_video_normal(fb, stream, size);
#endif	/* ! NO_X */
	}

	if(res == TC_ERROR) {
		tclient_msg("Error: %s: Couldn't read video data\n", func);
		return TC_ERROR;
	}

	tclient_dbg(1, ("%s: got data, updating screen\n", func));
	gettime_us(stream->instr_stamps.playstamp);
	tc_timestamp(("Video "
		"%-u\t"
		"%-u\t"
		"%-u\t"
		"%-u\n"
		,stream->timestamp 
		,stream->instr_stamps.recvstamp
		,stream->instr_stamps.playstamp
		,next->size
	));

#ifdef TC_USESYNC
	res = tc_doVideoSync(fb, stream);
	if (res != TC_OK) {
		/* pause video */
		if (tc_waitSyncVideo(fb, stream, -res) != TC_OK) {
			tclient_msg("Error: problem occurred while "
				"pausing/sync'ing video.\n");
			return TC_ERROR;
		}
	}
#endif

#ifndef NO_X
	if(vid->ck_paint) {
		XFillRectangle(fb->x11->disp, fb->x11->Main, vid->ck_gc, 
			stream->curr_x, stream->curr_y, stream->curr_width,
			stream->curr_height);
	}
	if(fb->x11->use_shm)
		/* TODO: Need to wait until server is finished drawing
		 * (wait for completion event) */
		XvShmPutImage(fb->x11->disp, vid->port_id, fb->x11->Main, 
			fb->x11->theGC, stream->image, 0, 0, stream->width, 
			stream->height, stream->curr_x, stream->curr_y, 
			stream->curr_width, stream->curr_height, True);
	else 
		XvPutImage(fb->x11->disp, vid->port_id, fb->x11->Main, 
			fb->x11->theGC, stream->image, 0, 0, stream->width, 
			stream->height, stream->curr_x, stream->curr_y, 
			stream->curr_width, stream->curr_height);

#endif	/* ! NO_X */

	XFlush(fb->x11->disp);

	tclient_dbg(1, ("%s: finished\n", func));
	return TC_OK;
}

#ifndef NO_X
static int tc_video_get_video_normal(struct tc_fb *fb, 
	struct tc_video_stream *stream)
#else
static int tc_video_get_video_normal(struct tc_fb *fb, 
	struct tc_video_stream *stream, int size)
#endif	/* ! NO_X */
{
	int res;

#ifndef NO_X
	res = readData(fb, stream->image->data, stream->image->data_size, 
		T_TRUE);
#else
	res = readData(fb, stream->image, size, T_TRUE);
#endif	/* ! NO_X */
	if(res == TC_ERROR) {
		tclient_msg("Error: Couldn't read video data\n");
		return TC_ERROR;
	} else if(res == TC_CLOSED) {
		tclient_msg("Error: Connection closed while waiting "
			"for video data\n");
		return TC_ERROR;
	}
	return TC_OK;
}

int tc_handleVideoMove(struct tc_fb *fb, thinc_header *h)
{
	const char *func = __func__;
	thinc_vidMove move;
	int res;
	struct tc_video_stream *stream;
	struct tc_video *vid = &fb->vid;

	tclient_dbg(1, ("%s: reading video move message\n", func));
	res = tc_readMsgNoHdr(fb, T_VIDEO_MOVE, &move, T_VIDEO_MOVE_SIZE);
	if(res == TC_ERROR) {
		tclient_msg("Error: %s: Couldn't read Video Move message\n",
			func);
		return TC_ERROR;
	} else if(res == TC_CLOSED) {
		tclient_msg("Error: %s: Connection closed while waiting for "
			"video message\n", func);
		return TC_ERROR;
	}

	move.id = ntohl(move.id);
	move.x = ntohs(move.x);
	move.y = ntohs(move.y);

	if(move.id >= vid->num_streams) {
		tclient_msg("Error: %s: Invalid stream id %d received\n", func,
			move.id);
		return TC_ERROR;
	}
	stream = &vid->streams[move.id];
	if(!stream->image) {
		tclient_msg("Error: %s: Stream %d hasn't been initialized\n",
			func, move.id);
		return TC_ERROR;
	}

	tclient_dbg(1, ("%s: moving video stream %d from (%d,%d) to (%d,%d)\n",
		func, stream->id, stream->x, stream->y, move.x, move.y));

	stream->curr_x = stream->x = move.x;
	stream->curr_y = stream->y = move.y;

	if(fb->x11->resized)
		tc_resize_move_video(fb, move.id);

	tclient_dbg(1, ("%s: finished\n", func));
	return TC_OK;
}

int tc_handleVideoScale(struct tc_fb *fb, thinc_header *h)
{
	const char *func = __func__;
	thinc_vidScale scale;
	int res;
	unsigned char resized;
	struct tc_video_stream *stream;
	struct tc_video *vid = &fb->vid;

	tclient_dbg(1, ("%s: reading video scale message\n", func));
	res = tc_readMsgNoHdr(fb, T_VIDEO_SCALE, &scale, T_VIDEO_SCALE_SIZE);
	if(res == TC_ERROR) {
		tclient_msg("Error: %s: Couldn't read Video Scale message\n",
			func);
		return TC_ERROR;
	} else if(res == TC_CLOSED) {
		tclient_msg("Error: %s: Connection closed while waiting for "
				"video message\n", func);
		return TC_ERROR;
	}

	scale.id = ntohl(scale.id);
	scale.dst_width = ntohs(scale.dst_width);
	scale.dst_height = ntohs(scale.dst_height);

	if(scale.id >= vid->num_streams) {
		tclient_msg("Error: %s: Invalid stream id %d received\n", 
				func, scale.id);
		return TC_ERROR;
	}
	stream = &vid->streams[scale.id];
	if(!stream->image) {
		tclient_msg("Error: %s: Stream %d hasn't been initialized\n",
				func, scale.id);
		return TC_ERROR;
	}

	if( 	(stream->curr_width != stream->dst_width) ||
			(stream->curr_height != stream->dst_height) )
		resized = T_TRUE;
	else
		resized = T_FALSE;

	stream->curr_width = stream->dst_width = scale.dst_width;
	stream->curr_height = stream->dst_height = scale.dst_height;

	if(resized)
		tc_resize_video(fb, scale.id);

	tclient_dbg(1, ("%s: finished\n", func));
	return TC_OK;
}

int tc_handleVideoResize(struct tc_fb *fb, thinc_header *h)
{
	const char *func = __func__;
	thinc_vidResize resize;
	int res;

	struct tc_video_stream *stream;
	struct tc_video *vid = &fb->vid;

	tclient_dbg(1, ("%s: reading video scale message\n", func));
	res = tc_readMsgNoHdr(fb, T_VIDEO_RESIZE, &resize, T_VIDEO_RESIZE_SIZE);
	if(res == TC_ERROR) {
		tclient_msg("Error: %s: Couldn't read Video Resize message\n",
			func);
		return TC_ERROR;
	} else if(res == TC_CLOSED) {
		tclient_msg("Error: %s: Connection closed while waiting for "
			"video message\n", func);
		return TC_ERROR;
	}

	resize.id = ntohl(resize.id);
	resize.width = ntohs(resize.width);
	resize.height = ntohs(resize.height);

	if(resize.id >= vid->num_streams) {
		tclient_msg("Error: %s: Invalid stream id %d received\n", 
				func, resize.id);
		return TC_ERROR;
	}
	stream = &vid->streams[resize.id];
	if(!stream->image) {
		tclient_msg("Error: %s: Stream %d hasn't been initialized\n",
				func, resize.id);
		return TC_ERROR;
	}

	stream->width = resize.width;
	stream->height = resize.height;

	switch(stream->format) {
		case T_FOURCC_YV12:
			res = tc_video_readinfo_yv12(fb, stream);
			break;
		case T_FOURCC_YUY2:
			res = tc_video_readinfo_yuy2(fb, stream);
			break;
		case T_FOURCC_UYVY:
			res = tc_video_readinfo_uyvy(fb, stream);
			break;
		default:
			tc_error_msg("Unknown video format %#x received\n",
				stream->format);
			return TC_ERROR;
	}
	if(res == TC_ERROR) {
		tc_error_msg("Couldn't read new video info after video "
			"stream was resized\n");
		return TC_ERROR;
	}

	tclient_dbg(1, ("%s: changing orig size to %ux%u\n", func,
		stream->width, stream->height));

	/* Destroy current XvImage and create a new one with new dimensions */
	tc_video_destroy_xvimage(fb, stream);
	if(tc_video_create_xvimage(fb, vid, stream) == TC_ERROR) {
		tclient_msg("Error: %s: Couldn't create XV Image\n", func);
		return TC_ERROR;
	}

	tclient_dbg(1, ("%s: finished\n", func));
	return TC_OK;
}


/** next 4 functions used when the client's window is resized */
inline int tc_video_client_scale_all(struct tc_fb *fb, double factor_x, 
		double factor_y)
{
	int i;
	struct tc_video *vid = &fb->vid;

	for(i=0; i<vid->num_streams; ++i) {
		tc_video_client_scale(fb, i, factor_x, factor_y);
	}

	return TC_OK;
}

inline int tc_video_client_scale(struct tc_fb *fb, int id, double factor_x,
		double factor_y)
{
	struct tc_video *vid = &fb->vid;
	struct tc_video_stream *stream;

	stream = &vid->streams[id];
	stream->curr_width = MIN(CEILING(stream->dst_width * factor_x), 
			fb->curr_width);
	stream->curr_height = MIN(CEILING(stream->dst_height * factor_y),
			fb->curr_height);

	return TC_OK;
}

inline int tc_video_client_move_all(struct tc_fb *fb, double factor_x, 
		double factor_y)
{
	int i;
	struct tc_video *vid = &fb->vid;

	for(i=0; i<vid->num_streams; ++i) {
		tc_video_client_move(fb, i, factor_x, factor_y);
	}

	return TC_OK;
}

int tc_video_client_move(struct tc_fb *fb, int id, double factor_x,
		double factor_y)
{
	struct tc_video *vid = &fb->vid;
	struct tc_video_stream *stream;

	stream = &vid->streams[id];
	stream->curr_x = ((double)stream->x) * factor_x;
	stream->curr_y = ((double)stream->y) * factor_y;

	return TC_OK;
}

/***** format dependent functions *****/
/* YV12 */
static int tc_video_readinfo_yv12(struct tc_fb *fb, 
	struct tc_video_stream *stream)
{
	int err;
	thinc_vidInfoYV12 info_yv12;
	struct tc_video_yv12 *yv12 = &stream->fmt.yv12;

	err = readData(fb, &info_yv12, T_VIDEO_INFO_YV12_SIZE, T_TRUE);
	if(err == TC_ERROR || err == TC_CLOSED) {
		return TC_ERROR;
	}

	yv12->yuv_offsets[0] = ntohl(info_yv12.yuv_yoffset);
	yv12->yuv_offsets[1] = ntohl(info_yv12.yuv_uoffset);
	yv12->yuv_offsets[2] = ntohl(info_yv12.yuv_voffset);
	yv12->yuv_pitches[0] = ntohl(info_yv12.yuv_ypitch);
	yv12->yuv_pitches[1] = ntohl(info_yv12.yuv_upitch);
	yv12->yuv_pitches[2] = ntohl(info_yv12.yuv_vpitch);

	return TC_OK;
}

static int tc_video_init_image_yv12(struct tc_fb *fb, struct tc_video *vid,
	struct tc_video_stream *stream)
{
	tc_dbg_name(func)
	struct tc_video_yv12 *yv12 = &stream->fmt.yv12;

	tclient_dbg(1, ("%s: init yv12 image\n", func));

#ifndef NO_X
	tclient_dbg(1, ("width: %u, height %u, offsets %u %u %u => %d %d %d. "
		"pitches %u %u %u => %d %d %d\n", stream->width, 
		stream->height,
		yv12->yuv_offsets[0], 
		yv12->yuv_offsets[1],
		yv12->yuv_offsets[2],
		stream->image->offsets[0], 
		stream->image->offsets[1], 
		stream->image->offsets[2], 
		yv12->yuv_pitches[0],
		yv12->yuv_pitches[1],
		yv12->yuv_pitches[2],
		stream->image->pitches[0], 
		stream->image->pitches[1], 
		stream->image->pitches[2] 
	));

	if(	(stream->image->offsets[0] != yv12->yuv_offsets[0]) ||
		(stream->image->offsets[1] != yv12->yuv_offsets[1]) ||
		(stream->image->offsets[2] != yv12->yuv_offsets[2]) ||
		(stream->image->pitches[0] != yv12->yuv_pitches[0]) ||
		(stream->image->pitches[1] != yv12->yuv_pitches[1]) ||
		(stream->image->pitches[2] != yv12->yuv_pitches[2])
	  ) {
		tc_error_msg("Local YUV offsets and pitches don't "
			"match with those from the server\n");
		return TC_ERROR;
	}
#else
	stream->image = malloc(stream->width*stream->height +
			(2 * (stream->width >> 1) * (stream->height >> 1)) );
	if(!stream->image) {
		tclient_msg("Error: %s: Couldn't allocate data for video "
			"data\n", func);
		return TC_ERROR;
	}
#endif	/* ! NO_X */

	tclient_dbg(1, ("%s: finished\n", func));
	return TC_OK;
}

static void tc_video_destroy_image_yv12(struct tc_fb *fb, 
	struct tc_video_stream *stream)
{
	/* noop */
}

/* YUY2 */
static int tc_video_readinfo_yuy2(struct tc_fb *fb, 
	struct tc_video_stream *stream)
{
	int err;
	thinc_vidInfoYUY2 info_yuy2;
	struct tc_video_yuy2 *yuy2 = &stream->fmt.yuy2;

	err = readData(fb, &info_yuy2, T_VIDEO_INFO_YUY2_SIZE, T_TRUE);
	if(err == TC_ERROR || err == TC_CLOSED) {
		return TC_ERROR;
	}

	yuy2->pitch = ntohl(info_yuy2.pitch);

	return TC_OK;
}

static int tc_video_init_image_yuy2(struct tc_fb *fb, struct tc_video *vid,
	struct tc_video_stream *stream)
{
	tc_dbg_name(func)
	struct tc_video_yuy2 *yuy2 = &stream->fmt.yuy2;

	tclient_dbg(1, ("%s: init yuy2 image\n", func));

#ifndef NO_X
	tclient_dbg(1, ("width: %u, height %u, pitch %u => %d\n", 
		stream->width, stream->height,
		yuy2->pitch, stream->image->pitches[0]
	));

	if((stream->image->pitches[0] != yuy2->pitch)) {
		tc_error_msg("Local YUV pitch doesn't "
			"match with those from the server\n");
		return TC_ERROR;
	}
#else
	stream->image = malloc((stream->width<<1)*stream->height);
	if(!stream->image) {
		tclient_msg("Error: %s: Couldn't allocate data for video "
			"data\n", func);
		return TC_ERROR;
	}
#endif	/* ! NO_X */

	tclient_dbg(1, ("%s: finished\n", func));
	return TC_OK;
}

static void tc_video_destroy_image_yuy2(struct tc_fb *fb, 
	struct tc_video_stream *stream)
{
	stream->fmt.yuy2.pitch = 0;
}

/* UYVY */
static int tc_video_readinfo_uyvy(struct tc_fb *fb, 
	struct tc_video_stream *stream)
{
	int err;
	thinc_vidInfoUYVY info_uyvy;
	struct tc_video_uyvy *uyvy = &stream->fmt.uyvy;

	err = readData(fb, &info_uyvy, T_VIDEO_INFO_UYVY_SIZE, T_TRUE);
	if(err == TC_ERROR || err == TC_CLOSED) {
		return TC_ERROR;
	}

	uyvy->pitch = ntohl(info_uyvy.pitch);

	return TC_OK;
}

static int tc_video_init_image_uyvy(struct tc_fb *fb, struct tc_video *vid,
	struct tc_video_stream *stream)
{
	tc_dbg_name(func)
	struct tc_video_uyvy *uyvy = &stream->fmt.uyvy;

	tclient_dbg(1, ("%s: init uyvy image\n", func));

#ifndef NO_X
	tclient_dbg(1, ("width: %u, height %u, pitch %u => %d\n", 
		stream->width, stream->height,
		uyvy->pitch, stream->image->pitches[0]
	));

	if((stream->image->pitches[0] != uyvy->pitch)) {
		tc_error_msg("Local YUV pitch doesn't "
			"match with those from the server\n");
		return TC_ERROR;
	}
#else
	stream->image = malloc((stream->width<<1)*stream->height);
	if(!stream->image) {
		tclient_msg("Error: %s: Couldn't allocate data for video "
			"data\n", func);
		return TC_ERROR;
	}
#endif	/* ! NO_X */

	tclient_dbg(1, ("%s: finished\n", func));
	return TC_OK;
}

static void tc_video_destroy_image_uyvy(struct tc_fb *fb, 
	struct tc_video_stream *stream)
{
	stream->fmt.uyvy.pitch = 0;
}
